//===------------- DefUseSites_Fer.cpp - Printing prototype ---------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it stores the definition and use sites of virtual registers.
//
// Date: August 22th, 2007
//
//===----------------------------------------------------------------------===//
//
// This class implements a mapping of virtual registers to instructions.
// For each virtual, this class gives its definition site, which is unique,
// due to the SSA property, and all the use sites. A definition site is the
// index of the instruction where the register is defined, and a use site is
// the index of the instruction where the register is used.
// These indices are defined by the live interval analysis.
//
//===----------------------------------------------------------------------===//

#ifndef DEF_USE_SITES_FER_H
#define DEF_USE_SITES_FER_H

#include "llvm/PassSupport.h"
#include <set>
#include <iostream>

namespace llvm {

    class LiveIntervals;

    /// This class is used to order the variables according to the index of
    /// the site where they are used.
    class UseVec_Fer {
        public:

        UseVec_Fer() {
            ptr_ = 0;
            reg_ = 0;
            num_uses_ = 0;
        }

        ~UseVec_Fer() {
            indices_.clear();
        }

        unsigned current() const {
            if(ptr_ < indices_.size()) {
                return indices_[ptr_];
            } else {
                return 0;
            }
        }

        void insert(unsigned idx) {
            if(indices_.size() > 0) {
                if(indices_[indices_.size() - 1] != idx) {
                    indices_.push_back(idx);
                    num_uses_++;
                }
            } else {
                indices_.push_back(idx);
                num_uses_++;
            }
        }

        void advance(unsigned idx) {
            while(ptr_ < indices_.size()) {
                if(idx <= indices_[ptr_]) {
                    break;
                } else {
                    ptr_++;
                }
            }
        }

        unsigned end() {
            if(indices_.size() == 0) {
                return 0;
            } else {
                return indices_[indices_.size() - 1];
            }
        }

        unsigned num_uses_;

        unsigned reg_;

        unsigned ptr_;

        std::vector<unsigned> indices_;
    };


    inline std::ostream & operator<<(std::ostream & os, const UseVec_Fer & u) {
        os << u.reg_ << " @ [ ";
        for(unsigned i = 0; i < u.indices_.size(); i++) {
            os << u.indices_[i] << " ";
        }
        os << "]";
        return os;
    }

    class UseVecComp_Fer {
        public :
        bool operator()(const UseVec_Fer * x, const UseVec_Fer * y) const {
            unsigned x_ind = x->current();
            unsigned y_ind = y->current();
            if(x_ind != y_ind) {
                return x_ind < y_ind;
            } else {
                return x->reg_ < y->reg_;
            }
        }
    };

    class DefUseSites_Fer : public MachineFunctionPass {

        public:

        static char ID; // Pass identification, replacement for typeid
        DefUseSites_Fer() : MachineFunctionPass((intptr_t)&ID) {}

        /// reporst that this pass does not change anything on the control
        /// flow graph.
        virtual void getAnalysisUsage(AnalysisUsage &AU) const;

        /// Informs the name of the pass, so it can appear in tools like
        /// llc.
        virtual const char * getPassName() const {
            return "Virtual to def/use mapping - Fernando.";
        }

        /// Prints in the standard error device all the information contained
        /// in this data structure.
        virtual void dump() const;

        /// Prints in the standard error device all the information about one
        /// single register.
        void dump(unsigned v_reg) const;

        /// This method inserts all the uses of register v_reg into the queue
        /// of possible spill candidates
        void set_active(unsigned v_reg, unsigned v_end);

        // This method returns the index of the next use of virtual v_reg.
        // Notice that the DefUseSites class keeps a state that is the next
        // usage of each virtual register.
        unsigned next_usage(unsigned v_reg) const;

        /// This method removes all the uses of spill candidates that are
        /// in indices higher than idx
        void advance_index(unsigned idx);

        std::set<UseVec_Fer * , UseVecComp_Fer >::iterator begin() {
            return active_.begin();
        }

        std::set<UseVec_Fer * , UseVecComp_Fer >::iterator end() {
            return active_.end();
        }

        unsigned get_num_uses(unsigned v_reg) const;

        private:

        // The interval representation of the target program.
        LiveIntervals * lis_;

        /// Mapping of registers to definition sites. v2def_[i] is the index
        /// of the definition site of virtual i+FirstVirtual;
        std::vector< unsigned > v2d_;

        /// Mapping of registers to use sites.
        std::vector< UseVec_Fer > v2u_;

        /// This data structure is used to quicly obtain the next register
        /// that must be spilled.
        std::set<UseVec_Fer * , UseVecComp_Fer > active_;

        /// This method visits a basic block and collects all the uses and defs
        /// present in this basic block.
        void visit_basic_block(MachineBasicBlock & mbb);

        /// Free the memory, so the operating system can use it in other
        /// applications.
        void releaseMemory();

        /// runOnMachineFunction - collects the use/def sites of each register,
        /// and computes their weights.
        bool runOnMachineFunction(MachineFunction & mf);

        /// This instruction marks the instruction mi as the definition site of
        /// the virtual register v_reg.
        void set_def_site(MachineInstr * mi, unsigned v_reg);

        /// This method adds the instruction mi as one of the use sites of the
        /// virtual v_reg.
        void add_use_site(MachineInstr * mi, unsigned v_reg);

        /// This method computes the number of times each variable is used
        void compute_statistics();

    };  // end of EdgeLiveness_Fernando declaration

} // end llvm namespace

#endif
