//===------------- EqvPhiClass_Fer.h - Phi Equivalence Class --------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This class represents the equivalence class of phi-related virtuals.
//
// Date: Jun 30th, 2006
//
//===----------------------------------------------------------------------===//
//
// Two virtuals are said to be phi-related according to the following
// properties:
// (i - reflexivity) any virtual v is phi-related to itself;
// (ii - symmetry) if virtual v is phi-related to virtual u, then u is
// phi-related to v;
// (iii - transitivity) if v_1 is phi-related to v_2, and v_2 is phi-related
// to v_3, then v_1 is phi-related to v_3.
// (iv) given a phi-function such as v_i = phi(v_{i1}, v_{i2}, \ldots,
// v_{im}), the virtuals v_i, v_{i1}, v_{i2}, \ldots, v_{im} are
// phi-related;
//
//===----------------------------------------------------------------------===//

#ifndef PHI_EQV_CLASS_H
#define PHI_EQV_CLASS_H

#include <set>
#include <list>

namespace llvm {

  class LiveInterval;

  struct Line {
    unsigned start;
    unsigned end;
    Line(unsigned s, unsigned e) : start(s), end(e) {
        assert(s < e && "Cannot create empty or backwards interval.");
    }
    /// contains - Return true if the index is covered by this range.
    bool contains(unsigned I) const {
      return start <= I && I < end;
    }
    ///
    bool operator<(const Line & L) const {
      return end <= L.start;
    }
    ///
    bool overlap(const Line & L) const {
      return L.contains(start) || this->contains(L.start);
    }
  };

  class EqvPhiClass_Fer {

    /// The Lines data-type:
    typedef std::list<Line> Lines;
 
    public:
 
    ///
    ~EqvPhiClass_Fer() {
        lines.clear();
    }
 
    ///
    void dump() const;
 
    ///
    bool overlap(const EqvPhiClass_Fer & pec) const;
 
    ///
    void merge(EqvPhiClass_Fer & pec);
 
    ///
    void add(LiveInterval & li);

    ///
    bool overlap(LiveInterval & li);
 
    ///
    void add(unsigned start, unsigned end);
 
    typedef Lines::iterator iterator;
    iterator begin() { return lines.begin(); }
    iterator end()   { return lines.end(); }
 
    typedef Lines::const_iterator const_iterator;
    const_iterator begin() const { return lines.begin(); }
    const_iterator end() const   { return lines.end(); }

    std::set<unsigned> * get_virtuals() {
        return & this->virtuals;
    }
 
    private:
 
    ///
    std::set<unsigned> virtuals;
 
    /// The Lines data-type:
    Lines lines;

  };

}

#endif
