//===------------------- Partition_Fer - Pass template -------------------===//
//
//                     The LLVM Compiler Infrastructure
//
// Performs register allocation by partitioning the CFG
//
//===---------------------------------------------------------------------===//
//
// This pass implements a register allocator that partitions the live ranges
// between each instruction, so that the interference graph of the transformed
// program will be a collection of clique substitutions of P_3. This family of
// graphs admits optimum coloring, even in the presence of precoloring and
// aliasing.
//
//===---------------------------------------------------------------------===//

#ifndef MACHINE_DOM_TREE_H
#define MACHINE_DOM_TREE_H

#include "llvm/PassSupport.h"

namespace llvm {

class MachineDomTree_Fer : public MachineFunctionPass {

  public:

    static char ID; // Pass identification, replacement for typeid
    MachineDomTree_Fer() : MachineFunctionPass((intptr_t)&ID) {}

    // This method specifies which other passes are required, or are modified
    // by this pass.
    virtual void getAnalysisUsage(AnalysisUsage & AU) const;

    // Prints the list of machine basic blocks and their immediate dominators.
    void print_dominators() const;

    // This method prints the dominance tree.
    void print_dom_tree() const;

    // This method returns the immediate dominator of the basic block which has
    // number mbb_num.
    unsigned get_idom(const unsigned mbb_num) const {
        assert(mbb_num < this->rtree_.size() && "Illegal basic block.");
        return this->rtree_[mbb_num];
    }

  private:

    /// This matrix will contain the dominators of each basic block.
    std::vector<std::vector<bool> > dominators_;

    /// This vector contains the reverse dominance tree: rtree[x] contains the
    /// node that dominates x.
    std::vector<unsigned> rtree_;

    /// The bank of register keeps a reference to the machine function in
    /// order to get the architecture specific information.
    MachineFunction * mf_;

    /// Returns true if mbb1 dominates mbb2, and false otherwise. If these
    /// basic blocks are not in the same dominance chain, the method throws an
    /// exception.
    bool dominates(unsigned mbb1, unsigned mbb2) const {
        if(dominators_[mbb2][mbb1]) {
            return true;
        }
        assert(dominators_[mbb1][mbb2] &&
                          "Basic blocks are not in the same dominance chain.");
        return false;
    }

    /// Visits the basic blocks and lists of intervals assigning registers to
    /// virtual variables.
    virtual bool runOnMachineFunction(MachineFunction & mf);

    /// Free the memory, so the operating system can use it in other
    /// applications.
    void releaseMemory();

    // This method performs one iteration of the simple algorithm to compute
    // dominators. The method returns true if the iteration has changed the
    // information that was previously known.
    bool execute_one_iteration();

    // This method builds the dominance tree of the target function.
    void build_dominance_tree();

};

}

#endif
