//===----------------- NeatPrint_Fer.cpp - Layout output ------------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This class contains some handy functions to generate strings describing
// instructions, register, etc.
//
// Date: June 17th, 2006
//
//===----------------------------------------------------------------------===//
//
// This file contains functions that produce strings describing instructions,
// operands, and other data structures of the LLVM platform.
//
//===----------------------------------------------------------------------===//

#ifndef NEAT_PRINTER_FER
#define NEAT_PRINTER_FER
 
#include <iostream>                              // cerr
#include <sstream>                               // ostringstream (RegClique)

#include "llvm/Function.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"

namespace llvm {

class NeatPrint_Fer {

public:


//===--------------------------------------------------------------------------
// This method prints two types of informations:
// 1) A description of the register classes present in the target architecture.
// 2) The name of each physical register.
//===--------------------------------------------------------------------------
static void print_reg_info(const MachineFunction & mf) {

    const TargetMachine & target_machine = mf.getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();

    // First, prints the information about the register classes present in
    // the target architecture:
    MRegisterInfo::regclass_iterator rci = reg_info->regclass_begin();
    MRegisterInfo::regclass_iterator rce = reg_info->regclass_end();
    while(rci != rce) {
        const TargetRegisterClass * trc = *rci;

        std::cout << "-----------------------------------------------------\n";
        std::cout << "ID: " << trc->getID() << "\n";
        std::cout << "Num regs: " << trc->getNumRegs() << "\n";
        std::cout << "Size: " << trc->getSize() << "\n";
        std::cout << "Alignment: " << trc->getAlignment() << "\n";
        TargetRegisterClass::iterator trci = trc->allocation_order_begin(mf);
        TargetRegisterClass::iterator trce = trc->allocation_order_end(mf);
        for (; trci != trce; ++trci) {
            std::cout << *trci << " - ";
        }
        std::cout << std::endl;

/*
        std::cout << std::endl << "<TR>" << std::endl;
        std::cout << "<TD>" << trc->getID() << "</TD>\n";
        std::cout << "<TD>" << trc->getNumRegs() << "</TD>\n";
        std::cout << "<TD>" << trc->getSize() << "</TD>\n";
        std::cout << "<TD>" << trc->getAlignment() << "</TD>\n";
*/
        rci++;
    }

    // Second, prints the name and number of each physical register:
    for(unsigned p_reg = 1; p_reg < reg_info->getNumRegs(); p_reg++) {
        std::ostringstream reg_text;
        reg_text << p_reg << ": " << reg_info->getName(p_reg);
        std::cout.width(16);
        std::cout << reg_text.str();
    }
    std::cout << std::endl;
}


//===--------------------------------------------------------------------------
/// Returns a string describing the name of the register. If it is a virtual
/// register, it will be printed in the format tn, e.g. t0, t1, ..., where n
/// is given by (reg - MRegisterInfo::FirstVirtualRegister). If it is physical
/// register, it will be printed in the format reg_name(reg), where reg_name is
/// the register's name as seen by the target architecture, e.g EAX, r1, etc.
//===--------------------------------------------------------------------------
static std::string op2string(const unsigned reg, const MachineFunction & mf) {
    std::ostringstream reg_name;
    if(MRegisterInfo::isVirtualRegister(reg)) {
        reg_name << "t" << (reg - MRegisterInfo::FirstVirtualRegister);
    } else {
        const TargetMachine & target_machine = mf.getTarget();
        const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
        reg_name << reg_info->getName(reg);
    }
    reg_name << "(" << reg << ")";
    return reg_name.str();
}


//===--------------------------------------------------------------------------
/// Returns a string describing the name of the instruction.
//===--------------------------------------------------------------------------
static std::string opcode2string
                        (const MachineInstr & mi, const MachineFunction & mf) {
    const TargetMachine & tm = mf.getTarget();
    const TargetInstrInfo * tii = tm.getInstrInfo();
    const TargetInstrDescriptor & tid = tii->get(mi.getOpcode());
    std::ostringstream aux;
    aux << tid.Name;
    return aux.str();
}


//===--------------------------------------------------------------------------
/// Returns a string describing the instruction opcode, plus all the operands
/// that are part of the instruction.
//===--------------------------------------------------------------------------
static std::string mi2string(const MachineInstr & mi) {
    const MachineFunction & mf = * mi.getParent()->getParent();
    std::ostringstream mi_str;
    mi_str << NeatPrint_Fer::opcode2string(mi, mf);
    unsigned i = 0;
    while(i != mi.getNumOperands()) {
        const MachineOperand & mo = mi.getOperand(i);
        if (mo.isRegister() && mo.getReg() && mo.isDef()) {
            mi_str << " " << op2string(mo.getReg(), mf);
        } else if(mo.isRegister() && mo.getReg() && mo.isUse()) {
            break;
        }
        i++;
    }
    mi_str << " := ";
    while(i != mi.getNumOperands()) {
        const MachineOperand & mo = mi.getOperand(i);
        if(mo.isRegister() && mo.getReg()) {
            mi_str << " " << op2string(mo.getReg(), mf);
        }
        i++;
    }
    return mi_str.str();
}


//===--------------------------------------------------------------------------
/// Returns a string describing the instruction opcode, plus all the operands
/// that are part of the instruction.
//===--------------------------------------------------------------------------
static std::string mi2string
                        (const MachineInstr & mi, const MachineFunction & mf) {
    std::ostringstream mi_str;
    mi_str << NeatPrint_Fer::opcode2string(mi, mf);
    unsigned i = 0;
    while(i != mi.getNumOperands()) {
        const MachineOperand & mo = mi.getOperand(i);
        if (mo.isRegister() && mo.getReg() && mo.isDef()) {
            mi_str << " " << op2string(mo.getReg(), mf);
        } else if(mo.isRegister() && mo.getReg() && mo.isUse()) {
            break;
        }
        i++;
    }
    mi_str << " := ";
    while(i != mi.getNumOperands()) {
        const MachineOperand & mo = mi.getOperand(i);
        if(mo.isRegister() && mo.getReg()) {
            mi_str << " " << op2string(mo.getReg(), mf);
        }
        i++;
    }
    return mi_str.str();
}


//===--------------------------------------------------------------------------
/// This method prints the header information of the dot graph. This
/// information defines the font used for nodes and edges, plus the shape of
/// nodes. In this case, nodes have rectangular shape, and the font is
/// "Courier".
//===--------------------------------------------------------------------------
static void print_graph_header(const MachineFunction & mf) {
    // first prints he format of each node.
    const Function * f = mf.getFunction();
    std::cerr << "digraph " << f->getName() << " {\n\n";
    std::cerr << "node [\n";
    std::cerr << "  fontname = \"Courier\"\n";
    std::cerr << "  fontsize = \"10\"\n";
    std::cerr << "  shape = \"box\"\n";
    std::cerr << "]\n\n";

    // second prints he format of each edge.
    std::cerr << "edge [\n";
    std::cerr << "  fontname = \"Courier\"\n";
    std::cerr << "  fontsize = \"8\"\n";
    std::cerr << "]\n\n";
}


//===--------------------------------------------------------------------------
/// Prints the cfg of the given machine function using the dot format.
//===--------------------------------------------------------------------------
static void print_cfg(const MachineFunction & mf) {
    print_graph_header(mf);
    for (MachineFunction::const_iterator mbb = mf.begin(), mbbe = mf.end();
                                                          mbb != mbbe; ++mbb) {
        print_edges(*mbb);
        print_block_label(mf, *mbb);
    }
    std::cerr << "}\n";
}


//===--------------------------------------------------------------------------
/// Prints all the edges that leave the given basic block. The output format
/// is dot. The label of each basic block is its block number; thus, an edge
/// is like "n1" -> "n2", where n1 and n2 are integer non-negative numbers.
//===--------------------------------------------------------------------------
static void print_edges(const MachineBasicBlock & mbb) {
    unsigned mbb_num = mbb.getNumber();
    for(MachineBasicBlock::const_succ_iterator succ = mbb.succ_begin();
                                              succ != mbb.succ_end(); succ++) {
        const MachineBasicBlock* mbb_succ = *succ;
        unsigned succ_num = mbb_succ->getNumber();
        std::cerr << "\"" << mbb_num << "\" -> \"" << succ_num << "\"\n";
    }
}


//===--------------------------------------------------------------------------
/// Prints the label of the basic block in the control flow graph description.
/// The label of the basic block includes the block number, and all the
/// instructions that are in the block, in the sequence in which they are
/// executed.
//===--------------------------------------------------------------------------
static void print_block_label
                  (const MachineFunction & mf, const MachineBasicBlock & mbb) {
    unsigned block_number = mbb.getNumber();
    std::cerr << "\"" << block_number << "\"" << " [ label=\"* "
              << block_number << " *\\l";

    MachineBasicBlock::const_iterator mii = mbb.begin();
    while (mii != mbb.end()) {
        std::cerr << NeatPrint_Fer::mi2string(*mii, mf) << " \\l";
        mii++;
    }
    std::cerr << "\"]\n";

}

private:

};

}

#endif
