//===--------------- P3_Fer.h - Clique substitution of P_3 ----------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This class represents a clique substitution of P_3
//
// Date: May 27th, 2007
//
//===----------------------------------------------------------------------===//
//
// Clique substitutions of P_3 are graphs that can be divided into three sets
// of nodes: X, Y and Z. Nodes in X, Y and Z form cliques. Also, nodes of X and
// Y are pairwise adjance. The same is true for nodes of Y and Z. However,
// nodes of X and Z are not adjacent. Thus, any of these graphs has two maximal
// cliques: X union Y and Y union Z.
// Notice that this class has only been implemented for the X86 architecture.
//
//===----------------------------------------------------------------------===//


#ifndef P3_FER_H
#define P3_FER_H

#include <set>
#include "FacadeP3_Fer.h"

namespace llvm {

  #define NUM_USES 4

  class vector;
  class MachineInstr;
  class Partition_Fer;
  class RA_Result_Fer;
  class RegMapping_Fer;
  class PhiAnalysis_Fer;
  class MachineFunction;

  class P3_Fer : public FacadeP3_Fer {

    public:

    struct State {
        enum {
            X = 0,
            Y = 1,
            Z = 2
        };
    };

    /// Constructor method.
    P3_Fer(
        MachineFunction & mf,
        MachineInstr * mi,
        Partition_Fer *partition,
        RegMapping_Fer *rm,
        RA_Result_Fer *rs,
        PhiAnalysis_Fer * pa
    );

    /// Destructor method.
    ~P3_Fer() {
        this->Xs_.clear();
        this->Ys_.clear();
        this->XL_.clear();
        this->YL_.clear();

        this->othersX_.clear();
        this->othersY_.clear();

        this->p_regs_X_.clear();
        this->p_regs_Z_.clear();

        this->fixed_regs_.clear();
        this->allocable_regs_.clear();
        this->in_X_.clear();
        this->in_Y_.clear();
    }

    /// Add 'reg' as a node of component X
    void addX(unsigned reg);

    /// Add 'reg' as a node of component Y
    void addY(unsigned reg);

    /// Add 'reg' as a node of component Z
    void addZ(unsigned reg);

    /// The number of virtuals in this partition:
    unsigned size() const {
        return this->size_Xs_ + this->Ys_.size() + this->Zs_ +
               this->size_XL_ + this->YL_.size() + this->ZL_ +
               this->size_othersX_ + this->othersY_.size() + this->othersZ_;
    }

    // This method prints the location of each variable that is alive at the
    // beginning of the program point that this graph represents.
    void print_in_set() const;

    /// This method colors this clique substitution of P_3 with the free
    /// registers of this object. The array preferences associates virtual
    /// variables to colors. This vector determines a color assignment that
    /// would reduce the number of copies in the final assembly code. Given
    /// two successive KP3 (Clique substitution of P_3), q1 and q2, the
    /// color assignment in q1 is used to build the preference vector that will
    /// be passed to q2. This method returns the machine instruction that was
    /// visited. It is possible that this is not the same instruction passed
    /// to the object constructor because of instruction folding.
    MachineInstr * color();

    /// This method inserts copies/loads/stores at the program point that this
    /// graph covers in order to preserve the semantics of the program.
    void transfer_colors();

    // This method return true if the graph has precolored nodes in any of its
    // maximal cliques.
    bool contains_precol() const {
        return this->num_precol_ > 0;
    }

    // This method return true if the graph has nodes of type short in any of
    // its maximal cliques.
    bool contains_short() const {
        return this->size_Xs_ > 0 || this->Ys_.size() > 0 || this->Zs_ > 0;
    }
   
    // This method return true if the graph has nodes of type long in any of its
    // maximal cliques.
    bool contains_long() const {
        return this->size_XL_ > 0 || this->YL_.size() > 0 || this->ZL_ > 0;
    }

    // This method returns true if the contents of p_reg can be evicted to
    // memory. The only situations when this cannot happen is if p_reg is
    // precolored, or if p_reg is holding a virtual that is used in the current
    // instruction.
    bool can_spill(unsigned p_reg) const;

    unsigned get_num_iterations() const {
        return numIter_;
    }

    /// Prints the assignment of physical registers to virtuals in that exist
    /// in the X component, before this object instruction has been executed.
    void dump_X_colors() const;

    /// Prints the assignment of physical registers to virtuals in the Z
    /// component, that is, after the instruction that this object surrounds
    /// has been executed.
    void dump_Z_colors() const;

    private:

    /// The reference to the target architecture. This class has only been
    /// implemented for the X86 architecture.
    MachineFunction * mf_;

    /// Each of these graphs represent the live ranges that cross one
    /// instruction. This is a reference to that instruction.
    MachineInstr * mi_;

    /// This object is effectively the register allocator. It is called to
    /// choose a register to spill.
    Partition_Fer *partition_;

    // This class represents the mappings between virtual variables and
    // physical registers.
    RegMapping_Fer * rm_;

    // The result of the register allocation.
    RA_Result_Fer * rs_;

    // The phi analysis is necessary for implementing global coalescing.
    PhiAnalysis_Fer *pa_;

    /// Number of times the coloring algorithm had to spill variables.
    unsigned numIter_;

    /// The reigsters used in mi_. The set makes it more efficient
    /// to query these registers.
    std::set<unsigned> fixed_regs_;

    /// This set stores the physical registers than contains virtuals that can
    /// be spilled if necessary  to make room for other virtuals.
    std::vector<bool> allocable_regs_;

    /// Shorts in the X component.
    SmallVector<unsigned, NUM_USES> Xs_;
    unsigned size_Xs_;

    /// Longs in the X component.
    SmallVector<unsigned, NUM_USES> XL_;
    unsigned size_XL_;

    /// Shorts in the Y component.
    std::set<unsigned> Ys_;

    /// Longs in the Y component.
    std::set<unsigned> YL_;

    /// Shorts in the Z component.
    unsigned Zs_;

    /// Longs in the Z component.
    unsigned ZL_;

    /// Registers that have classes in which aliasing does not apply should be
    /// placed in these vectors:
    SmallVector<unsigned, NUM_USES> othersX_;
    unsigned size_othersX_;

    std::set<unsigned> othersY_;

    unsigned othersZ_;

    /// This vector associates physical registers with the virtual registers
    /// that have been allocated into each physical. For instance, if
    /// p_regs_[1] == 1024, then the virtual %reg1024 has been stored into
    /// physical register 'AH', which correspond to index 1.
    std::vector<unsigned> p_regs_X_;
    std::vector<unsigned> p_regs_Z_;

    /// These maps store the variables that are live in at the program point
    /// that this graph represents. Each variable is mapped to its location.
    std::map<unsigned, unsigned> in_X_;
    std::map<unsigned, unsigned> in_Y_;

    /// This variable will hold the number of precolored nodes in this graph.
    /// It is a way to answer the query contains_precol() quickly.
    unsigned num_precol_;

    /// The state can be either X, Y or Z, depending on which part of the
    /// puzzle board is being colored.
    unsigned state_;

    // This method return true if p_reg is neither a precolored, nor a fixed
    // register.
    bool is_spillable(unsigned p_reg) const;

    // This method colors the registers in the given set X. There are three
    // such sets: shorts, longs and others.
    // All the elements in X sets are fixed, and, once colored, they must be
    // removed from the fixed list.
    void color_set_X(SmallVector<unsigned, NUM_USES> & vec, unsigned size);

    // 1) The variable V is in register and its color can be reused:
    // - do it
    //
    // 2) Else it is possible to transfer V to another register in X/Z:
    // - do it
    //
    // 3) Else V is in register, its color X can be used, but Z can't:
    // - Use the color X, and spill V
    //
    // 4) Else the variable is in register, but its colors X/Z can't be used:
    // - Else, spill some variable V' and use give its color to V.
    void handle_fixed_Y(unsigned v_reg);

    // This method colors the registers in the given Y set. The vector
    // "preference" guides the coloring. Every time it is possible, the method
    // trys to assign to an element 'e' of "set" the color 'c' that is assigned
    // to 'e' in "preference".
    // The only situation in which it is not possible is if 'c' is precolored in
    // this graph.
    void color_fixed_Y(std::set<unsigned> & vec);

    // This method colors the registers in the given Y set. The vector
    // "preference" guides the coloring. Every time it is possible, the method
    // trys to assign to an element 'e' of "set" the color 'c' that is assigned
    // to 'e' in "preference".
    // The only situation in which it is not possible is if 'c' is precolored
    // in this graph.
    void color_set_Y(std::set<unsigned> & set);

    // This method frees the colors used by the elements in "set". Colors are
    // freed by invoking the method free_color on the RegMapping object.
    void free_color(SmallVector<unsigned, NUM_USES> & vec, unsigned size) const;

    // This method colors nodes in the set Z. These nodes must not have been
    // assigned colors before. If this happens, an exception is thrown. After
    // coloring a node 'v' with color 'c', the method makes 'c' as the
    // preference for 'v' in the "preference" vector.
    void color_set_Z(unsigned v_reg);

    // This method marks the given register as precolored. If p_reg was being
    // used to store any virtual, the color of this virtual will be reseted.
    void mark_precol(unsigned p_reg, std::vector<unsigned> & vec);

    // This method attempts to coalesce the instruction of this graph. If the
    // instruction is a move, and the coalescing is possible, the method
    // returns true, otherwise it returns false.
    bool coalescible_move();

    // This method assigns the physical register p_reg to the variable v_reg in
    // the vectors rb1 and rb2, but only if p_reg is not in use by any other
    // register.
    void test_and_set_color(
        unsigned v_reg,
        std::vector<unsigned> & rb1,
        std::vector<unsigned> & rb2
    );

    // This method calls the necessary routines to spill v_reg.
    unsigned spill_variable(unsigned v_reg, bool self_spill);

  };

}

#endif
