//===------------------- Partition_Fer - Pass template -------------------===//
//
//                     The LLVM Compiler Infrastructure
//
// Performs register allocation by partitioning the CFG
//
//===---------------------------------------------------------------------===//
//
// This pass implements a register allocator that partitions the live ranges
// between each instruction, so that the interference graph of the transformed
// program will be a collection of clique substitutions of P_3. This family of
// graphs admits optimum coloring, even in the presence of precoloring and
// aliasing.
//
//===---------------------------------------------------------------------===//


#ifndef PARTITION_H
#define PARTITION_H

#include <set>
#include <queue>

#include "llvm/PassSupport.h"
#include "Line_Fer.h"

namespace llvm {

#define NO_DELIM ~0U

// These constants are used to classify the interference graph at each program
// point of the target function. The constants SHRT, LONG and STLG are
// exclusive. The constants PRCL and UNCL are exclusive. Constants of different
// sets are orthogonal, that is, a graph can have pre-colored nodes that are
// all short, all long or mixed. The same for a graph without fixed-colors.

// The graph contains only short nodes:
#define SHRT  1
// The graph contains only long nodes:
#define LONG  2
// The graph contains short and long nodes:
#define STLG  3
// The graph contains pre-colored nodes:
#define PRCL  4
// The graph does not contain pre-colored nodes:
#define UNCL  8

class P3_Fer;
class Line_Fer;
class FacadeP3_Fer;
class MRegisterInfo;
class LiveIntervals;
class RA_Result_Fer;
class RegMapping_Fer;
class LineCompareEnd;
class DefUseSites_Fer;
class PhiAnalysis_Fer;
class MachineDomTree_Fer;


class Partition_Fer : public MachineFunctionPass {
  public:

    static char ID; // Pass identification, replacement for typeid
    Partition_Fer() : MachineFunctionPass((intptr_t)&ID) {}

    // This method specifies which other passes are required, or are modified
    // by this pass.
    virtual void getAnalysisUsage(AnalysisUsage & AU) const;

    // This method chooses, among the intervals in the active set, the best
    // candidate for spilling, and then send it to memory. It is used as a
    // criterion for choosing a good register to spill. The method returns the
    // physical register that was previously used by the spilled register.
    unsigned choose_reg_and_spill(
        unsigned v_reg,
        FacadeP3_Fer & p3_graph,
        std::vector<unsigned> & sp_regs,
        bool self_spill
    ) const;

    // This method prints the mappings of variables to locations that will be
    // stored at the end of each basic block once the function has been
    // traversed.
    void print_bb_boundaries() const;

    // This method prints the mapping that must be joined due to the presence of
    // phi-functions in the target control flow graph.
    void print_join_points() const;

    // This method returns the mapping of virtuals to locations at the end of
    // basic block mbb_num, once register allocation has been performed.
    const std::map<unsigned, unsigned> & get_tail_map(unsigned mbb_num) const {
        assert(mbb_num < bb2Pref_.size() && "Invalid basic block");
        return bb2Pref_[mbb_num];
    }

    // This method returns the mapping of virtuals to locations at the program
    // point immediately after the definitios in phi-functions.
    const std::map<unsigned, unsigned> & get_head_map(unsigned mbb_num) const {
        assert(mbb_num < bb_phi_.size() && "Invalid basic block");
        return bb_phi_[mbb_num];
    }

    // Returns a reference to the register mapping used by this object. This
    // reference contains the mapping between spilled virtuals and memory
    // addresses.
    RegMapping_Fer * get_register_mapping() const {
        return rm_;
    }

    /// Returns a reference to the internal RA_Result_Fer object
    RA_Result_Fer * get_RA_result() {
        return rs_;
    }

    /// Returns the point in the interval list that is been currently scanned.
    unsigned get_index() const {
        return c_indx_;
    }

    // This method implements instruction folding. The value in slot must be
    // loaded into register, or the value in register must be stored in the
    // memory slot. Instead of performing the reload directly, this method
    // attempts to replace the instruction old_mi by an equivalent instruction
    // that accesses memory directly. The parameter op_num denotes the operand
    // number of the register in which the instruction must be loaded.
    MachineInstr * foldMemoryOperand(
        MachineInstr * old_mi,
        unsigned op_num,
        int slot
    );

  private:

    // This data structure maps basic blocks to preference maps. A preference
    // map is a mapping of variables to the physical registers that they should
    // be possibily assigned, e.g assume v_reg == 1025, and v_reg is allocated
    // to physical BL. Then bb2Pref_[1025-1] == bb2Pref_[1] == BL. If the
    // location of v_reg were a stack slot, then bb2Pref_[1] == RF_SPILL.
    std::vector< std::map<unsigned, unsigned> > bb2Pref_;

    // This data structure associates basic blocks that contain phi-functions
    // with the mapping of live variables after all the variables defined by
    // phi-functions have gotten a location. It is used to destroy phi
    // functions.
    std::vector< std::map<unsigned, unsigned> > bb_phi_;

    // Makes a priority queue of Line_Fer. The priority is defined by the
    // operator '>' from Line_Fer. So, the interval with the lowest starting
    // point will be removed first.
    std::priority_queue<Line_Fer *, std::deque<Line_Fer *>,
                                            greater_ptr<Line_Fer> > unhandled_;

    // This vector marks the instructions tat start basic blocks.
    // if headers_[u] != NO_DELIM, the instruction of index u*NUM is the first
    // in a basic block, otherwise it is not.
    std::vector<unsigned> headers_;

    // Each position of this vector corresponds to an instruction. If this
    // position holds a value b different from NO_DELIM, then b is the index of
    // a basic block that the corresponding instruction finishes.
    std::vector<unsigned> terminators_;

    // The result of the register allocation.
    RA_Result_Fer * rs_;

    // This class represents the mappings between virtual variables and
    // physical registers.
    RegMapping_Fer * rm_;

    // Makes another priority queue containing Line_Fer. The priority is
    // defined by the LineCompareEnd class. Thus, the interval with the lowest
    // end will be removed first.
    std::set<Line_Fer * , LineCompareEnd > active_;

    // The interval representation of the target program.
    LiveIntervals * live_intervals_;

    /// The bank of register keeps a reference to the machine function in
    /// order to get the architecture specific information.
    MachineFunction * mf_;

    /// This object contains informations about the bank of registers of the
    /// target architecture.
    const MRegisterInfo * ri_;

    /// This data structure holds the dominance tree of the target machine
    /// function.
    MachineDomTree_Fer * dt_;

    /// This variable maps virtual registers to the instructions where they are
    /// defined or used.
    DefUseSites_Fer * du_;

    /// This variable ranges over the indices of instructions.
    unsigned c_indx_;

    /// This class collects information about which nodes are related by phi
    /// functions, in order to allow coalescing during register allocation;
    PhiAnalysis_Fer * pa_;

    // This method makes the interval lf to be active. An active interval has
    // been scanned by the register allocator.
    void activate_interval(Line_Fer *lf);

    /// Free the memory, so the operating system can use it in other
    /// applications.
    void releaseMemory();

    // This method builds the preference list of a basic block with null
    // instructions.
    void build_preference_list
                       (MachineBasicBlock * mbb, MachineBasicBlock * succ_mbb);

    // This method fills the data structures with the data that they must have
    // in order to start the processing of intervals.
    void initialize_data_structures(MachineFunction & mf);

    // This method scans all the intervals, divide them into sets of clique
    // substitutions of P_3 and assign registers to them.
    void scan_intervals();

    // This method process all the intervals that exist between two consecutive
    // program points, or crossbars. This name, crossbar, was taken from David
    // Koes PLDI paper, Incremental Register Allocation.
    void process_intervals(P3_Fer & p3_graph);

    // This method removes the top intervals from unhandled_ that are in the
    // range of curr_index and add them to active_.
    void remove_top_intervals();

    // This method marks in the header_ vector which instructions are the
    // first instruction in each basic block.
    void mark_delimiters();

    // If the instruction at index is the header of a basic block b, this
    // method returns the index of b, otherwise it returns NO_DELIM.
    unsigned getHeadedBB(unsigned index) const;

    // If the instruction at index is the terminator of a basic block b, this
    // method returns the index of b, otherwise it returns NO_DELIM.
    unsigned getTerminatedBB(unsigned index) const;

    // This method refills the preference vector with the registers in use at
    // the end of basic block 'b'.
    void reset_preference(unsigned b);

    // Visits the basic blocks and lists of intervals assigning registers to
    // virtual variables..
    virtual bool runOnMachineFunction(MachineFunction & mf);

    // This method is used to find liveness information for each basic block
    // that has no instructions. For these basic blocks, the method
    // scan_intervals will not assign a liveness mapping. This mapping is
    // necessary to destroy phi-functions.
    void store_liveness_for_null_blocks();

    // This method computes the type of the clique substitution of P3, based on
    // the types of each of its nodes: short/long or precolored/uncolored.
    // This method returns 0 if, and only if, the puzzle is empty.
    unsigned compute_statistics(P3_Fer & p3_graph);

    // This method prints all the variables defined/used in the instruction
    // mi in a format suitable for debugging.
    void print_instruction(MachineInstr * mi) const;
};

}

#endif
