//===-------------- PhiAnalysis_Fer.cpp - Printing prototype --------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it groups virtual registers that are related by phi functions.
//
// Date: August 1th, 2006
//
//===----------------------------------------------------------------------===//
//
// This class groups together virtual registers that are related by phi
// functions in the control flow graph. Such virtuals describe the same
// variable in the original program. A good register allocator should try to
// assign the same physical register to these virtuals. For instance, if
// v1 := PHI (v2, v3, v3), then {v1, v2, v3} are phi related virtuals.
//
//===----------------------------------------------------------------------===//

#ifndef PHI_ANALYSIS_H
#define PHI_ANALYSIS_H

#include <set>

#include "llvm/PassSupport.h"
#include "llvm/ADT/SmallVector.h"

namespace llvm {

  class DefUseSites_Fer;
  class MachineFunction;
  class MachineInstr;
  class PhiMemCoalescer_Fer;
  class SeparatePhiLiveRanges_Fer;

  /// I am assuming that no more than six registers are used. If more, then
  /// some will not be recorded in the class.
  #define SIZE 7
  class Preference {
    typedef SmallVector<unsigned,SIZE> SVec;
    public:
    /// The physical register that has been used most often in the phi-class
    unsigned preference;
    /// Constructor:
    Preference() {
      preference = 0;
      pref_occur = 0;
      physicals.resize(SIZE, 0);
      occurrences.resize(SIZE, 0);
    }
    /// Destructor:
    ~Preference() {
      physicals.clear();
      occurrences.clear();
    }
    /// Adds p_reg to the preference list
    void add(unsigned p_reg) {
      for(unsigned u = 0; u < SIZE; u++) {
        if(physicals[u] == 0 || physicals[u] == p_reg) {
          physicals[u] = p_reg;
          occurrences[u]++;
          unsigned aux = occurrences[u];
          if(aux > pref_occur) {
            preference = physicals[u];
            pref_occur = aux;
          }
          break;
        }
      }
    }
    private:
    /// The number of times the physical register preference was used to
    /// allocate members of the phi-family.
    unsigned pref_occur;
    /// This vector contains the physical registers used to color the
    /// equivalence class. 
    SVec physicals;
    /// For each i, 1 <= i <= 8, this vector holds the number of times the
    /// register in physicals[i] was used to color a phi-function.
    SVec occurrences;
  };

  class PhiAnalysis_Fer : public MachineFunctionPass {

    public:

    static char ID; // Pass identification, replacement for typeid
    PhiAnalysis_Fer() : MachineFunctionPass((intptr_t)&ID) {}


    /// reporst that this pass does not change anything on the control flow
    /// graph.
    virtual void getAnalysisUsage(AnalysisUsage & AU) const;


    /// Informs the name of the pass, so it can appear in tools like llc.
    virtual const char * getPassName() const {
        return "Phi Analysis - Fernando.";
    }


    /// This method points that the physical p_reg has been assigned to the
    /// virtual v_reg. This contributes to set p_reg as the prefered register
    /// of any virtual that is phi-related to v_reg.
    void set_preference(unsigned v_reg, unsigned p_reg);


    /// Returns true if the virtual register is phi-related, and has preference
    /// for some particular physical register.
    bool has_preference(unsigned v_reg) const;


    /// This method returns the physical register that should be assigned to
    /// v_reg if possible.
    unsigned get_preference(unsigned v_reg) const;


    /// This method returns true if v_reg is part of a phi function somewhere
    /// in the control flow graph, and false otherwise. 
    bool is_phi_related(unsigned v_reg) const;


    /// Prints the set of phi-related virtuals
    void dump();

  
    /// This method maps the data from the PhiMemCoalesce analysis into this
    /// analysis:
    void get_new_data(PhiMemCoalescer_Fer * pmc);


    /// This method returns true if the given virtual (or, for the same effect,
    /// any other phi-related virtual) has been mapped to a stack slot.
    bool has_stack_slot(unsigned v);


    /// This method returns the stack slot that has been assigned to virtual
    /// v, if it has been spilled to memory. All the phi-related virtuals have
    /// the same stack slot.
    int get_stack_slot(unsigned v);


    /// This method maps the virtual v to the stack slot s. It also has the
    /// effect of assigning s to all other virtuals that are phi-related to v.
    /// If v has already been mapped to a stack slot, the function will produce
    /// an error.
    void map_to_stack_slot(unsigned v, int s);


    /// This method removes the phi-function associated to old_v, which must be
    /// unique, and makes it associated to new_v.
    void update_use_list(unsigned old_v, unsigned new_v);


    /// This method removes the phi-function associated to old_v, which must be
    /// unique, and makes it associated to new_v. The only difference of this
    /// method to 'update_use_list' is that in this case old_v is expected to be
    /// defined by a phi-function.
    void update_def_list(unsigned old_v, unsigned new_v);


    /// This method returns the set of virtuals that are phi-related to v_reg
    std::set<unsigned> & get_phi_group(unsigned v_reg);


    /// This method returns the phi-function that defines the virtual, and
    /// NULL if the virtual is not define in a phi-function
    const MachineInstr * get_phi_def(unsigned v_reg);


    /// This method returns true if v_reg is defined in a phi-function.
    bool is_defined_in_phi(unsigned v_reg) const;


    /// The number of virtual registers found in this analysis:
    unsigned get_num_virtuals() const {
        return this->max_virtual;
    }


    /// This function merges the phi-group of v1 with the phi-group of v2, so
    /// that each virtual in any of these groups will be phi-related.
    void merge(unsigned v1, unsigned v2);


    // This method prints the assignment of virtuals to stack slots.
    void print_v2s_mapping() const;


    // This method prints the groups of phi-related virtuals. If "sub_base" is
    // set to true, the method prints virtuals as in t0, t1, etc. Otherwise, it
    // prints the virtuals as in 1024, 1025, etc.
    void print_phi_groups(bool sub_base) const;


    // Prints all phi-functions that have been found during this analysis.
    void print_def_list(MachineFunction & mf) const;


    // This method returns the number of phi-related variables in the source
    // program. Any variable that is part of a phi-function is phi-related.
    unsigned get_number_phi_related_variables() const;


    // This method returns the number of phi-groups in the source program.
    // Phi-groups constitute equivalence classes. This method only returns the
    // number of non-singleton groups.
    unsigned get_number_phi_groups() const;

    private:

    /// This integer holds the index of the virtual with the biggest id. It
    /// is useful when scanning the vector rltd_list.
    unsigned max_virtual;


    /// Number of machine registers in the target architecture. It will never
    /// change during program execution.
    unsigned num_p_regs;


    /// This vector contains a collection of sets. Each set groups together all
    /// the virtuals that are related by some phi-function. In theory, two
    /// virtuals are phi-related if they were part of the same variable
    /// before the SSA-deconstruction phase. This information is lost in LLVM,
    /// and we do an approaximation here.
    std::vector< std::set<unsigned> * > phi_related_virtuals;


    /// This vector contains the Preference objects related to each phi-class.
    std::vector< Preference > pref_list;


    /// Each cell of this vector contains an index of the phi_related_virtuals
    /// vector, or the NO_REG value. If a cell c contains a valid index i, then
    /// it means that the collection of virtuals phi-related to virtual c is
    /// stored in phi_related_virtuals[i].
    std::vector<unsigned> v2phi;


    /// This vector associates virtuals with the phi functions in which they
    /// are defined. Notice that, although the vector will have an entry to all
    /// the virtuals, only those defined in phi-functions will be valid.
    std::vector< const MachineInstr * > def_list;


    /// This vector associates virtuals with the phi-functions in which they
    /// are used.
    std::vector< std::vector<MachineInstr *> > use_list;


    /// This vector associates virtuals to instructions. Each virtual is
    /// defined by a phi-function. def_sites[v] = {i0, i1,..., i2} is the list
    /// of instructions where the parameters of the phi-function that defines
    /// v are defined.
    std::vector< std::set<unsigned> > def_sites;


    /// This vector maps virtuals to stack slots. All the phi-related virtuals
    /// are mapped to the same stack slot.
    std::vector<int> v2s;


    /// The reference to the machine function is useful for printing and
    /// debugging:
    const MachineFunction * machine_function;


    /// runOnMachineFunction - Groups together the phi related virtuals.
    bool runOnMachineFunction(MachineFunction & mf);


    /// This method allocates the data structures.
    void initialize_data_structures(MachineFunction & mf);


    /// This method reads an instruction, and marks it as related to all the
    /// virtuals defined or used in the instruction.
    void visit_phi_function(MachineInstr * phi);


    /// This method marks the virtuals v_def and v_reg as being related. This
    /// method must be called in such a way that v_def is the register defined
    /// in the phi-function, and v_reg is the register used in the same
    /// phi-function.
    void add_to_related_set(unsigned v_def, unsigned v_reg);


    /// Free the memory, so the operating system can use it in other
    /// applications.
    void releaseMemory() {
        this->phi_related_virtuals.clear();
        this->v2phi.clear();
        this->def_list.clear();
        this->use_list.clear();
        this->def_sites.clear();
        this->v2s.clear();
        this->pref_list.clear();
    }

  };

}

#endif
