//===------------- PhiMemCoalescer_Fer.cpp - Printing prototype -----------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This pass attempts to group together virtuals whose live range does not
// overlap, so that they can be inserted in the same memory spot.
//
// Date: March 3rd, 2007
//
//===----------------------------------------------------------------------===//
//
// The combinantion of PhiSplit and PhiAnalysis passes group together all the
// virtuals that are part of the same phi-function. This is fine for
// implementing the phi-deconstruction algorithm, but it is to conservative for
// assigning the same memory address to phi-related virtuals. This pass
// attempts to increase these groups by checking if the live ranges of virtuals
// do not overlap, and then assigning the virtuals to the same group.
//
//===----------------------------------------------------------------------===//

#ifndef PHI_MEM_COALESCER_H
#define PHI_MEM_COALESCER_H

namespace llvm {

  class LiveIntervals;
  class PhiAnalysis_Fer;
  class EqvPhiClass_Fer;
  class SplitPhiGroups_Fer;
  class RegisterCoalescer_Fer;

  class PhiMemCoalescer_Fer : public MachineFunctionPass {

    public:

    static char ID; // Pass identification, replacement for typeid
    PhiMemCoalescer_Fer() : MachineFunctionPass((intptr_t)&ID) {}

    virtual void getAnalysisUsage(AnalysisUsage & AU) const;

    /// Returns a pointer to the vector of virtuals:
    std::vector<unsigned> * get_virtuals() {
        return & this->v2pec;
    }

    /// Return the set of phi-related virtuals for a given virtual
    std::set<unsigned> * get_virtual_set(unsigned v_reg);

    /// This method prints the internal state of this object.
    void dump() const;

    /// Return the number of phi-equivalence classes in this object
    inline unsigned get_num_eqv_classes() const {
        return this->phi_related_virtuals.size();
    }

    private:

    /// This object contains methods to join the live ranges of variables.
    RegisterCoalescer_Fer * rc_;

    /// This vector contains a collection of Phi-equivalence classes. In theory,
    /// two virtuals are phi-related if they were part of the same variable
    /// before the SSA-deconstruction phase. This information is lost in LLVM,
    /// and we do an approaximation here.
    std::vector< EqvPhiClass_Fer * > phi_related_virtuals;

    /// Each cell of this vector contains an index of the phi_related_virtuals
    /// vector, or the NO_REG value. If a cell c contains a valid index i, then
    /// it means that the collection of virtuals phi-related to virtual c is
    /// stored in phi_related_virtuals[i].
    std::vector<unsigned> v2pec;

    /// This class collects information about which nodes are related by phi
    /// functions, in order to allow coalescing during register allocation;
    PhiAnalysis_Fer * phi_analysis;

    /// This pass splits the live range of phi-related virtuals in such a way
    /// that they the transformed program does not contain interferences
    /// between phi-related virtuals. We need a pointer to this class because
    /// it contains the move instructions inserted to split live ranges.
    SplitPhiGroups_Fer * split_phi;

    /// The results of the interval analysis:
    LiveIntervals * interval_analysis;

    /// runOnMachineFunction - Groups together the phi related virtuals.
    bool runOnMachineFunction(MachineFunction & mf);

    /// Free the memory, so the operating system can use it in other
    /// applications.
    void releaseMemory();

    /// This function joins the equivalence classes of v1 and v2.
    void merge(unsigned v1, unsigned v2);

    /// This method initializes the data structures used during register
    /// allocation. Among the variables initilized, are ig, split_phi and
    /// def_use_sites.
    void initialize_data_structures(MachineFunction & mf);

    /// This method returns the phi-class of the virtual v_reg.
    EqvPhiClass_Fer * get_phi_class(unsigned v_reg);

    /// This method return true if any of the virtuals in these sets are
    /// adjacent in the interference graph.
    bool overlap(unsigned use_reg, unsigned def_reg);


    /// This method removes the instruction mi from the basic block, and updates
    /// the liveness analysis information and the def/use sites. It must be used
    /// when removing a move instruction created to break the live range of a
    /// use in a phi-function.
    void remove_instruction_prm(
        MachineInstr * mi,
        unsigned def_reg,
        unsigned use_reg
    );

    /// This method removes the instruction mi from the basic block, and updates
    /// the liveness analysis information and the def/use sites. It must be used
    /// when removing a move instruction created to break the live range of a
    /// use in a phi-function.
    void remove_instruction_rst(
        MachineInstr * mi,
        unsigned def_reg,
        unsigned use_reg
    );

  };

}

#endif
