//===-------------- PhiP3_Fer.h - Clique substitution of P_3 --------------===//
//
//            Register Allocation via Coloring of Chordal Graphs 
//
// This class represents a clique substitution of P_3
//
// Date: Oct 8th, 2007
//
//===----------------------------------------------------------------------===//
//
// This class represents the clique substitution of P3 that forms the head of
// a basic block that contains phi-functions. It is interestly to treat this
// graph differently than the other clique substitutions because it allows to
// design better spilling heuristics.
//
//===----------------------------------------------------------------------===//


#ifndef PHI_FER_H
#define PHI_FER_H

#include <iostream>
#include "FacadeP3_Fer.h"

namespace llvm {

  class vector;
  class MachineInstr;
  class Partition_Fer;
  class RA_Result_Fer;
  class RegMapping_Fer;
  class DefUseSites_Fer;
  class PhiAnalysis_Fer;
  class MachineFunction;

  class VirtualCost_Fer {
    public:
      // Constructor
      VirtualCost_Fer(
        unsigned v_reg
      ) : v_reg_(v_reg), cost_(0) {}
      // Adds the given value to the cost associtated to this virtual
      inline void add_cost(int cost) {
          this->cost_ += cost;
      }
      // Sets the cost associtated to this virtual
      inline void set_cost(int cost) {
          this->cost_ = cost;
      }
      // True if this object has same cost:
      bool operator==(const VirtualCost_Fer &vc) const {
          return cost_ == vc.cost_;
      }
      // True if this object has lower cost:
      bool operator<(const VirtualCost_Fer &vc) const {
          return cost_ < vc.cost_;
      }
      // True if this object has higher cost:
      bool operator>(const VirtualCost_Fer &vc) const {
          return cost_ > vc.cost_;
      }
      // Returns the cost associated to this object:
      inline int get_cost() const {
          return cost_;
      }
      // Returns the virtual represented by this object:
      inline unsigned get_virtual() const {
          return v_reg_;
      }
    private:
      unsigned v_reg_;
      int cost_;
  };

  inline std::ostream & operator<<(
    std::ostream & os,
    const VirtualCost_Fer & vc
  ) {
    os << vc.get_virtual() << ":" << vc.get_cost();
    return os;
  }

  class PhiP3_Fer : public FacadeP3_Fer {

    public:

    /// Constructor method.
    PhiP3_Fer(
        MachineInstr * phi,
        MachineFunction & mf,
        Partition_Fer * partition,
        RegMapping_Fer *rm,
        RA_Result_Fer *rs,
        PhiAnalysis_Fer * pa,
        DefUseSites_Fer * du
    );

    /// Destructor method.
    ~PhiP3_Fer() {
        this->Z_.clear();

        this->Y_.clear();

        this->p_regs_X_.clear();
        this->p_regs_Z_.clear();

        this->bb_.clear();

        this->phi_virtuals_.clear();

        this->v2phi_.clear();
    }

    /// Add 'v_reg' as a node of component X
    void addX(unsigned v_reg) {}

    /// Add 'reg' as a node of component Z
    void addZ(unsigned v_reg);

    /// Add 'reg' as a node of component Y
    void addY(unsigned v_reg);

    /// The number of virtuals in this partition:
    unsigned size() const {
        return this->Z_.size() + this->Y_.size();
    }

    /// This method colors this clique substitution of P_3 with the free
    /// registers of this object. The array preferences associates virtual
    /// variables to colors. This vector determines a color assignment that
    /// would reduce the number of copies in the final assembly code. Given
    /// two successive KP3 (Clique substitution of P_3), q1 and q2, the
    /// color assignment in q1 is used to build the preference vector that will
    /// be passed to q2.
    MachineInstr * color();

    // This method return true if the graph has nodes of type short in any of
    // its maximal cliques.
    bool contains_short() const {
        // TODO: fix this
        return false;
    }
   
    // This method return true if the graph has nodes of type long in any of its
    // maximal cliques.
    bool contains_long() const {
        // TODO: fix this
        return false;
    }

    unsigned get_num_iterations() const {
        return numIter_;
    }

    // This method prints the location of each variable that is alive at the
    // beginning of the program point that this graph represents.
    void print_in_set() const;

    /// Prints the assignment of physical registers to virtuals that exist
    /// in the X component.
    void dump_X_colors() const {}

    /// Prints the assignment of physical registers to virtuals in the Z
    /// component, that is, after the instruction that this object surrounds.
    void dump_Z_colors() const {}

    /// Add this phi-function to the graph. The register defined in the phi
    /// function will contribute to the Z group.
    void add_phi(MachineInstr * phi);

    private:

    /// The reference to the target architecture. This class has only been
    /// implemented for the X86 architecture.
    MachineFunction * mf_;

    /// This object is effectively the register allocator. It is called to
    /// choose a register to spill.
    Partition_Fer *partition_;

    // This class represents the mappings between virtual variables and
    // physical registers.
    RegMapping_Fer * rm_;

    // The result of the register allocation.
    RA_Result_Fer * rs_;

    // The phi analysis is necessary for implementing global coalescing.
    PhiAnalysis_Fer *pa_;

    // The sites of definition and use of each virtual.
    DefUseSites_Fer *du_;

    /// Number of times the coloring algorithm had to spill variables.
    unsigned numIter_;

    /// Shorts in the Y component.
    std::vector<unsigned> Y_;

    /// This vector will hold virtuals that are defined in a phi-function.
    std::vector<unsigned> Z_;

    /// Basic blocks that feed the phi-function:
    std::vector<unsigned> bb_;

    /// This set holds the virtual registers defined in phi-functions.
    std::set<unsigned> phi_virtuals_;
    std::map<unsigned, MachineInstr *> v2phi_;

    /// This vector associates physical registers with the virtual registers
    /// that have been allocated into each physical. For instance, if
    /// p_regs_[1] == 1024, then the virtual %reg1024 has been stored into
    /// physical register 'AH', which correspond to index 1.
    std::vector<unsigned> p_regs_X_;
    std::vector<unsigned> p_regs_Z_;

    // This method returns an index of a basic block that has already been
    // visited by the register allocator. This basic block contains a non-empty
    // vector of locations. The number of the basic block that is feeding the
    // map will be placed into the 'mbb_num' reference.
    const std::map<unsigned, unsigned> * get_incoming_map
                                                    (unsigned & mbb_num) const;

  };

}

#endif
