//===------------- RA_Result_Fer.h - Bank of Machine Registers ------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This class holds the results of register allocation.
//
// Date: June 27th, 2007
//
//===----------------------------------------------------------------------===//
//
// This class holds the results of register allocation. These results consist
// in:
// Instructions -> {Virtual X Register}
// Instructions -> {Event}, where an event is a copy, load, store or swap
// Basic block -> {Event}
//
//===----------------------------------------------------------------------===//

#ifndef RA_RESULT_FER_H
#define RA_RESULT_FER_H

namespace llvm {

  class MachineFunction;
  class RegMapping_Fer;
  class LiveIntervals;
  class MachineInstr;
  class TargetInstrInfo;
  class Partition_Fer;
  class TargetRegisterClass;
  class DefUseSites_Fer;

  typedef enum Events {CP, ST, LD, SW} Event;

  // CP, LD, ST: (l_dst, v_dst) = (l_src, v_src)
  // SW: (l_dst, v_src) (l_src, v_dst) = (l_dst, v_dst) (l_src, v_src)
  class Event_Fer {
    public:

    Event_Fer(
        Event type,
        unsigned l_dst,
        unsigned v_dst,
        unsigned l_src,
        unsigned v_src
    ) : type(type),
        l_dst(l_dst),
        v_dst(v_dst),
        l_src(l_src),
        v_src(v_src)
    {
    }

    Event type;

    unsigned l_dst, v_dst, l_src, v_src;
  };

  // Variable v_reg is bound to register p_reg
  class Binding_Fer {
    public:

    Binding_Fer(unsigned v_reg, unsigned p_reg) : v_reg(v_reg), p_reg(p_reg) {
    }

    unsigned v_reg, p_reg;
  };

  class RA_Result_Fer {

    public:

    /// Constructor method.
    RA_Result_Fer(
        MachineFunction & mf,
        const LiveIntervals * li,
        const RegMapping_Fer * rm,
        Partition_Fer * pf,
        DefUseSites_Fer * du
    );

    /// Destructor method.
    ~RA_Result_Fer() {
        num_loads_.clear();
        binding_list_.clear();
        mbb_event_list_.clear();
        mock_event_list_.clear();
        instr_event_list_.clear();
        mbb_beginning_list_.clear();
    }

    /// Prints a representation of this object.
    void dump(MachineFunction & mf) const;

    // This method converts the mapping from virtual register to physical
    // register by the register allocator to a real assignment of physical
    // registers in each operand of each instruction of the given machine
    // function.
    void hard_code_register_mapping();

    /// Adds the binding (loc, v_reg) at instructin i.
    void add_binding(unsigned loc, unsigned v_reg, MachineInstr * mi);

    /// Add the copy (l_dst, v_dst) = (l_src, v_src) before i
    void add_cp(
        unsigned l_dst,
        unsigned v_dst,
        unsigned l_src,
        unsigned v_src,
        MachineInstr * mi
    );

    /// Add the load (l_dst, v_dst) = (l_src, v_src) before i
    void add_ld(
        unsigned l_dst,
        unsigned v_dst,
        unsigned l_src,
        unsigned v_src,
        MachineInstr * mi
    );

    /// Add the store (l_dst, v_dst) = (l_src, v_src) after i
    void add_st(
        unsigned l_dst,
        unsigned v_dst,
        unsigned l_src,
        unsigned v_src,
        MachineInstr * mi
    );

    /// Add the copy (l_dst, v_dst) = (l_src, v_src) at the end of machine
    /// basic block b if add_at_end is true, and at the beginning otherwise.
    void add_cp2mbb(
        unsigned l_dst,
        unsigned v_dst,
        unsigned l_src,
        unsigned v_src,
        unsigned b,
        bool add_at_end
    );

    /// Add the load (l_dst, v_dst) = (l_src, v_src) at the end of machine
    /// basic block b if add_at_end is true, and at the beginning otherwise.
    void add_ld2mbb(
        unsigned l_dst,
        unsigned v_dst,
        unsigned l_src,
        unsigned v_src,
        unsigned b,
        bool add_at_end
    );

    /// This method adds a 'mock' load at the end of basic block b. This load
    /// will be implemented only if v_src is marked as a 'local load' by the
    /// load reduction heuristics.
    /// Add the load (l_dst, v_dst) = (l_src, v_src) at the end of machine basic
    /// block b.
    void add_mock_load(
        unsigned l_dst,
        unsigned v_dst,
        unsigned l_src,
        unsigned v_src,
        unsigned b
    );

    /// Add the store (l_dst, v_dst) = (l_src, v_src) at the end of machine
    /// basic block b if add_at_end is true, and at the beginning otherwise.
    void add_st2mbb(
        unsigned l_dst,
        unsigned v_dst,
        unsigned l_src,
        unsigned v_src,
        unsigned b,
        bool add_at_end
    );

    /// SW: (l_dst, v_src) (l_src, v_dst) = (l_dst, v_dst) (l_src, v_src) at
    /// the end of basic block mbb
    void add_sw2mbb(
        unsigned l_dst,
        unsigned v_dst,
        unsigned l_src,
        unsigned v_src,
        unsigned b
    );

    // This method returns true if mi has been already visited by the register
    // allocator and physical registers have been assigned to its variables.
    bool has_binding(MachineInstr * mi) const;

    // This method determines where we should insert loads for spilled
    // variables. The puzzle solver tends to split the live range of variables
    // into many smaller pieces, and inserts loads to keep the splitting
    // consistent. One typical example happens at the beginning of basic blocks
    // where loads must be inserted to join the live ranges of variables that
    // reach the basic block at both memory and registers. However, if the
    // variable is used very few times, say once, it is not worth to insert all
    // these loads. In this case we could replace all these loads with a single
    // load. This method determines when we should do this.
    bool load_local(unsigned v_reg) const;

    // This method increments by one the number of loads inserted to split the
    // live range of v_reg. 
    void inc_load(unsigned v_reg);

    private:

    /// This variable is used to define which statistics to collect at each
    /// time. These statistics are the number of moves, loads and stores used
    /// to remove phi/pi-functions are parallel copies.
    struct StatsType {
        enum {
            COPY = 0,
            LOAD = 1,
            STORE = 2,
            SWAP = 3
        };
    };
    struct StatsState {
        enum {
            PHI = 0,
            PI = 1,
            PC = 2
        };
    };
    unsigned stat_st;

    /// The list that maps instructions to the bindings found by the register
    /// allocator for that instruction:
    std::vector< std::vector<Binding_Fer> > binding_list_;

    /// The list that binds instructions to the events that must be inserted
    /// before or after that instruction during code generation. Copies and
    /// loads  are always inserted before the instruction. Stores are inserted
    /// after.
    std::vector< std::vector<Event_Fer> > instr_event_list_;

    /// The list that maps basic blocks to the instructions that must be
    /// inserted after that basic block because of SSA-elimination.
    std::vector< std::vector<Event_Fer> > mbb_event_list_;

    /// This vector holds the mock loads. These loads are implemented only if
    /// the source register is marked as 'load local' by the load reduction
    /// heuristics.
    std::vector< std::vector<Event_Fer> > mock_event_list_;

    /// The list with the events that must be inserted at the beginning of the
    /// basic blocks during the SSA elimination phase.
    std::vector< std::vector<Event_Fer> > mbb_beginning_list_;

    // This vector holds the number of load instructios that were inserted
    // per each spilled variable:
    std::vector<unsigned> num_loads_;

    // The interval representation of the target program.
    const LiveIntervals * li_;

    // This class represents the mappings between virtual variables and
    // physical registers.
    const RegMapping_Fer * rm_;

    // The machine function that is the target of register allocation.
    MachineFunction * mf_;

    /// This class contains the result of the liveness analysis produced
    /// after register allocation.
    Partition_Fer * pf_;

    /// This variable maps virtual registers to the instructions where they are
    /// defined or used.
    DefUseSites_Fer * du_;

    /// This method computes statistics about the number of moves, stores,
    /// loads and swaps inserted due to phi-functions are parallel copies.
    void compute_statistics(unsigned stats_tp);

    // Print the instructions that preecede the instruction, such as loads and
    // copies.
    void print_event_list(const std::vector<Event_Fer> & vec) const;

    // Prints the binddings on this instruction. 'i' is the index of the
    // instruction.
    void print_bindings(
        const MachineInstr * mi,
        unsigned i,
        const TargetInstrInfo & tii
    ) const;

    // This method prints the register binding for a phi-function. Only the
    // definition is printed.
    void print_phi_function(const MachineInstr * phi, unsigned i) const;

    // This method prints all the instructions necessary at the end of the
    // basic block. It includes jumps, and also the instructions used to
    // destroy phi-functions.
    void print_tail_code(
        MachineFunction & mf,
        const MachineBasicBlock * mbb
    ) const;

    // This method removes all the occurrences of phi-functions from the given
    // basic block. Before this method is called, instructions must be inserted
    // in the control flow graph to reproduce the behavior of phi functions.
    MachineBasicBlock::iterator destroy_phi_functions(
        MachineBasicBlock & mbb,
        const MRegisterInfo & reg_info
    );

    // This method go through all the operands of mbbi that hold virtual
    // variables, and changed them into physical registers. If mbbi define
    // variables that have been spilled, then the method inserts a store.
    void hard_code_bindings(
        MachineBasicBlock & mbb,
        const MRegisterInfo & reg_info
    );

    // This method gets a list of events and an iterator and inserts code before
    // the iterator in order to implement these events.
    void hard_code_event_list(
        const std::vector<Event_Fer> & vec,
        MachineBasicBlock & mbb,
        MachineBasicBlock::iterator * mbbi,
        const MRegisterInfo & reg_info
    );

    // This method implements the mock loads. Only loads of variables marked as
    // 'load local' will be inserted at the end of the basic block mbb.
    void hard_code_mock_loads(
        const std::vector<Event_Fer> & vec,
        MachineBasicBlock & mbb,
        const MRegisterInfo & reg_info
    );

  };

} // end of llvm namespace

#endif
