//===------------ RegMapping_Fer.h - Bank of Machine Registers ------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This class represents the bank of machine registers for register allocation
// purposes.
//
// Date: May 27th, 2006
//
//===----------------------------------------------------------------------===//
//
// This class represents the bank of machine registers for register allocation
// purposes. Essentially, it marks which physical registers are being used,
// and which ones are free to use. For each physical register in use, it keeps
// track of which virtual registers they use.
//
//===----------------------------------------------------------------------===//

#ifndef REG_MAPPING_FER_H
#define REG_MAPPING_FER_H

namespace llvm {

  class Line_Fer;
  class LineCompareEnd;
  class PhiAnalysis_Fer;
  class DefUseSites_Fer;

  class RegMapping_Fer {

    public:

    /// Constructor method. Initializes the data structures. The machine
    /// function parameter is only used to provide initializing information,
    /// such as the number of virtual and physical registers.
    RegMapping_Fer(MachineFunction & mf, PhiAnalysis_Fer * pa);

    /// Destructor method. Gives back to the operating system the memory that
    /// was in use during the object's execution.
    ~RegMapping_Fer();

    /// This method sets p_reg as the physical location of reg, which can be
    /// either a virtual variable or a physical register itself, due to
    /// precoloring. If reg is a physical register, we require that reg==p_reg
    /// or an exception would be thrown.
    void allocate_color(unsigned reg, unsigned p_reg);

    /// This method makes the physical register p_reg available for allocating
    /// virtuals. If p_reg is not a physical register, an error is issued.
    void free_color(unsigned p_reg);

    /// This method returns a string in the format "t1(r1), t2(r2), ... tm(rm),
    /// where virtual registers ti are allocated in machine registers ri.
    std::string virtual_to_string() const;

    /// This method returns a string in the format "r1(t1), r2(t2), ... rm(tm),
    /// where physical registers ri is holding the value of virtual register
    /// ti.
    std::string physical_to_string() const;

    // This method returns the virtual register that was assigned to the given
    // physical register. If the physical register does not contain any virtual
    // register, the method will return RF_UNDEF.
    unsigned get_assigned_data(unsigned p_reg) const;

    // This method returns the physical register that was assigned to the
    // virtual register. If the virtual register has not been allocated to any
    // register, it will return null. It is also possible that the given
    // parameter describes a physical register. In this case, it is a
    // pre-colored register, and its location must be the physical register
    // itself.
    unsigned get_physical_location(unsigned reg) const;

    // This method stores the values in the 'preference' vector as the last
    // registers in use at the end of the basic block 'b'.
    void store_v2p(
        std::map<unsigned, unsigned> & map,
        std::set<Line_Fer * , LineCompareEnd > & live_set
    ) const;

    // This method cleans the current register mapping in this object, and
    // fills it with the mapping given in the "map" parameter.
    void reset_mapping(
        const std::map<unsigned, unsigned> & map,
        std::set<Line_Fer * , LineCompareEnd > & live_set
    );

    /// Clean the arrays used in this object.
    void releaseMemory();

    // This method scans the vector of available registers and returns a
    // register that can be assigned to 'v'. It will return RF_UNDEF if there
    // is no register available.
    unsigned get_free_register(unsigned v_reg) const;

    // This method determines if p_reg is not currently allocated to any virtual
    // register. In order to verify this, it does not suffice to check if the
    // cell of p_reg in the physical register table contains the RF_UNDEF value,
    // because physical registers contain alias. For instance, in the X86
    // architecture, the physical registers AL, AX and EAX are all alias, and if
    // one of them is in use, the others are not free. If neither p_reg nor any
    // of its aliases is been used to allocate virtuals, the method returns
    // true. It returns false otherwise.
    bool is_register_available(unsigned p_reg) const;

    // This method marks v_reg as a spilled variable. A stack slot will be
    // assigned to v_reg.
    void set_spill(unsigned v_reg);

    /// Returns true if the given virtual has been spilled.
    bool is_spill(unsigned v_reg) const;

    // This method marks v_reg as defined in a two address instruction.
    // Therefore, v_reg is defined twice: v_reg = x; v_reg = v_reg + y;
    void set_defined_in_2addr_instr(unsigned v_reg);

    // This method returns true if v_reg is defined in a two address
    // instruction, and false otherwise.
    bool is_defined_in_2addr_instr(unsigned v_reg) const;

    // This method assigns RF_UNDEF to the virtual v_reg.
    void reset_color(unsigned v_reg);

    /// This method prints the internal state of this register mapping, which
    /// includes the bank of physical registers plus the memory.
    void dump() const;

    /// This method prints in the standard output all the virtuals that are
    /// spilled, plus their stack position.
    void print_stack_allocation() const;

    /// Return the number of physical registers in this object.
    unsigned num_p_regs() {
        return this->p2v_.size();
    }

    /// Return the number of virtual registers in the target function:
    unsigned num_v_regs() {
        return this->v2p_.size();
    }

    /// Return the stack slot that has been assigned to the virtual
    unsigned get_stack_slot(unsigned v_reg) const;

    // This method returns the last physical register that has been allocated to
    // v_reg.
    unsigned get_last_location(unsigned v_reg) const;

    // This method prints a list with all the variables that have been spilled.
    void print_store_list() const;

    // This method spills the physical register associated to v_reg. If there
    // are different virtuals sharing aliases of the register that must be
    // freed, then spill all those virtuals.
    void free_register(
        unsigned p_reg,
        std::vector<unsigned> & sp_regs
    );

    // This method returns the number of aliases of p_reg that are being in use.
    unsigned number_of_occupied_aliases(unsigned p_reg) const;

    // This method modifies the vector aliases. For each physical register
    // p_reg in regs, it will insert the value 'p_reg' in every cell of aliases
    // that represents a physical register that is an alias of p_reg and is
    // being used.
    void fill_aliases(
        std::vector<unsigned> & regs,
        std::vector<unsigned> & aliases
    ) const;

    // This method clears the mapping between virtuals and physical registers.
    // It does not clear the last physical assigned to each virtual, neither the
    // memory slot of each virtual.
    void clear_mapping();

    // This method returns the virtual register that is the next to be used
    // among the virtual registers stored in aliases of p_reg.
    unsigned next_nearest_alias_used(unsigned p_reg, DefUseSites_Fer *du) const;

    private:

    /// This vector holds the last physical register that has been assigned to
    /// each virtual.
    std::vector<unsigned> last_location_;

    /// This is the table that maps physical registers to virtual registers.
    /// If p2v[p_reg] = NO_REG, than p_reg is not assigned to any
    /// virtual register.
    std::vector<unsigned> p2v_;

    /// This vector maps virtual registers to physical registers. A
    /// BankOfRegisters_Fer::NO_REG value indicates that the virtual register
    /// hasn't been assigned a value yet.
    std::vector<unsigned> v2p_;

    /// This vector marks which registers are spilled. For each virtual 'v',
    /// is_spl_[v - FirstVirtualRegister] will contain true if this virtual
    /// is currently in memory, or false otherwise. Notice that a virtual can
    /// be in memory at some intevals of its live range, and not in others.
    /// Thus, this variable is always reseted by the clear_mapping method.
    std::vector<bool> is_spill_;

    /// This vector maps virtuals to the stack slots where they are stored.
    /// Testing this v2s_[...] == RF_UNDEF is a good way to know if a virtual
    /// has a copy in memory or not.
    std::vector<unsigned> v2s_;

    /// This vector marks which virtuals are defined in two address
    /// instructions.
    std::vector<bool> is_defined_in_2addr_;

    /// The bank of register keeps a reference to the machine function in
    /// order to get the architecture specific information.
    MachineFunction * machine_function_;

    /// This class collects information about which nodes are related by phi
    /// functions, in order to allow coalescing during register allocation;
    PhiAnalysis_Fer * phi_analysis_;

  };

} // end of llvm namespace

#endif
