//===----------------------------------------------------------------------===//
//
// Copyright (c) 2007 The Regents of the University of California, Los
// Angeles Campus. All Rights Reserved.  
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted by nonprofit, research institutions for
// research use only, provided that the following conditions are met:
// 
// *         Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer. 
// 
// *         Redistributions in binary form must reproduce the above
// copyright notice, this list of conditions and the following disclaimer
// in the documentation and/or other materials provided with the
// distribution. 
// 
// *         Neither the name of The Regents nor the names of its
// contributors may be used to endorse or promote products derived from
// this software without specific prior written permission. 
// 
// The end-user understands that the program was developed for research
// purposes and is advised not to rely exclusively on the program for any
// reason.
// 
// THE SOFTWARE PROVIDED IS ON AN "AS IS" BASIS, AND THE REGENTS AND
// CONTRIBUTORS HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
// UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE REGENTS AND CONTRIBUTORS
// SPECIFICALLY DISCLAIM ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING BUT
// NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE REGENTS OR
// CONTRIBUTORS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
// INCIDENTAL, EXEMPLARY OR CONSEQUENTIAL DAMAGES, INCLUDING BUT NOT
// LIMITED TO  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES, LOSE OF USE,
// DATA OR PROFITS, OR BUSINESS INTERRUPTION, HOWEVER CAUSED AND UNDER ANY
// THEORY OF LIABILITY WHETHER IN CONTRACT, STRICT LIABILITY OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF ADVISED OF THE POSSIBILITY
// OF SUCH DAMAGE. 
// 
// If you do not agree to these terms, do not download or use the software.
// This license may be modified only in a writing signed by authorized
// signatory of both parties.
//
//===-------------------- Regs_Fer.cpp - Layout output --------------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This class contains some handy functions to manipulate LLVM registers.
//
// Date: June 2nd, 2007
//
//===----------------------------------------------------------------------===//
//
// This file contains functions that verify if a register is free.
//
//===----------------------------------------------------------------------===//

#ifndef REGS_FER_H
#define REGS_FER_H
 
#include <iostream>                              // cerr
#include <sstream>                               // ostringstream (RegClique)

#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/CodeGen/MachineFunction.h"
#include "llvm/CodeGen/MachineFunction.h"

#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetMachine.h"

#include "X86PuzzleBoard_Fer.h"

namespace llvm {

#define RF_UNDEF ~0U

#define RF_SPILL ~1U

class Regs_Fer {

public:

//===--------------------------------------------------------------------------
// This method scans the vector of available registers and returns a register
// that can be assigned to 'v'. It will return RF_UNDEF if there is no register
// available.
//===--------------------------------------------------------------------------
static unsigned get_free_register(
    unsigned v_reg,
    std::vector<unsigned> & rb1,
    std::vector<unsigned> & rb2,
    MachineFunction & mf
) {
    const TargetRegisterClass *trc = mf.getSSARegMap()->getRegClass(v_reg);
    TargetRegisterClass::iterator trci = trc->allocation_order_begin(mf);
    TargetRegisterClass::iterator trce = trc->allocation_order_end(mf);

    for (; trci != trce; ++trci) {
        if (is_register_available(*trci, rb1, rb2, mf)) {
            return *trci;
        }
    }

    return RF_UNDEF;
}


//===--------------------------------------------------------------------------
// This method determines if p_reg is not currently allocated to any virtual
// register. In order to verify this, it does not suffice to check if the
// cell of p_reg in the physical register table contains the RF_UNDEF value,
// because physical registers contain alias. For instance, in the X86
// architecture, the physical registers AL, AX and EAX are all alias, and if
// one of them is in use, the others are not free. If p_reg is not been
// used to allocate any virtual, the method returns true. It returns false
// otherwise.
//===--------------------------------------------------------------------------
static bool is_register_available(
    unsigned p_reg,
    std::vector<unsigned> & rb1,
    std::vector<unsigned> & rb2,
    MachineFunction & mf
) {
    assert(MRegisterInfo::isPhysicalRegister(p_reg) &&
                                           "Target register must be physical");

    if (rb1[p_reg] != RF_UNDEF || rb2[p_reg] != RF_UNDEF) {
        return false;
    }

    const TargetMachine & target_machine = mf.getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();

    for (const unsigned *a = reg_info->getAliasSet(p_reg); *a; ++a) {
        if (rb1[*a] != RF_UNDEF || rb2[*a] != RF_UNDEF) {
            return false;
        }
    }

    return true;
}


//===--------------------------------------------------------------------------
// This method returns the alias of p_reg that is in the class of v_reg.
// Parameters:
// p_reg: a physical register.
// v_reg: a virtual register.
// mf: the machine function in which register allocation is being performed.
//===--------------------------------------------------------------------------
static unsigned get_reg_class(
    unsigned p_reg,
    unsigned v_reg,
    MachineFunction & mf
) {
    // See if p_reg can hold v_reg:
    if(Regs_Fer::compatible_class(mf, v_reg, p_reg)) {
        return p_reg;
    } else {
        // Go through the aliases of p_reg, and try to find one that can hold
        // v_reg.
        const TargetMachine & target_machine = mf.getTarget();
        const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
        for (const unsigned *a = reg_info->getAliasSet(p_reg); *a; ++a) {
            if (Regs_Fer::compatible_class(mf, v_reg, *a)) {
                return *a;
            }   
        }
    }

    return RF_UNDEF;
}


//===--------------------------------------------------------------------------
// This method verifies if p_reg can be used to allocated v_reg. This will be
// the case if both registers belong to the same class.
//===--------------------------------------------------------------------------
static bool compatible_class(
    MachineFunction & mf,
    unsigned v_reg,
    unsigned p_reg
) {
    assert(MRegisterInfo::isPhysicalRegister(p_reg) &&
                                           "Target register must be physical");
    const TargetRegisterClass *trc = mf.getSSARegMap()->getRegClass(v_reg);

    TargetRegisterClass::iterator trci = trc->allocation_order_begin(mf);
    TargetRegisterClass::iterator trce = trc->allocation_order_end(mf);

    for (; trci != trce; ++trci) {
        if(*trci == p_reg) {
             return true;
        }
    }

    return false;
}


//===--------------------------------------------------------------------------
// This method returns true if none of the registers aliased to p_reg is
// occupied, or if any of them is, them it is occupied by v_reg. This means
// that v_reg can be assigned to p_reg if they have compatible classes.
// The method returns false otherwise.
//===--------------------------------------------------------------------------
static bool are_aliases_free(
    unsigned v_reg,
    unsigned p_reg,
    std::vector<unsigned> & rb,
    MachineFunction & mf
) {
    const TargetMachine & target_machine = mf.getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();

    for (const unsigned *a = reg_info->getAliasSet(p_reg); *a; ++a) {
        if (rb[*a] != RF_UNDEF && rb[*a] != v_reg) {
            return false;
        }
    }

    return true;
}


//===--------------------------------------------------------------------------
// This method prints the definition of a virtual register. In addition of
// printing the pair (p_reg, v_reg), the method prints a binding pair for each
// alias of p_reg.
//===--------------------------------------------------------------------------
static void print_def(MachineFunction & mf, unsigned p_reg, unsigned v_reg) {
    print_reg(p_reg, v_reg);
    if(MRegisterInfo::isVirtualRegister(v_reg)) {
        const TargetMachine & target_machine = mf.getTarget();
        const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
        for (const unsigned *a = reg_info->getAliasSet(p_reg); *a; ++a) {
            std::cout << " ";
            Regs_Fer::print_reg(*a, v_reg);
        }
    } else {
        const TargetMachine & target_machine = mf.getTarget();
        const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
        for (const unsigned *a = reg_info->getAliasSet(p_reg); *a; ++a) {
            std::cout << " ";
            Regs_Fer::print_reg(*a, *a);
        }
    }
}


//===--------------------------------------------------------------------------
// This method prints a register and its location in a notation like
// (loc, reg).
//===--------------------------------------------------------------------------
static void print_reg(unsigned loc, unsigned reg) {
    //unsigned p_reg = rm->get_physical_location(reg);
    std::cout << "(R" << loc << ", " << "P" << reg << ") ";
}


//===--------------------------------------------------------------------------
// This method returns the operand number that corresponds to the given
// virtual register in the instruction.
//===--------------------------------------------------------------------------
static unsigned get_operand_num(
    MachineInstr * mi,
    unsigned v_reg,
    bool is_def
) {
    for(unsigned u = 0; u < mi->getNumOperands(); u++) {
        const MachineOperand & mo = mi->getOperand(u);
        if(mo.isRegister() && mo.getReg() == v_reg && mo.isDef() == is_def) {
            return u;
        }
    }
    assert(false && "Operand not present in instruction.");
    return RF_UNDEF;
}

private:

};

}

#endif
