//===-------------------- RevVec_Fer.cpp - PHI Matrix ---------------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This class implements the reverse vector data structure.
//
// Date: Jun 22, 2007
//
//===----------------------------------------------------------------------===//
//
// A reverse vector provides a "contains" operation that is O(1). The only
// requirement is that the maximum element stored in the vector must be known.
// This class only stores unsigned ints.
//
//===----------------------------------------------------------------------===//

#ifndef REVERSE_VECTOR_H
#define REVERSE_VECTOR_H

namespace llvm {

#define RV_UNDEF ~0U

  class MRegisterInfo;
  class RegMapping_Fer;

  class RevVec_Fer {

  public:

    /// Constructor method.
    RevVec_Fer(
        unsigned size_phi_row,
        unsigned max_elem,
        const RegMapping_Fer * rm
    );

    /// Inserts loc at the "index"  position of the phi-vector. "loc" is the
    /// location of v_reg.
    void push(unsigned v_reg, unsigned loc, unsigned index);

    /// This method sets the location in index to be p_reg.
    void update_phys(unsigned p_reg, unsigned index) {
        assert(p_reg < this->phy_regs_.size() && "Invalid physical register");
        assert(index < this->phi_rows_.size() && "Invalid index");
        this->phi_rows_[index] = p_reg;
        this->phy_regs_[p_reg] = index;
    }

    /// This method prints the constents of this vector in the standard ouput.
    void dump(const MRegisterInfo * reg_info) const;

    /// Destructor method.
    ~RevVec_Fer() {
        this->phi_rows_.clear();
        this->phy_regs_.clear();
        this->vrt_regs_.clear();
    }

    /// True if this vector contains u;
    bool contains_phys(unsigned loc) const {
        if(loc < this->phy_regs_.size()) {
            if(this->phy_regs_[loc] != RV_UNDEF) {
                return true;
            }
        }
        return false;
    }

    /// This method returns true if this vector contains, either the physical
    /// register loc, or one of its alias. If loc is not a physical register,
    /// or if neither loc nor any of its alias is in the reverse vector,
    /// then the method returns false;
    bool contains_reg_or_alias
                          (unsigned loc, const MRegisterInfo & reg_info) const;

    /// Return the index of the element in the vector.
    unsigned get_index_phy(unsigned loc) const {
        if(loc < phy_regs_.size()) {
            return phy_regs_[loc];
        } else {
            return RV_UNDEF;
        }
    }

    /// Return the index of the element in the vector.
    unsigned get_virtual(unsigned index) const {
        assert(index < this->vrt_regs_.size() && "Invalid index");
        return vrt_regs_[index];
    }

    /// Returns the virtual that is allocated to p_reg.
    unsigned get_virtual_assign_to_phys(unsigned p_reg) const {
        assert(p_reg < phy_regs_.size());
        unsigned index = phy_regs_[p_reg];
        assert(index < this->vrt_regs_.size() && "Invalid index");
        return vrt_regs_[index];
    }

    /// This method replaces the old physical register with the new one.
    void replace_phys(unsigned old_reg, unsigned new_reg);

    /// Return the location stored at index.
    unsigned get_loc(unsigned index) const {
        assert(index < this->phi_rows_.size() && "Invalid index");
        return phi_rows_[index];
    }

    /// Returns the size of the row in the phi-matrix represented by this
    /// object.
    unsigned size() const {
        return phi_rows_.size();
    }

    /// Removes the location at the given index from this object, and returns
    /// the location stored at the removed index.
    unsigned remove(unsigned index);

    /// Swaps the contents of registers r1 and r2.
    void swap(unsigned r1, unsigned r2);

  private:

    // This vector represents the row of a phi-function. For instance, for
    // (R1, R2) phi:= [(R3, R4),(R5,R6)] we would have phi_rows = [R1,R2]
    // It is this vector that really implements the reverse index.
    std::vector<unsigned> phi_rows_;

    // This vector represents the physical registers. For instance,
    // phy_regs_[2] = 6 means that the physical register 2 is used to store the
    // data in the 6th row of the phi-function.
    std::vector<unsigned> phy_regs_;

    // This vector represents the virtual registers stored in this object.
    std::vector<unsigned> vrt_regs_;

    /// This class contains the assignment of spilled temporaries to memory
    /// addresses.
    const RegMapping_Fer *rm_;

  };

}

#endif
