//===------------- PhiDeconstruction.h - Phi-elimination Pass -------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it eliminate phi functions after register allocation has been
// performed.
//
// Date: June 23th, 2007
//
//===----------------------------------------------------------------------===//
//
// Traditional algorithms used to destroy phi functions normally replace them
// by copy instructions. Although the classical algorithms preserve the
// semantics of the SSA-form program, they may increase the number of registers
// found during the register assignment phase. To avoid this problem, we
// replace phi functions by parallel copies. Parallel copies are normally
// implemented with xor instructions (Remember the trick of swaping two
// variables without an auxiliary using xor?). The register swap is a costly
// operation, because it demands three instructions to be implemented;
// therefore, it should be avoided whenever possible. One way of avoiding
// swaps is by means of move instructions. Basically, swap operations are not
// necessary in every time that phi instructions are destroyed. This happens
// when the definition vector contains machine registers not used in the
// parameter vector. This pass tries to use such optimization  before inserting
//  swap operations. It performs phi elimination in time linear in the size of
// the definition vector.
//
//===----------------------------------------------------------------------===//

#ifndef SSA_ELIM_FER_H
#define SSA_ELIM_FER_H

#include "llvm/PassSupport.h"

namespace llvm {

class RevVec_Fer;
class Partition_Fer;
class RegMapping_Fer;
class RA_Result_Fer;
class PhiAnalysis_Fer;
class MachineFunction;
class MachineBasicBlock;

class SSAElim_Fer : public MachineFunctionPass {

  public:

    static char ID; // Pass identification, replacement for typeid
    SSAElim_Fer() : MachineFunctionPass((intptr_t)&ID) {}

    // This method specifies which other passes are required, or are modified
    // by this pass.
    virtual void getAnalysisUsage(AnalysisUsage & AU) const;

    /// Returns a reference to the internal RA_Result_Fer object
    RA_Result_Fer * get_RA_result() {
        return rs_;
    }

  private:

    /// This class contains the result of the liveness analysis produced
    /// after register allocation.
    Partition_Fer * pf_;

    /// This class collects information about which nodes are related by phi
    /// functions, in order to allow coalescing during register allocation;
    PhiAnalysis_Fer * pa_;

    /// This class contains the assignment of spilled temporaries to memory
    /// addresses.
    RegMapping_Fer *rm_;

    /// The bank of register keeps a reference to the machine function in
    /// order to get the architecture specific information.
    MachineFunction * mf_;

    // The result of the register allocation.
    RA_Result_Fer * rs_;

    // This vector will hold, for each spilled variable, how many loads were
    // inserted to maintain the live range of this variable.
    // std::vector<unsigned> loadsPerSpills;

    // Number of machine registers in the target architecture.
    unsigned num_p_regs_;

    /// Visits the basic blocks and lists of intervals assigning registers to
    /// virtual variables.
    virtual bool runOnMachineFunction(MachineFunction & mf);

    /// Free the memory, so the operating system can use it in other
    /// applications.
    void releaseMemory() {
    }

    // This method inserts instructions at all the predecessors of mbb in order
    // to transfer colors.
    void visit(MachineBasicBlock * mbb);

    // This method inserts code at the end of src_mbb to transfer values from
    // src_mbb to dst_mbb.
    void insert_transfer_at_end(
        MachineBasicBlock * dst_mbb,
        MachineBasicBlock * src_mbb,
        unsigned phi_column
    );

    // This method inserts code at the beginning of dst_mbb to transfer values
    // from src_mbb to dst_mbb. This happens because src_mbb has multiple
    // successors, and if code was added at its end, it could compromise the
    // semantics of the target program.
    void insert_transfer_at_beginning(
        MachineBasicBlock * dst_mbb,
        MachineBasicBlock * src_mbb
    );

    // This method inserts copies/loads/stores at the program point that this
    // graph covers in order to preserve the semantics of the program.
    void transfer_colors(
        RevVec_Fer & src,
        RevVec_Fer & dst,
        MachineBasicBlock * mbb
    );

    // This method fills the vector diff with indexes of dst. Each of these
    // indexes correspond to a safe operation that can be performed to transfer
    // values from the source vector to the destination vector.
    void find_safe_operations(
        RevVec_Fer & src,
        RevVec_Fer & dst,
        std::vector<unsigned> & diff
    ) const;

    // This method replaces each alias of p_dst in vec with the corresponding
    // alias of p_src.
    void replace_regs(
        RevVec_Fer & vec,
        unsigned p_src,
        unsigned p_dst,
        const MRegisterInfo & reg_info
    ) const;

    // This method inserts copy instructions at the end of the target basic
    // block in order to implement safe operations.
    void implement_safe_operations(
        RevVec_Fer & src,
        RevVec_Fer & dst,
        std::vector<unsigned> & diff,
        MachineBasicBlock * mbb
    );

    // When performing the tranfer of values, it is always safe to perform a
    // store, if the destination location is a memory address.
    void implement_stores(
        RevVec_Fer & src,
        RevVec_Fer & dst,
        MachineBasicBlock * mbb
    );

    // This function uses register swaps to implement the transfers that remain
    // between src and dst, once safe operations no longer apply.
    void implement_swaps(
        RevVec_Fer & src,
        RevVec_Fer & dst,
        MachineBasicBlock * mbb
    );

    // This function swaps registers of 32 and 16 bits, but do nothing about the
    // little registers of 8 bits. 
    bool swap_big_registers(
        RevVec_Fer & src,
        RevVec_Fer & dst,
        MachineBasicBlock * mbb
    );

    // This function swaps the eight bit registers.
    bool swap_small_registers(
        RevVec_Fer & src,
        RevVec_Fer & dst,
        MachineBasicBlock * mbb
    );

    // This pass inserts loads for the variables that are marked as load local
    // by RA_Results. This is a heuristics that aims at reducing the number of
    // loads in the target program.
    void reload_locals(
        RevVec_Fer & src,
        RevVec_Fer & dst,
        MachineBasicBlock * mbb
    );

    // This pass removes instances from the parallel copy where the source and
    // the destination have the same location.
    void implement_coalescing(
        RevVec_Fer & src,
        RevVec_Fer & dst,
        MachineBasicBlock * mbb
    );

    // Returns true if the element in the c-th column of phi is alive at the
    // end of basic block pred.
    bool is_column_alive(
        const MachineInstr * phi,
        MachineBasicBlock & src_mbb,
        unsigned c
    ) const;

    // This method fills the data structures with the data that they must have
    // in order to start the processing of intervals.
    void initialize_data_structures(MachineFunction & mf);

};

}

#endif
