//===-------------------- Regs_Fer.cpp - Layout output --------------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This class constains functios to manipulate the register classes of the x86
// computer architecture
//
// Date: June 26th, 2007
//
//===----------------------------------------------------------------------===//
//
// This file contains functions that distinguish the size of registers, the
// classes of registers, etc.
//
//===----------------------------------------------------------------------===//

#ifndef CLREGS_FER_H
#define CLREGS_FER_H

#include "NeatPrint_Fer.h"

namespace llvm {

#define RG_UNDEF ~0U

class X86PuzzleBoard_Fer {

public:

struct PzConst {
    enum {
        UNDEF = ~0U
    };
};

struct RegSz {
    enum {
        E = 8,
        X = 4,
        H = 2,
        L = 1
    };
};

// The families of registers in the x86 architecture.
struct RegFm {
    enum {
        A  =   1,
        B  =   2,
        C  =   4,
        D  =   8,
        SP =  16,
        BP =  32,
        SI =  64,
        DI = 128
    };
};

struct PuzzlePiece {
    enum {
        SIZE1 = 1,
        SIZE2 = 2,
        UNDEFINED = 4
    };
};


//===--------------------------------------------------------------------------
// This method returns true if v_reg is defined by mi, and mi is a two address
// instruction. In this case, v_reg will be both defined and used in mi.
//===--------------------------------------------------------------------------
static bool is_def_two_addr(const MachineInstr * mi, unsigned v_reg) {
    for (unsigned i = 0, e = mi->getNumOperands(); i != e; ++i) {
        const MachineOperand & mo1 = mi->getOperand(i);
        if (mo1.isRegister() && mo1.isDef() && mo1.getReg() == v_reg) {
            return true;
        }
    }
    return false;
/*
    for (unsigned i = 0, e = mi->getNumOperands(); i != e; ++i) {
        const MachineOperand & mo1 = mi->getOperand(i);
        if (mo1.isRegister() && mo1.isDef() && mo1.getReg() == v_reg) {
            for (unsigned j = i+1; j < e; ++j) {
                const MachineOperand & mo2 = mi->getOperand(j);
                if (
                     mo2.isRegister() &&
                     mo2.isUse() &&
                     mo2.getReg() == v_reg &&
                     mi->getInstrDescriptor()->
                                getOperandConstraint(j, TOI::TIED_TO) == (int)i
                ) {
                    return true;
                }
            }
        }
    }
    return false;
*/
}


//===--------------------------------------------------------------------------
// This method returns true if the given instruction is two address code.
//===--------------------------------------------------------------------------
static bool is_two_addr(const MachineInstr * mi) {
    const TargetInstrDescriptor * tid = mi->getInstrDescriptor();
    for (unsigned j = 1; j < tid->numOperands; ++j) {
        int ti = tid->getOperandConstraint(j, TOI::TIED_TO);
        if(ti != -1) {
            return true;
        }
    }
    return false;
}


//===--------------------------------------------------------------------------
// This method returns the piece that corresponds to the register class trc.
//===--------------------------------------------------------------------------
static unsigned cl2piece(const TargetRegisterClass *trc){
    if(trc->getID() == 9) {
        return PuzzlePiece::SIZE1;
    } else if(trc->getID() >= 4 && trc->getID() <= 8) {
        return PuzzlePiece::SIZE2;
    } else {
        return PuzzlePiece::UNDEFINED;
    }
}


//===--------------------------------------------------------------------------
// This method finds a free short register, e.g 8-bit register, that can be
// packed with another 8-bit register already in use. This method implements
// the puzzle rule that attempts to pack two short registers together as much
// as possible.
//===--------------------------------------------------------------------------
static unsigned find_available_short(
    std::vector<unsigned> & rb1,
    std::vector<unsigned> & rb2
) {
    unsigned first_candidate = RG_UNDEF;
    if(rb1[1] == RG_UNDEF && rb2[1] == RG_UNDEF) {
        first_candidate = 1;
        if(rb1[2] != RG_UNDEF || rb2[2] == RG_UNDEF) {
            return 2;
        }
    }
    if(rb1[2] == RG_UNDEF && rb2[2] == RG_UNDEF) {
        first_candidate = 2;
        if(rb1[1] != RG_UNDEF || rb2[1] == RG_UNDEF) {
            return 1;
        }
    }
    if(rb1[4] == RG_UNDEF && rb2[4] == RG_UNDEF) {
        first_candidate = 4;
        if(rb1[5] != RG_UNDEF || rb2[5] == RG_UNDEF) {
            return 5;
        }
    }
    if(rb1[5] == RG_UNDEF && rb2[5] == RG_UNDEF) {
        first_candidate = 5;
        if(rb1[4] != RG_UNDEF || rb2[4] == RG_UNDEF) {
            return 4;
        }
    }
    if(rb1[9] == RG_UNDEF && rb2[9] == RG_UNDEF) {
        first_candidate = 9;
        if(rb1[10] != RG_UNDEF || rb2[10] == RG_UNDEF) {
            return 10;
        }
    }
    if(rb1[10] == RG_UNDEF && rb2[10] == RG_UNDEF) {
        first_candidate = 10;
        if(rb1[9] != RG_UNDEF || rb2[9] == RG_UNDEF) {
            return 9;
        }
    }
    if(rb1[12] == RG_UNDEF && rb2[12] == RG_UNDEF) {
        first_candidate = 12;
        if(rb1[15] != RG_UNDEF || rb2[15] == RG_UNDEF) {
            return 15;
        }
    }
    if(rb1[15] == RG_UNDEF && rb2[15] == RG_UNDEF) {
        first_candidate = 15;
        if(rb1[12] != RG_UNDEF || rb2[12] == RG_UNDEF) {
            return 12;
        }
    }
    return first_candidate;
}


//===--------------------------------------------------------------------------
// Returns true if the parameter register is 8-bit.
//===--------------------------------------------------------------------------
static bool is_short(unsigned pr) {
    if(
        pr == 1 || pr == 4 || pr == 9 || pr == 12 ||
        pr == 2 || pr == 5 || pr == 10 || pr == 15
    ) {
        return true;
    } else {
        return false;
    }
}


//===--------------------------------------------------------------------------
// Returns a number associated to the class of the register passed as
// parameter:
// E: EAX, EBX, ECX, EDX, EDI
// X: AX, BX, CX, DX, DI
// H: AH, BH, CH, DH
// L: AL, BL, CL, DL
//===--------------------------------------------------------------------------
static unsigned reg2size(unsigned pr) {
    assert(MRegisterInfo::isPhysicalRegister(pr) &&
                           "Attempt to obtain class of non-physical register");
    if(pr == 1 || pr == 4 || pr == 9 || pr == 12) {
        return RegSz::H;
    } else if(pr == 2 || pr == 5 || pr == 10 || pr == 15) {
        return RegSz::L;
    } else if(pr == 3 || pr == 6 || pr == 8 || pr == 11 || pr == 13 ||
                                            pr == 16 || pr == 84 || pr == 86) {
        return RegSz::X;
    } else if( pr >= 17 && pr <= 22 || pr == 25 || pr == 26 ) {
        return RegSz::E;
    } else {
        return RG_UNDEF;
    }
}


//===--------------------------------------------------------------------------
// This method returns the 32-bit extended register that corresponds to the
// 16-bit register x_reg.
//===--------------------------------------------------------------------------
static unsigned get_ex_reg(unsigned x_reg) {
    switch(x_reg) {
        case  3: return 17; // AX -> EAX
        case  6: return 18; // BP -> EBP
        case  8: return 19; // BX -> EBX
        case 11: return 20; // CX -> ECX
        case 13: return 21; // DI -> EDI
        case 16: return 22; // DX -> EDX
        case 84: return 25; // SI -> ESI
        case 86: return 26; // SP -> ESP
        default: return RG_UNDEF;
    }
}


//===--------------------------------------------------------------------------
// This method returns the size of a x86 register, in bits.
//===--------------------------------------------------------------------------
static unsigned size(unsigned p_reg) {
    switch(p_reg) {
        case  1: return  8; // AH
        case  2: return  8; // AL
        case  3: return 16; // AX
        case  4: return  8; // BH
        case  5: return  8; // BL
        case  6: return 16; // BP
        case  8: return 16; // BX
        case  9: return  8; // CH
        case 10: return  8; // CL
        case 11: return 16; // CX
        case 12: return  8; // DH
        case 13: return 16; // DI
        case 15: return  8; // DL
        case 16: return 16; // DX
        case 17: return 32; // EAX
        case 18: return 32; // EBP
        case 19: return 32; // EBX
        case 20: return 32; // ECX
        case 21: return 32; // EDI
        case 22: return 32; // EDX
        case 25: return 32; // ESI
        case 26: return 32; // ESP
        case 84: return 16; // SI
        case 86: return 16; // SP
        default: return RG_UNDEF;
    }
}


//===--------------------------------------------------------------------------
// Given a register family, its minimum size is the size of the smallest alias.
// For instance, the minimum size of {AH, AL, AX, EAX} is 8, the size of either
// AH or AL.
//===--------------------------------------------------------------------------
static unsigned minimum_size(unsigned pr) {
    assert(MRegisterInfo::isPhysicalRegister(pr) &&
                    "Attempt to obtain minimum size of non-physical register");
    if(pr == 1 || pr == 2 || pr == 3 || pr == 17) {
        // AH, AL, AX, EAX:
        return 8;
    } else if(pr == 4  || pr == 5  || pr == 8  || pr == 19) {
        // BH, BL, BX, EBX:
        return 8;
    } else if(pr == 9  || pr == 10  || pr == 11 || pr == 20) {
        // CH, CL, CX, ECX:
        return 8;
    } else if(pr == 12 || pr == 15 || pr == 16 || pr == 22) {
        // DH, DL, DX, EDX:
        return 8;
    } else if(pr == 6  || pr == 18) {
        // BP, EBP:
        return 16;
    } else if(pr == 13 || pr == 21) {
        // DI, EDI:
        return 16;
    } else if(pr == 25 || pr == 84) {
        // ESI, SI:
        return 16;
    } else if(pr == 26 || pr == 86) {
        // ESP, SP:
        return 16;
    } else {
        return RG_UNDEF;
    }
}


//===--------------------------------------------------------------------------
// Returns the family associated to a particular register:
// {AH, AL, AX, EAX} => A
// {BH, BL, BX, EBX} => B
// {CH, CL, CX, ECX} => C
// {DH, DL, DX, EDX} => D
// {BP, EBP} => P
// {DI, EDI} => I
//===--------------------------------------------------------------------------
static unsigned reg2family(const unsigned pr) {
    assert(MRegisterInfo::isPhysicalRegister(pr) &&
                           "Attempt to obtain class of non-physical register");
    if(pr == 1 || pr == 2 || pr == 3 || pr == 17) {
        // AH, AL, AX, EAX:
        return RegFm::A;
    } else if(pr == 4  || pr == 5  || pr == 8  || pr == 19) {
        // BH, BL, BX, EBX:
        return RegFm::B;
    } else if(pr == 9  || pr == 10  || pr == 11 || pr == 20) {
        // CH, CL, CX, ECX:
        return RegFm::C;
    } else if(pr == 12 || pr == 15 || pr == 16 || pr == 22) {
        // DH, DL, DX, EDX:
        return RegFm::D;
    } else if(pr == 6  || pr == 18) {
        // BP, EBP:
        return RegFm::BP;
    } else if(pr == 13 || pr == 21) {
        // DI, EDI:
        return RegFm::DI;
    } else if(pr == 25 || pr == 84) {
        // ESI, SI:
        return RegFm::SI;
    } else if(pr == 26 || pr == 86) {
        // ESP, SP:
        return RegFm::SP;
    } else {
        return RG_UNDEF;
    }
}


//===--------------------------------------------------------------------------
// This method returns a machine register given a family and a size. A family
// is a collection of all the registers that alias.
//===--------------------------------------------------------------------------
static unsigned find_replacement(unsigned sz, unsigned fm) {
    switch(fm) {
        case RegFm::A:
            if(sz == RegSz::H) {
                return 1;
            } else if(sz == RegSz::L) {
                return 2;
            } else if(sz == RegSz::X) {
                return 3;
            } else if(sz == RegSz::E) {
                return 17;
            }
        case RegFm::B:
            if(sz == RegSz::H) {
                return 4;
            } else if(sz == RegSz::L) {
                return 5;
            } else if(sz == RegSz::X) {
                return 8;
            } else if(sz == RegSz::E) {
                return 19;
            }
        case RegFm::C:
            if(sz == RegSz::H) {
                return 9;
            } else if(sz == RegSz::L) {
                return 10;
            } else if(sz == RegSz::X) {
                return 11;
            } else if(sz == RegSz::E) {
                return 20;
            }
        case RegFm::D:
            if(sz == RegSz::H) {
                return 12;
            } else if(sz == RegSz::L) {
                return 15;
            } else if(sz == RegSz::X) {
                return 16;
            } else if(sz == RegSz::E) {
                return 22;
            }
        case RegFm::BP:
            if(sz == RegSz::X) {
                return 6;
            } else if(sz == RegSz::E) {
                return 18;
            }
        case RegFm::DI:
            if(sz == RegSz::X) {
                return 13;
            } else if(sz == RegSz::E) {
                return 21;
            }
        case RegFm::SI:
            if(sz == RegSz::E) {
                return 25;
            } else if(sz == RegSz::X) {
                return 84;
            }
        case RegFm::SP:
            if(sz == RegSz::E) {
                return 26;
            } else if(sz == RegSz::X) {
                return 86;
            }
    }
    return RG_UNDEF;
}

private :

};

}

#endif
