//===----------- CriticalEdgeRemoval_Fer - Break critical edges ----------===//
//
//                     The LLVM Compiler Infrastructure
//
// This files was originally developed by the LLVM research group. Fernando
// at cs dot ucla dot edu has changed the file a little, so it could be used as
// a pre-register allocation pass.
//
//===---------------------------------------------------------------------===//
//
// CriticalEdgeRemoval pass - Break all of the critical edges in the CFG by
// inserting a dummy basic block.  This pass may be "required" by passes that
// cannot deal with critical edges. Notice that this pass invalidates the
// CFG, because the same BasicBlock is used as parameter for the src
// MachineBasicBlock and the new dummy MachineBasicBlock.
//
//===---------------------------------------------------------------------===//

#include <iostream>

#include "llvm/Transforms/Scalar.h"
#include "llvm/Transforms/Utils/BasicBlockUtils.h"
#include "llvm/Target/TargetData.h"
#include "llvm/CodeGen/MachineJumpTableInfo.h"
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/Function.h"
#include "llvm/Instructions.h"
#include "llvm/Type.h"
#include "llvm/Support/CFG.h"
#include "llvm/CodeGen/MachineFunctionPass.h"            // MachineBasicBlock
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"

#include "llvm/fernando/CriticalEdgeRemoval_Fer.h"       // class definition

// Only to read the "IGNORE.txt" file
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/fernando/NeatPrint_Fer.h"
#include <fstream>
#include <sstream>
#include <string>
//---

using namespace llvm;

char CriticalEdgeRemoval_Fer::ID = 0;

namespace {
    RegisterPass<CriticalEdgeRemoval_Fer>
    X("rm_crit", "Critical Edge Removal - Fernando", true);
}


//===---------------------------------------------------------------------===//
// I am not sure that this pass conserves all these properties, but I am
// assuming that critical edges between BasicBlocks have already been broken
// before. So, I only have to break critical edges between basic blocks
// split due to some optimizations. If the underline BasicBlock is the same
// for both source and target of the critical edge, this pass will not
// invalidate the other passes.
//===---------------------------------------------------------------------===//
void CriticalEdgeRemoval_Fer::getAnalysisUsage(AnalysisUsage & AU) const {
    AU.setPreservesAll();
}


//===---------------------------------------------------------------------===//
// This method opens a file called "IGNORE.txt" and, if it finds the name of
// the given function there, the method returns true.
//===---------------------------------------------------------------------===//
bool found_function_name(MachineFunction & mf) {
    std::ifstream file("IGNORE.txt");
    if(file.is_open()) {
        const Function * f = mf.getFunction();
        std::ostringstream f_str;
        f_str << f->getName();
        std::string f_name = f_str.str();
        while(!file.eof()) {
            std::string line;
            std::getline(file, line);
            if(line == f_name) {
                file.close();
                return true;                   
            }
        }
        file.close();
    }
    return false;
}


//===---------------------------------------------------------------------===//
// This method split all the critical edges in the machine function control
// flow graph.
//===---------------------------------------------------------------------===//
bool CriticalEdgeRemoval_Fer::runOnMachineFunction(MachineFunction & mf) {
/*
    if(found_function_name(mf)) {
        return true;
    }
*/

    std::vector<MachineBasicBlock *> src_blocks;
    std::vector<MachineBasicBlock *> dst_blocks;

    for(unsigned u = 0; u < mf.size(); u++) {
        MachineBasicBlock * mbb = mf.getBlockNumbered(u);
        for(MachineBasicBlock::succ_iterator succ = mbb->succ_begin();
                                             succ != mbb->succ_end(); succ++) {
            MachineBasicBlock * mbb_succ = *succ;
            // necessary and sufficient condition that characterizes a critical
            // edge: predecessor with many successors, successor with many
            // predecessors.
            if( (mbb->succ_size() > 1) && (mbb_succ->pred_size() > 1) ) {
                src_blocks.push_back(mbb);
                dst_blocks.push_back(mbb_succ);
            }
        }
    }

    for(unsigned u = 0; u < src_blocks.size() > 0; u++) {
        SplitCriticalMachineEdge(src_blocks[u], dst_blocks[u]);
    }

    return false;
}


//===---------------------------------------------------------------------===//
// This method scans the basic block and returns the index of the jump table
// that is referenced in that basic block.
//===---------------------------------------------------------------------===//
unsigned getTableIndex(MachineBasicBlock & mbb) {
    unsigned index = ~0U;
    MachineBasicBlock::iterator I, E;
    for (I = mbb.begin(), E = mbb.end(); I != E; ++I) {
        for (unsigned u = 0, e = I->getNumOperands(); u != e; ++u) {
            MachineOperand & op = I->getOperand(u);
            if (!op.isJumpTableIndex()) {
                continue;
            } else {
                unsigned aux_index = op.getJumpTableIndex();
                assert( (index == ~0U || aux_index == index) &&
                          "Basic block with two different jump table indices");
                index = aux_index;
            }
        }
    }
    return index;
}


//===---------------------------------------------------------------------===//
// This method breaks the critical edge that links src to dst. It will insert
// a dummy block in between these two machine blocks. Originally, I though
// about reporting an error if the source and destiny basic block do not share
// the same underling BasicBlock. Later on, it turned out that this restriction
// was too much for the client methods, and I decided to remove it. The source
// code still has the old assertion comented. One way that the client could
// use to ensure the absence of critical edges was to call
// call PM.add(createBreakCriticalEdgesPass()); in the target machine. But
// this seems not to be enough. I've discovered this error when compiling the
// Ackermann function.
//===---------------------------------------------------------------------===//
MachineBasicBlock* CriticalEdgeRemoval_Fer::SplitCriticalMachineEdge(
  MachineBasicBlock* src,
  MachineBasicBlock* dst
) {
  const BasicBlock* srcBB = src->getBasicBlock();

  MachineBasicBlock* crit_mbb = new MachineBasicBlock(srcBB);

  // modify the llvm control flow graph
  src->removeSuccessor(dst);
  src->addSuccessor(crit_mbb);
  crit_mbb->addSuccessor(dst);

  // insert the new block into the machine function.
  src->getParent()->getBasicBlockList().insert(src->getParent()->end(),
                                               crit_mbb);

  // insert a unconditional branch linking the new block to dst
  const TargetMachine& TM = src->getParent()->getTarget();
  const TargetInstrInfo* TII = TM.getInstrInfo();
  std::vector<MachineOperand> emptyConditions;
  TII->InsertBranch(*crit_mbb, dst, (MachineBasicBlock*)0, emptyConditions);

  // modify every branch in src that points to dst to point to the new
  // machine basic block instead:
  MachineBasicBlock::iterator mii = src->end();
  bool found_branch = false;
  while (mii != src->begin()) {
    mii--;
    // if there are no more branches, finish the loop
    if (!TII->isTerminatorInstr(mii->getOpcode())) {
      break;
    }
    
    // Scan the operands of this branch, replacing any uses of dst with
    // crit_mbb.
    for (unsigned i = 0, e = mii->getNumOperands(); i != e; ++i) {
      MachineOperand & mo = mii->getOperand(i);
      if (mo.isMachineBasicBlock() &&
          mo.getMachineBasicBlock() == dst) {
        found_branch = true;
        mo.setMachineBasicBlock(crit_mbb);
      }
    }
  }

  // TODO: This is tentative. It may be necessary to fix this code. Maybe
  // I am inserting too many gotos, but I am trusting that the asm printer
  // will optimize the unnecessary gotos.
  if(!found_branch) {
    TII->InsertBranch(*src, crit_mbb, (MachineBasicBlock*)0, emptyConditions);
  }

  /// Change all the phi functions in dst, so that the incoming block be
  /// crit_mbb, instead of src
  for(mii = dst->begin(); mii != dst->end(); mii++) {
    /// the first instructions are always phi functions.
    if(mii->getOpcode() != TargetInstrInfo::PHI)
      break;
    
    for (unsigned u = 0; u != mii->getNumOperands(); ++u)
      if (mii->getOperand(u).isMachineBasicBlock() &&
          mii->getOperand(u).getMachineBasicBlock() == src)
        mii->getOperand(u).setMachineBasicBlock(crit_mbb);
  }
  
  return crit_mbb;
}
