//===------------- DefUseSites_Fer.cpp - Printing prototype ---------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it stores the definition and use sites of virtual registers.
//
// Date: August 22th, 2007
//
//===----------------------------------------------------------------------===//
//
// This class implements a mapping of virtual registers to instructions.
// For each virtual, this class gives its definition site, which is unique,
// due to the SSA property, and all the use sites. A definition site is the
// index of the instruction where the register is defined, and a use site is
// the index of the instruction where the register is used.
// These indices are defined by the live interval analysis.
//
//===----------------------------------------------------------------------===//

#define DEBUG_TYPE "defUse"

#include <set>
#include <vector>
#include <sstream>                                // ostringstream
#include <iostream>                               // cerr
#include <algorithm>                              // std::sort(...)

#include "llvm/Instructions.h"                    // MachineOperand

#include "llvm/CodeGen/Passes.h"                  // EdgeLivenessFernandoPassID
#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/CodeGen/MachineInstr.h"
#include "llvm/CodeGen/MachineFunctionPass.h"     // MachineBasicBlock
#include "llvm/CodeGen/LiveIntervalAnalysis.h"

#include "llvm/ADT/Statistic.h"
#include "llvm/ADT/DepthFirstIterator.h"         // df_ext_iterator

#include "llvm/Support/Debug.h"

#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"

#include "llvm/fernando/RA_Result_Fer.h"
#include "llvm/fernando/NeatPrint_Fer.h"
#include "llvm/fernando/SplitPhiGroups_Fer.h"
#include "llvm/fernando/PhiMemCoalescer_Fer.h"

#include "llvm/fernando/DefUseSites_Fer.h"        // class definition


using namespace llvm;

STATISTIC(num0, "Variables used zero times.");
STATISTIC(num1, "Variables used one time.");
STATISTIC(num2, "Variables used two times.");
STATISTIC(num3, "Variables used three times.");
STATISTIC(num4, "Variables used four times.");
STATISTIC(num5Plus, "Variables used five or more times.");

char DefUseSites_Fer::ID = 0;

namespace {
    RegisterPass<DefUseSites_Fer>
    X("Virtual2DefUseMapping", "Virtual to def/use mapping - Fernando.");
}


//===--------------------------------------------------------------------------
// reporst that this pass does not change anything on the control flow graph.
//===--------------------------------------------------------------------------
void DefUseSites_Fer::getAnalysisUsage(AnalysisUsage &AU) const {
    AU.addRequired<LiveIntervals>();
    AU.addRequired<SplitPhiGroups_Fer>();
    AU.addRequired<PhiMemCoalescer_Fer>();
    AU.setPreservesAll();
}


//===--------------------------------------------------------------------------
// Free the memory, so the operating system can use it in other applications.
// TODO: it is necessary to verify if I am freeing the memory correctly.
//===--------------------------------------------------------------------------
void DefUseSites_Fer::releaseMemory() {
    this->v2d_.clear();
    this->v2u_.clear();
    this->active_.clear();
}


//===--------------------------------------------------------------------------
// This instruction marks the instruction mi as the definition site of the
// virtual register v_reg.
//===--------------------------------------------------------------------------
void DefUseSites_Fer::set_def_site(MachineInstr * mi, unsigned v_reg) {
    assert(MRegisterInfo::isVirtualRegister(v_reg) &&
                          "Assigning definition site to non-virtual register");
    unsigned m_index = this->lis_->getInstructionIndex(mi);
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    this->v2d_[base_reg] = m_index;
    this->v2u_[base_reg].reg_ = v_reg;
}


//===--------------------------------------------------------------------------
// This method adds the instruction mi as one of the use sites of the virtual
// v_reg.
//===--------------------------------------------------------------------------
void DefUseSites_Fer::add_use_site(MachineInstr * mi, unsigned v_reg) {
    assert(MRegisterInfo::isVirtualRegister(v_reg) &&
                                 "Assigning use site to non-virtual register");
    unsigned m_index = this->lis_->getInstructionIndex(mi);
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    UseVec_Fer & uv = this->v2u_[base_reg];
    uv.insert(m_index);
}


//===--------------------------------------------------------------------------
// This method visits a basic block and collects all the uses and defs present
// in this basic block.
//===--------------------------------------------------------------------------
void DefUseSites_Fer::visit_basic_block(MachineBasicBlock & mbb) {

    MachineBasicBlock::iterator mii = mbb.begin();
    while (mii != mbb.end()) {
        MachineInstr *mi = mii++;
        for (unsigned i = 0; i != mi->getNumOperands(); ++i) {
            MachineOperand & mo = mi->getOperand(i);
            if (mo.isRegister() && mo.getReg()) {
                if(MRegisterInfo::isVirtualRegister(mo.getReg())) {
                    unsigned v_reg = mo.getReg();
                    if (mo.isDef()) {
                        this->set_def_site(mi, v_reg);
                    } else {
                        this->add_use_site(mi, v_reg);
                    }
                }
            }
        }
    }
}


//===--------------------------------------------------------------------------
// This method removes all the uses of spill candidates that are in indices
// higher than idx
//===--------------------------------------------------------------------------
void DefUseSites_Fer::advance_index(unsigned idx) {
    std::vector<UseVec_Fer *> aux;
    std::set<UseVec_Fer * , UseVecComp_Fer >::iterator i;
    std::set<UseVec_Fer * , UseVecComp_Fer >::iterator e;
    for(i = this->active_.begin(), e = this->active_.end(); i != e;) {
        UseVec_Fer * ud = *i;
        if(ud->current() > idx) {
            break;
        } else {
            ud->advance(idx);
            this->active_.erase(i++);
            aux.push_back(ud);
        }
    }

    for(unsigned u = 0; u < aux.size(); u++) {
        if(aux[u]->current() >= idx) {
            this->active_.insert(aux[u]);
        }
    }

    // Printing top of list:
/*
    std::cout << "Printing active vector after advancing, idx = " << idx << "\n";
    for(i = this->active_.begin(), e = this->active_.end(); i != e; i++) {
        UseVec_Fer * ud = *i;
        std::cout << " " << ud->reg_ << " @ " << ud->current() << std::endl;
    }
*/
}


//===--------------------------------------------------------------------------
// This method inserts all the uses of register v_reg into the queue of
// possible spill candidates
//===--------------------------------------------------------------------------
void DefUseSites_Fer::set_active(unsigned v_reg, unsigned v_end) {
    assert(MRegisterInfo::isVirtualRegister(v_reg) &&
                                 "Assigning use site to non-virtual register");
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    std::sort(v2u_[base_reg].indices_.begin(), v2u_[base_reg].indices_.end());
    if(v2u_[base_reg].end() < v_end) {
        v2u_[base_reg].indices_.push_back(v_end);
    }
    this->active_.insert(& this->v2u_[base_reg]);
}


//===--------------------------------------------------------------------------
// This method returns the index of the next use of virtual v_reg. Notice
// that the DefUseSites class keeps a state that is the next usage of each
// virtual register.
//===--------------------------------------------------------------------------
unsigned DefUseSites_Fer::next_usage(unsigned v_reg) const {
    assert(MRegisterInfo::isVirtualRegister(v_reg) &&
                                 "Assigning use site to non-virtual register");
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    return this->v2u_[base_reg].current();
}


//===--------------------------------------------------------------------------
// Prints in the standard error device all the information contained in this
// data structure.
//===--------------------------------------------------------------------------
void DefUseSites_Fer::dump() const {
    for(unsigned u = 0; u < this->v2d_.size(); u++) {
        this->dump(u + MRegisterInfo::FirstVirtualRegister);
    }
}


//===--------------------------------------------------------------------------
// Prints in the standard error device all the information about one single
// register.
//===--------------------------------------------------------------------------
void DefUseSites_Fer::dump(unsigned v_reg) const {
    assert(MRegisterInfo::isVirtualRegister(v_reg) &&
                     "Only virtual registers are registered in def use sites");
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    std::cout << "Uses: " << this->v2u_[base_reg];
    std::cout << ", Def : " << this->v2d_[base_reg] << std::endl;
}


//===--------------------------------------------------------------------------
// This method computes the number of times each variable is used
//===--------------------------------------------------------------------------
void DefUseSites_Fer::compute_statistics() {
    for(unsigned u = 0; u < this->v2u_.size(); u++) {
        UseVec_Fer & uv = this->v2u_[u];
        switch(uv.indices_.size()) {
            case 0:
                num0++;
                break;
            case 1:
                num1++;
                break;
            case 2:
                num2++;
                break;
            case 3:
                num3++;
                break;
            case 4:
                num4++;
                break;
            default :
                num5Plus++;
                break;
        }
    }
}


//===--------------------------------------------------------------------------
// This method returns the number of times that the parameter variable is used
// in the source program.
//===--------------------------------------------------------------------------
unsigned DefUseSites_Fer::get_num_uses(unsigned v_reg) const {
    assert(MRegisterInfo::isVirtualRegister(v_reg) &&
                     "Only virtual registers are registered in def use sites");
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    return this->v2u_[base_reg].num_uses_;
}


//===--------------------------------------------------------------------------
// runOnMachineFunction - collects the use/def sites of each register, and
// computes their weights.
//===--------------------------------------------------------------------------
bool DefUseSites_Fer::runOnMachineFunction(MachineFunction & mf) {
//std::cout << "Running DefUseSistes_Fer\n";
    // These next lines only get the number of virtual registers in mf:
    SSARegMap * srm = mf.getSSARegMap();
    unsigned last_v_reg = srm->getLastVirtReg() + 1;
    unsigned num_v_regs = last_v_reg - MRegisterInfo::FirstVirtualRegister;

    this->v2d_.resize(num_v_regs);
    this->v2u_.resize(num_v_regs);

    this->lis_ = & getAnalysis<LiveIntervals>();

    MachineBasicBlock *entry = mf.begin();
    std::set<MachineBasicBlock*> visited;
    for (df_ext_iterator<MachineBasicBlock*> dfi = df_ext_begin(entry,visited),
                       endi = df_ext_end(entry, visited); dfi != endi; ++dfi) {
        MachineBasicBlock *mbb = *dfi;
        visit_basic_block(*mbb);
    }

//this->dump();
    this->compute_statistics();

    return false;
}
