//===------------- EqvPhiClass_Fer.h - Phi Equivalence Class --------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This class represents the equivalence class of phi-related virtuals.
//
// Date: Jun 30th, 2006
//
//===----------------------------------------------------------------------===//
//
// Two virtuals are said to be phi-related according to the following
// properties:
// (i - reflexivity) any virtual v is phi-related to itself;
// (ii - symmetry) if virtual v is phi-related to virtual u, then u is
// phi-related to v;
// (iii - transitivity) if v_1 is phi-related to v_2, and v_2 is phi-related
// to v_3, then v_1 is phi-related to v_3.
// (iv) given a phi-function such as v_i = phi(v_{i1}, v_{i2}, \ldots,
// v_{im}), the virtuals v_i, v_{i1}, v_{i2}, \ldots, v_{im} are
// phi-related;
//
//===----------------------------------------------------------------------===//

#include <iostream>

#include "llvm/CodeGen/LiveInterval.h"
#include "llvm/CodeGen/LiveIntervalAnalysis.h"

#include "llvm/fernando/EqvPhiClass_Fer.h"

using namespace llvm;

//===--------------------------------------------------------------------------
//
//===--------------------------------------------------------------------------
void EqvPhiClass_Fer::merge(EqvPhiClass_Fer & pec) {
    // First, merge the intervals:
    Lines::iterator it1 = pec.begin();
    Lines::iterator it1_end = pec.end();
    Lines::iterator it2 = this->begin();
    Lines::iterator it2_end = this->end();
    while(it1 != it1_end) {
        while(it2 != it2_end && it2->end < it1->start) {
            it2++;
        }
        if(it2 == it2_end) {
            this->lines.push_back(*it1);
        } else if(it2->end == it1->start) {
            Lines::iterator it_aux = it2;
            it_aux++;
            if(it_aux != it2_end) {
                assert(it_aux->start >= it1->end && "Intervals overlap.");
                if(it_aux->start == it1->end) {
                    it_aux->start = it2->start;
                    this->lines.erase(it2);
                    it2 = it_aux;
                } else {
                    it2->end = it1->end;
                }
            } else {
                it2->end = it1->end;
            }
        } else if(it1->end == it2->start) { // correct!
            it2->start = it1->start;
        } else {
            assert(it2->start > it1->end && "Intervals overlap.");
            this->lines.insert(it2, *it1);
        }
        it1++;
    }
    // Now, merge the set of virtuals:
    std::set<unsigned> * set = pec.get_virtuals();
    this->virtuals.insert(set->begin(), set->end());
}


//===--------------------------------------------------------------------------
//
//===--------------------------------------------------------------------------
void EqvPhiClass_Fer::add(LiveInterval & li) {
    LiveInterval::Ranges::const_iterator it1 = li.begin();
    LiveInterval::Ranges::const_iterator it1_end = li.end();
    Lines::iterator it2 = this->begin();
    Lines::iterator it2_end = this->end();
    while(it1 != it1_end) {
        while(it2 != it2_end && it2->end < it1->start) {
            it2++;
        }
        if(it2 == it2_end) {
            this->lines.push_back(Line(it1->start, it1->end));
        } else if(it2->end == it1->start) {
            Lines::iterator it_aux = it2;
            it_aux++;
            if(it_aux != it2_end) {
                assert(it_aux->start >= it1->end && "Intervals overlap.");
                if(it_aux->start == it1->end) {
                    it_aux->start = it2->start;
                    this->lines.erase(it2);
                    it2 = it_aux;
                } else {
                    it2->end = it1->end;
                }
            } else {
                it2->end = it1->end;
            }
        } else if(it1->end == it2->start) { // correct!
            it2->start = it1->start;
        } else {
            if(it2->start < it1->end) {
                std::cerr << "Register: " << li.reg << "\n";
                std::cerr << "it1->start: " << it1->start << ", it1->end: " << it1->end << "\n";
                std::cerr << "it2->start: " << it2->start << ", it2->end: " << it2->end << "\n";
            }
            assert(it2->start >= it1->end && "Intervals overlap.");
            this->lines.insert(it2, Line(it1->start, it1->end));
        }
        it1++;
    }
    this->virtuals.insert(li.reg);
}


//===--------------------------------------------------------------------------
//
//===--------------------------------------------------------------------------
void EqvPhiClass_Fer::add(unsigned start, unsigned end) {
    Lines::iterator it = this->begin();
    Lines::iterator end_it = this->end();
    while(it != end_it && it->end < start) {
        it++;
    }
    if(it == end_it) {
        this->lines.push_back(Line(start, end));
    } else if(it->end == start) {
        Lines::iterator it_aux = it;
        it_aux++;
        assert(it_aux->start >= end && "Intervals overlap.");
        it->end = end;
        if(it->end == it_aux->start) {
            it_aux->start = it->start;
            this->lines.erase(it);
        }
    } else if(it->start == end) {
        it->start = start;
    } else {
        assert((start > it->end) || (end < it->start) && "Intervals overlap.");
        this->lines.insert(it, Line(start, end));
    }
}


//===--------------------------------------------------------------------------
//
//===--------------------------------------------------------------------------
bool EqvPhiClass_Fer::overlap(LiveInterval & li) {
    LiveInterval::Ranges::const_iterator it1 = li.begin();
    LiveInterval::Ranges::const_iterator it1_end = li.end();
    Lines::const_iterator it2_end = end();
    Lines::const_iterator it2 = begin();
    while(true) {
        if(it1 == it1_end || it2 == it2_end) {
            break;
        }
        while( (it1 != it1_end) && (it1->end <= it2->start) ) {
            it1++;
        }
        if(it1 == it1_end) {
            return false;
        }
        while( (it2 != it2_end) && (it2->end <= it1->start) ) {
            it2++;
        }
        if(it2 == it2_end) {
            return false;
        }
        if((it1->start <= it2->start && it2->start < it1->end) ||
                         (it2->start <= it1->start && it1->start < it2->end)) {
            return true;
        }
    }
    return false;
}


//===--------------------------------------------------------------------------
//
//===--------------------------------------------------------------------------
bool EqvPhiClass_Fer::overlap(const EqvPhiClass_Fer & pec) const {
    Lines::const_iterator it1 = pec.begin();
    Lines::const_iterator it1_end = pec.end();
    Lines::const_iterator it2_end = end();
    Lines::const_iterator it2 = begin();
    while(true) {
        if(it1 == it1_end || it2 == it2_end) {
            break;
        }
        while( (it1 != it1_end) && ((*it1) < (*it2)) ) {
            it1++;
        }
        if(it1 == it1_end) {
            return false;
        }
        while( (it2 != it2_end) && ((*it2) < (*it1)) ) {
            it2++;
        }
        if(it2 == it2_end) {
            return false;
        }
        if(it1->overlap(*it2)) {
            return true;
        }
    }
    return false;
}


//===--------------------------------------------------------------------------
//
//===--------------------------------------------------------------------------
void EqvPhiClass_Fer::dump() const {
    Lines::const_iterator it = this->begin();
    Lines::const_iterator end = this->end();
    while(it != end) {
        std::cerr << "[" << it->start << ", " << it->end << ")  ";
        it++;
    }
    std::cerr << "\nVirtuals: (";
    std::set<unsigned>::iterator pos;
    for (pos = virtuals.begin(); pos != virtuals.end(); ++pos) {
        unsigned r_src = * pos;
        std::cerr << "t" << r_src << "  ";
    }
    std::cerr << ")\n\n";
}
