//===-------------- FillEmptyMBB_Fer - Break critical edges --------------===//
//
//                     The LLVM Compiler Infrastructure
//
// Inserts a no-op into each basic block that is empty.
//
//===---------------------------------------------------------------------===//
//
// After this pass runs, the target function will not have any empty basic
// block. Each basic block that was previously empty will have a no-op
// instruction.
//
//===---------------------------------------------------------------------===//

#define DEBUG_TYPE "fillEmptyMBB"

#include <iostream>                                      // cerr

#include "llvm/Function.h"
#include "llvm/Instructions.h"
#include "llvm/CodeGen/MachineFunctionPass.h"            // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"

#include "llvm/Support/Debug.h"

#include "llvm/ADT/Statistic.h"

#include "llvm/fernando/CriticalEdgeRemoval_Fer.h"

#include "llvm/fernando/FillEmptyMBB_Fer.h"       // class definition

using namespace llvm;

STATISTIC(numBlocks, "Number of basic blocks in the source program.");
STATISTIC(numEdges,  "Number of edges in the source program.");

char FillEmptyMBB_Fer::ID = 0;

namespace {
    RegisterPass<FillEmptyMBB_Fer>
    X("fill_mbb", "Fill Emtpy MBB - Fernando", true);
}


//===---------------------------------------------------------------------===//
// This pass certainly changes the control flow graph, in the sense that it
// inserts no-op instruction in some basic blocks. However, I am assuming that
// this does not change any property of the target program.
//===---------------------------------------------------------------------===//
void FillEmptyMBB_Fer::getAnalysisUsage(AnalysisUsage & AU) const {
    AU.addRequired<CriticalEdgeRemoval_Fer>();
    AU.setPreservesAll();
}

//===---------------------------------------------------------------------===//
// Visits each basic block and inserts no-op into those that do not contain
// any instruction.
//===---------------------------------------------------------------------===//
bool FillEmptyMBB_Fer::runOnMachineFunction(MachineFunction & mf) {
    const TargetMachine & target_machine = mf.getTarget();
    const TargetInstrInfo & tii = * (target_machine.getInstrInfo());

    MachineFunction::iterator mbbi, mbbe;
    for (mbbi = mf.begin(), mbbe = mf.end(); mbbi != mbbe; mbbi++) {
        MachineBasicBlock * mbb = mbbi;
        numBlocks++;
        numEdges += mbb->succ_size();
        if(mbb->begin() == mbb->end()) {
            tii.insertNoop(*mbbi, mbb->end());
        } else if(mbb->back().getOpcode() == TargetInstrInfo::PHI) {
            tii.insertNoop(*mbbi, mbb->end());
        }
    }

    return false;
}
