//===---------------- MachineDomTree_Fer - Pass Template -----------------===//
//
//                     The LLVM Compiler Infrastructure
//
// This is a template to use when necessary to develope new passes.
//
//===---------------------------------------------------------------------===//
//
// PassTemplate is a file that should be used as a model for the implementation
// of new passes.
//
//===---------------------------------------------------------------------===//

#include <iostream>                              // cerr

#include "llvm/Analysis/Dominators.h"

#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/Target/TargetInstrInfo.h"

#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/CodeGen/LiveInterval.h"
#include "llvm/CodeGen/MachineFunctionPass.h"    // MachineBasicBlock
#include "llvm/CodeGen/LiveIntervalAnalysis.h"

#include "llvm/Instructions.h"                   // MachineOperand, etc

#include "llvm/ADT/DepthFirstIterator.h"         // df_ext_iterator

#include "llvm/fernando/MachineDomTree_Fer.h"    // class definition

using namespace llvm;

char MachineDomTree_Fer::ID = 0;

namespace {
    RegisterPass<MachineDomTree_Fer>
    X("m-dom-tree", "Dominator tree for machine basic blocks - Fernando");
}


//===--------------------------------------------------------------------------
// This method specifies which other passes are required, or are modified by
// this pass.
//===--------------------------------------------------------------------------
void MachineDomTree_Fer::getAnalysisUsage(AnalysisUsage & AU) const {
    AU.addRequired<DominatorTree>();
    AU.setPreservesAll();
}

//===--------------------------------------------------------------------------
// Free the memory, so the operating system can use it in other applications.
//===--------------------------------------------------------------------------
void MachineDomTree_Fer::releaseMemory() {
    this->dominators_.clear();
}


//===--------------------------------------------------------------------------
// Prints the list of machine basic blocks and their immediate dominators.
//===--------------------------------------------------------------------------
void MachineDomTree_Fer::print_dominators() const {
    for(unsigned u = 0; u < this->dominators_.size(); u++) {
        std::cout << "Basic block " << u << ": ";
        for(unsigned v = 0; v < this->dominators_[u].size(); v++) {
            if(this->dominators_[u][v]) {
                std::cout << v << " ";
            }
        }
        std::cout << std::endl;
    }
}


//===--------------------------------------------------------------------------
// This method prints the dominance tree.
//===--------------------------------------------------------------------------
void MachineDomTree_Fer::print_dom_tree() const {
    std::cout << "Dominator tree:\n";
    for(unsigned u = 1; u < this->rtree_.size(); u++) {
        std::cout << u << " <= " << this->rtree_[u] << std::endl;
    }
}


//===--------------------------------------------------------------------------
// This method performs one iteration of the simple algorithm to compute
// dominators. The method returns true if the iteration has changed the
// information that was previously known.
//===--------------------------------------------------------------------------
bool MachineDomTree_Fer::execute_one_iteration() {
    bool has_changed = false;

    for (MachineFunction::const_iterator i = this->mf_->begin(),
                                           e = this->mf_->end(); i != e; ++i) {
        const MachineBasicBlock * mbb = i;
        unsigned mbb_num = mbb->getNumber();
        std::vector<bool> & mbb_vec = this->dominators_[mbb_num];

        // compute intersection with predecessors:
        unsigned mbb_size = mbb_vec.size();
        std::vector<bool> intersec(mbb_size);

        for(MachineBasicBlock::const_pred_iterator pred_aux = mbb->pred_begin();
                                     pred_aux != mbb->pred_end(); pred_aux++) {
            const MachineBasicBlock *pred = *pred_aux;
            intersec.assign(mbb_size, true);
            unsigned pred_num = pred->getNumber();
            std::vector<bool> & pred_vec = this->dominators_[pred_num];
            for(unsigned u = 0; u < mbb_size; u++) {
                if(!pred_vec[u]) {
                    intersec[u] = false;
                }
            }
            for(unsigned u = 0; u < intersec.size(); u++) {
                if(u != mbb_num && mbb_vec[u] && !intersec[u]) {
                    mbb_vec[u] = false;
                    has_changed = true;
                }
            }
        }
    }

    return has_changed;
}


//===--------------------------------------------------------------------------
// This method builds the dominance tree of the target function.
//===--------------------------------------------------------------------------
void MachineDomTree_Fer::build_dominance_tree() {
    // First, builds the reverse dominance tree.
    for(unsigned u = 0; u < this->dominators_.size(); u++) {
        unsigned dom = 0;
        for(unsigned v = 0; v < this->dominators_[u].size(); v++) {
            if(v != u && this->dominators_[u][v] && this->dominates(dom, v)) {
                dom = v;
            }
        }
        this->rtree_[u] = dom;
    }
}


//===--------------------------------------------------------------------------
// Add any processing on the machine function here.
//===--------------------------------------------------------------------------
bool MachineDomTree_Fer::runOnMachineFunction(MachineFunction & mf) {
    this->mf_ = & mf;

    // Gets the number of basic blocks and initializes the dominators_ vec
    unsigned num_mbb = mf.size();
    this->dominators_.resize(num_mbb);
    this->rtree_.resize(num_mbb, 0);

    // Initialize each vector to contain all the basic blocks. This is a must
    // analysis
    for(unsigned u = 0; u < num_mbb; u++) {
        this->dominators_[u].resize(num_mbb, true);
    }

    // Gives the value of the initial basic block. It is set to be all false,
    // except for the first basic block.
    this->dominators_[0].assign(num_mbb, false);
    this->dominators_[0][0] = true;

    // Compute the dominators of each basic block:
    bool has_changed = true;
    do {
        has_changed = this->execute_one_iteration();
    } while(has_changed);

    this->build_dominance_tree();

//    this->print_dominators();

//    this->print_dom_tree();

    return false;
}
