//===--------------- P3_Fer.h - Clique substitution of P_3 ----------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This class represents a clique substitution of P_3
//
// Date: May 27th, 2007
//
//===----------------------------------------------------------------------===//
//
// Clique substitutions of P_3 are graphs that can be divided into three sets
// of nodes: X, Y and Z. Nodes in X, Y and Z form cliques. Also, nodes of X and
// Y are pairwise adjance. The same is true for nodes of Y and Z. However,
// nodes of X and Z are not adjacent. Thus, any of these graphs has two maximal
// cliques: X union Y and Y union Z.
//
//===----------------------------------------------------------------------===//

#define DEBUG_TYPE "p3graph"

#include <iostream>                              // cerr
#include <sstream>                               // ostringstream (RegClique)
#include <iomanip>                               // setw
#include <queue>
#include <set>
#include <map>

#include "llvm/Instructions.h"                   // MachineOperand, etc
#include "llvm/CodeGen/Passes.h"                 // RegAllocChord_Fer
#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/CodeGen/LiveInterval.h"
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/CodeGen/MachineFunctionPass.h"    // MachineBasicBlock
#include "llvm/CodeGen/LiveIntervalAnalysis.h"

#include "llvm/Support/Debug.h"

#include "llvm/ADT/SmallVector.h"
#include "llvm/ADT/STLExtras.h"
#include "llvm/ADT/Statistic.h"
#include "llvm/ADT/DepthFirstIterator.h"         // df_ext_iterator

#include "llvm/fernando/Regs_Fer.h"
#include "llvm/fernando/X86PuzzleBoard_Fer.h"
#include "llvm/fernando/Partition_Fer.h"
#include "llvm/fernando/RA_Result_Fer.h"
#include "llvm/fernando/RegMapping_Fer.h"
#include "llvm/fernando/PhiAnalysis_Fer.h"

#include "llvm/fernando/FacadeP3_Fer.h"
#include "llvm/fernando/P3_Fer.h"

using namespace llvm;

STATISTIC(zeroIter, "Puzzles solved without iterations.");
STATISTIC(maxIterations, "Maximum number of iterations due to spilling.");
STATISTIC(numMvs, "Number of moves fixing puzzles.");

//===--------------------------------------------------------------------------
// Constructor method.
//===--------------------------------------------------------------------------
P3_Fer::P3_Fer(
    MachineFunction & mf,
    MachineInstr * mi,
    Partition_Fer * partition,
    RegMapping_Fer *rm,
    RA_Result_Fer *rs,
    PhiAnalysis_Fer * pa
) : 
    mf_(& mf),
    mi_(mi),
    partition_(partition),
    rm_(rm),
    rs_(rs),
    pa_(pa),
    numIter_(0)
{
    // Discover the number of physical registers that must be handled:
    const TargetMachine & target_machine = mf.getTarget(); 
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    unsigned num_p_regs = reg_info->getNumRegs();
    this->p_regs_X_.resize(num_p_regs, RF_UNDEF);
    this->p_regs_Z_.resize(num_p_regs, RF_UNDEF);

    // Mark the fixed registers:
    for(unsigned u = 0; u < mi->getNumOperands(); u++) {
        MachineOperand & mo = mi->getOperand(u);
        if(mo.isRegister() && mo.getReg() && mo.isUse()) {
            if(MRegisterInfo::isVirtualRegister(mo.getReg())) {
                this->fixed_regs_.insert(mo.getReg());
            }
        }
    }

    // Initialize the data structures that will contain the pieces of the
    // puzzle:
    this->Xs_.resize(NUM_USES, 0);
    this->size_Xs_ = 0;
    this->XL_.resize(NUM_USES, 0);
    this->size_XL_ = 0;
    this->othersX_.resize(NUM_USES, 0);
    this->size_othersX_ = 0;

    this->Zs_ = 0;
    this->ZL_ = 0;
    this->othersZ_ = 0;

    this->num_precol_ = 0;

    this->state_ = P3_Fer::State::X;
}


//===--------------------------------------------------------------------------
// This method prints a vector of unsigned elements.
//===--------------------------------------------------------------------------
void print_vec(const std::vector<unsigned> & v, const MachineFunction & mf) {
    const TargetMachine & target_machine = mf.getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    for(unsigned u = 0; u < v.size(); u++) {
        unsigned reg = v[u];
        std::ostringstream str;
        str << reg << ":" << reg_info->getName(u);
        std::cout.width(10);
        if(reg != RF_UNDEF) {
            std::cout << str.str();
        }
    }
}


//===--------------------------------------------------------------------------
// This method prints the location of each variable that is alive at the
// beginning of the program point that this graph represents.
//===--------------------------------------------------------------------------
void P3_Fer::print_in_set() const {
    const TargetMachine & target_machine = this->mf_->getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    std::map<unsigned, unsigned>::const_iterator i;
    std::map<unsigned, unsigned>::const_iterator e;
    for(i = this->in_X_.begin(), e = this->in_X_.end(); i != e; i++) {
        std::ostringstream str;
        if(i->second == RF_SPILL) {
            str << i->first << ":[MEM]";
        } else if(i->second == RF_UNDEF) {
            str << i->first << ":[UDF]";
        } else {
            str << i->first << ":" << reg_info->getName(i->second);
        }
        std::cout.width(16);
        std::cout << str.str();
    }
    for(i = this->in_Y_.begin(), e = this->in_Y_.end(); i != e; i++) {
        std::ostringstream str;
        if(i->second == RF_SPILL) {
            str << i->first << ":[MEM]";
        } else if(i->second == RF_UNDEF) {
            str << i->first << ":[UDF]";
        } else {
            str << i->first << ":" << reg_info->getName(i->second);
        }
        std::cout.width(16);
        std::cout << str.str();
    }
    std::cout << std::endl;
}


//===--------------------------------------------------------------------------
// Prints the assignment of physical registers to virtuals in that exist in the
// X component, before this object instruction has been executed.
//===--------------------------------------------------------------------------
void P3_Fer::dump_X_colors() const {
    // Print the colors in the X component first:
    print_vec(this->p_regs_X_, * this->mf_);
    std::cout << "\n";
}


//===--------------------------------------------------------------------------
// Prints the assignment of physical registers to virtuals in the Z component,
// that is, after the instruction that this object surrounds has been executed.
//===--------------------------------------------------------------------------
void P3_Fer::dump_Z_colors() const {
    // Prints the colors in the Z component
    print_vec(this->p_regs_Z_, * this->mf_);
    std::cout << "\n";
}


//===--------------------------------------------------------------------------
// This method marks the given register as precolored in the X component of
// the graph.
//===--------------------------------------------------------------------------
void P3_Fer::mark_precol(unsigned p_reg, std::vector<unsigned> & vec) {
//std::cout << "Marking precolored " << p_reg << std::endl;
    assert(MRegisterInfo::isPhysicalRegister(p_reg) &&
                      "Only physical registers can be marked as precolored\n");
    this->num_precol_++;

    // Reset any virtual in p_reg:
    if(
        ! MRegisterInfo::isPhysicalRegister(vec[p_reg]) &&
        vec[p_reg] != RF_UNDEF &&
        vec[p_reg] != RF_SPILL
    ) {
        this->rm_->reset_color(vec[p_reg]);
    }
    vec[p_reg] = p_reg;

    // Reset virtuals in the aliases of p_reg
    const TargetMachine & target_machine = this->mf_->getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    unsigned p_size = X86PuzzleBoard_Fer::size(p_reg);
    for (const unsigned *a = reg_info->getAliasSet(p_reg); *a; ++a) {
        if(
            X86PuzzleBoard_Fer::size(*a) < p_size &&
            ! MRegisterInfo::isPhysicalRegister(vec[*a]) &&
            vec[*a] != RF_UNDEF &&
            vec[*a] != RF_SPILL
        ) {
            this->rm_->reset_color(vec[*a]);
            vec[*a] = *a;
        }
    }
}


//===--------------------------------------------------------------------------
// This method assigns the physical register p_reg to the variable v_reg in
// the vectors rb1 and rb2, where p_reg is the physical location of v_reg.
// Notice that this happens only if p_reg is not in use by any other
// register.
//===--------------------------------------------------------------------------
void P3_Fer::test_and_set_color(
    unsigned v_reg,
    std::vector<unsigned> & rb1,
    std::vector<unsigned> & rb2
) {
    unsigned p_reg = this->rm_->get_physical_location(v_reg);
    if(&rb1 == &rb2) {
        this->in_X_[v_reg] = p_reg;
    } else {
        this->in_Y_[v_reg] = p_reg;
    }
    if(MRegisterInfo::isPhysicalRegister(p_reg)) {
        if(rb1[p_reg] == RF_UNDEF && rb2[p_reg] == RF_UNDEF) {
            rb1[p_reg] = v_reg;
            rb2[p_reg] = v_reg;
        } else {
            this->rm_->reset_color(v_reg);
        }
    }
}


//===--------------------------------------------------------------------------
// Add 'reg' as a node of component X
//===--------------------------------------------------------------------------
void P3_Fer::addX(unsigned reg) {
//std::cout << "Add X " << reg << "\n";
    if(MRegisterInfo::isVirtualRegister(reg)) {
        const TargetRegisterClass *trc = mf_->getSSARegMap()->getRegClass(reg);
        if(X86PuzzleBoard_Fer::cl2piece(trc) ==
                                      X86PuzzleBoard_Fer::PuzzlePiece::SIZE1) {
            this->Xs_[this->size_Xs_] = reg;
            this->size_Xs_++;
        } else if(X86PuzzleBoard_Fer::cl2piece(trc) ==
                                       X86PuzzleBoard_Fer::PuzzlePiece::SIZE2){
            this->XL_[this->size_XL_] = reg;
            this->size_XL_++;
        } else {
            this->othersX_[this->size_othersX_] = reg;
            this->size_othersX_++;
        }
        this->test_and_set_color(reg, this->p_regs_X_, this->p_regs_X_);
    } else {
        this->mark_precol(reg, this->p_regs_X_);
    }
}


//===--------------------------------------------------------------------------
// Add 'reg' as a node of component Y
//===--------------------------------------------------------------------------
void P3_Fer::addY(unsigned reg) {
//std::cout << "Add Y " << reg << "\n";
    if(MRegisterInfo::isVirtualRegister(reg)) {
        const TargetRegisterClass *trc = mf_->getSSARegMap()->getRegClass(reg);
        if(X86PuzzleBoard_Fer::cl2piece(trc) ==
                                      X86PuzzleBoard_Fer::PuzzlePiece::SIZE1) {
            this->Ys_.insert(reg);
        } else if(X86PuzzleBoard_Fer::cl2piece(trc) ==
                                      X86PuzzleBoard_Fer::PuzzlePiece::SIZE2) {
            this->YL_.insert(reg);
        } else {
            this->othersY_.insert(reg);
        }
        this->test_and_set_color(reg, this->p_regs_X_, this->p_regs_Z_);
    } else {
        this->mark_precol(reg, this->p_regs_X_);
        this->mark_precol(reg, this->p_regs_Z_);
    }
}


//===--------------------------------------------------------------------------
// Add 'reg' as a node of component Z
//===--------------------------------------------------------------------------
void P3_Fer::addZ(unsigned reg) {
//std::cout << "Add Z " << reg << "\n";
    if(MRegisterInfo::isVirtualRegister(reg)) {
        const TargetRegisterClass *trc = mf_->getSSARegMap()->getRegClass(reg);
        if(X86PuzzleBoard_Fer::cl2piece(trc) ==
                                      X86PuzzleBoard_Fer::PuzzlePiece::SIZE1) {
            assert( (this->Zs_ == 0 || this->Zs_ == reg)
                              && "Instruction defines more than one virtual.");
            this->Zs_ = reg;
        } else if(X86PuzzleBoard_Fer::cl2piece(trc) ==
                                      X86PuzzleBoard_Fer::PuzzlePiece::SIZE2) {
            assert( (this->ZL_ == 0 || this->ZL_ == reg)
                              && "Instruction defines more than one virtual.");
            this->ZL_ = reg;
        } else {
            assert( (this->othersZ_ == 0 || this->othersZ_ == reg)
                              && "Instruction defines more than one virtual.");
            this->othersZ_ = reg;
        }
    } else {
        this->mark_precol(reg, this->p_regs_Z_);
    }
}


//===--------------------------------------------------------------------------
// This method attempts to coalesce the instruction of this graph. If the
// instruction is a move, and the coalescing is possible, the method
// returns true, otherwise it returns false.
//===--------------------------------------------------------------------------
bool P3_Fer::coalescible_move() {
    const TargetMachine & target_machine = this->mf_->getTarget();
    const TargetInstrInfo & tii = * (target_machine.getInstrInfo());
    unsigned use_reg, def_reg;
    if(tii.isMoveInstr(*this->mi_, use_reg, def_reg)) {
        // In order to coalesce this instruction, it is necessary that use_reg
        // be present in one of the X sets. We must look for it now:
        if(!MRegisterInfo::isVirtualRegister(def_reg)) {
            return false;
        }
        bool is_coalescible = false;
        if(this->size_Xs_) {
            is_coalescible = *(this->Xs_.begin()) == use_reg;
        } else if(this->XL_.size()) {
            is_coalescible = *(this->XL_.begin()) == use_reg;
        } else if(this->othersX_.size()) {
            is_coalescible = *(this->othersX_.begin()) == use_reg;
        }
        unsigned p_reg = this->rm_->get_physical_location(use_reg);
        if(
            is_coalescible &&
            p_reg != RF_UNDEF &&
            p_reg != RF_SPILL &&
            Regs_Fer::compatible_class(*this->mf_, def_reg, p_reg)
        ) {
            this->p_regs_X_[p_reg] = use_reg;
            this->p_regs_Z_[p_reg] = def_reg;
            this->rm_->allocate_color(def_reg, p_reg);
            this->rs_->add_binding(p_reg, def_reg, this->mi_);
            this->rs_->add_binding(p_reg, use_reg, this->mi_);
            return true;
        }
    }
    return false;
}


//===--------------------------------------------------------------------------
// This method calls the necessary routines to spill v_reg.
//===--------------------------------------------------------------------------
unsigned P3_Fer::spill_variable(unsigned v_reg, bool self_spill) {
    std::vector<unsigned> vec;
    unsigned new_color = this->partition_->choose_reg_and_spill
                                              (v_reg, * this, vec, self_spill);
    numIter_++;
    if(numIter_ > maxIterations) {
        maxIterations = numIter_;
    }
    // clean the spilled registers:
    for(unsigned u = 0; u < vec.size(); u++) {
        this->p_regs_X_[vec[u]] = RF_UNDEF;
        this->p_regs_Z_[vec[u]] = RF_UNDEF;
    }

    return new_color;
}


//===--------------------------------------------------------------------------
// This method colors the registers in the given set X. There are three such
// sets: shorts, longs and others.
// All the elements in X sets are fixed, and, once colored, they must be
// removed from the fixed list.
//===--------------------------------------------------------------------------
void P3_Fer::color_set_X(SmallVector<unsigned, NUM_USES> &vec, unsigned size) {
//std::cerr << "Color X\n";
    if(this->mi_->getOpcode() == TargetInstrInfo::PHI) {
        this->fixed_regs_.clear();
        assert(size == 0 && "Error: X component of phi instruction not empty");
    } else {
        for(unsigned i = 0; i < size; i++) {
            const unsigned v_reg = vec[i];
            const unsigned p_reg = this->rm_->get_physical_location(v_reg);
            unsigned new_color = p_reg;
            if(p_reg == RF_SPILL) {
                // v_reg is in memory. Let's try to fold it in the instruction.
                int slot = this->rm_->get_stack_slot(v_reg);
                unsigned op_num = Regs_Fer::get_operand_num(mi_, v_reg, false);
                MachineInstr * new_instr =
                  this->partition_->foldMemoryOperand(this->mi_, op_num, slot);
                if(new_instr) {
                    this->mi_ = new_instr;
                    // It is necessary to remove the operand from the in_map:
                    this->in_X_.erase(v_reg);
//std::cout << "Folding load of " << v_reg << " into " << NeatPrint_Fer::mi2string(*new_instr) << std::endl;
                    continue;
                }
            }
            if(
               p_reg == RF_UNDEF || p_reg == RF_SPILL ||
               this->p_regs_X_[p_reg] != v_reg ||
               ! Regs_Fer::are_aliases_free(v_reg, p_reg, p_regs_X_, * mf_)
            ) {
                // The virtual was in memory, or its current location is now a
                // precolored; therefore, it needs a new location:
                new_color = Regs_Fer::get_free_register
                         (v_reg, this->p_regs_X_, this->p_regs_X_, *this->mf_);
                if(new_color == RF_UNDEF) {
                    // There is no free color, and we must handle the spill.
                    new_color = this->spill_variable(v_reg, false);
                }
                this->p_regs_X_[new_color] = v_reg;
                this->rm_->allocate_color(v_reg, new_color);
            }
            this->rs_->add_binding(new_color, v_reg, this->mi_);
        }
    }
}


//===--------------------------------------------------------------------------
// This method colors the registers in the given Y set.
// The only situation in which it is not possible is if 'c' is precolored in
// this graph.
//===--------------------------------------------------------------------------
void P3_Fer::color_fixed_Y(std::set<unsigned> & set) {
    std::set<unsigned>::const_iterator i = set.begin();
    std::set<unsigned>::const_iterator e = set.end();
    std::set< unsigned > to_erase;
    while(i != e) {
        unsigned v_reg = *i;
        if(this->fixed_regs_.find(v_reg) != this->fixed_regs_.end()) {
            this->handle_fixed_Y(v_reg);
            // We erase this register, so that it is not colored again when the
            // component Y is colored.
            to_erase.insert(v_reg);
        }
        i++;
    }
    for(i = to_erase.begin(), e = to_erase.end(); i != e; i++) {
        unsigned v_reg = *i;
        std::set<unsigned>::const_iterator v_i = set.find(v_reg);
        set.erase(*i);
    }
}


//===--------------------------------------------------------------------------
// 1) The variable V is in register and its color can be reused:
// - do it
//
// 2) Else it is possible to transfer V to another register in X/Z:
// - do it
//
// 3) Else V is in register, its color X can be used, but Z can't:
// - Use the color X, and spill V
//
// 4) Else the variable is in register, but its colors X/Z can't be used:
// - Else, spill some variable V' and use give its color to V.
//===--------------------------------------------------------------------------
void P3_Fer::handle_fixed_Y(unsigned v_reg) {
//std::cerr << "Color fixed Y " << v_reg << std::endl;
    unsigned new_color = this->rm_->get_physical_location(v_reg);
    if(
        new_color == RF_UNDEF ||
        new_color == RF_SPILL ||
        this->p_regs_X_[new_color] != v_reg ||
        this->p_regs_Z_[new_color] != v_reg ||
        ! Regs_Fer::are_aliases_free(v_reg, new_color, p_regs_X_, * mf_) ||
        ! Regs_Fer::are_aliases_free(v_reg, new_color, p_regs_Z_, * mf_)
    ) {
        new_color = RF_UNDEF;
        if(this->pa_->has_preference(v_reg)) {
            unsigned pref_reg = this->pa_->get_preference(v_reg);
            if(
                Regs_Fer::is_register_available
                       (pref_reg, this->p_regs_X_, this->p_regs_Z_, *this->mf_)
            ) {
                new_color = pref_reg;
            }
        }
        if(new_color == RF_UNDEF) {
            new_color = Regs_Fer::get_free_register
                                          (v_reg, p_regs_X_, p_regs_Z_, * mf_);
        }
        if(new_color == RF_UNDEF) {
            new_color = this->spill_variable(v_reg, false);
        }
        this->p_regs_X_[new_color] = v_reg;
        this->p_regs_Z_[new_color] = v_reg;
        this->rm_->allocate_color(v_reg, new_color);
    }
    this->rs_->add_binding(new_color, v_reg, this->mi_);
}


//===--------------------------------------------------------------------------
// This method colors the registers in the given Y set. The vector "preference"
// guides the coloring. Every time it is possible, the method trys to assign to
// an element 'e' of "set" the color 'c' that is assigned to 'e' in
// "preference".
// The only situation in which it is not possible is if 'c' is precolored in
// this graph.
//===--------------------------------------------------------------------------
void P3_Fer::color_set_Y(std::set<unsigned> & set) {
//std::cout << "color_set_Y\n";
    std::set<unsigned>::const_iterator i = set.begin();
    std::set<unsigned>::const_iterator e = set.end();
    for(i = set.begin(), e = set.end(); i != e; i++) {
        unsigned v_reg = *i;
        const unsigned p_reg = this->rm_->get_physical_location(v_reg);
        // We have to test if p_reg != RF_UNDEF, because the reset_mapping
        // function assigns this value to spilled variables.
        if(
            p_reg == RF_SPILL ||
            (
                p_reg != RF_UNDEF &&
                this->p_regs_X_[p_reg] == v_reg &&
                this->p_regs_Z_[p_reg] == v_reg &&
                Regs_Fer::are_aliases_free(v_reg, p_reg, p_regs_X_, * mf_) &&
                Regs_Fer::are_aliases_free(v_reg, p_reg, p_regs_Z_, * mf_)
            )
        ) {
            continue;
        }
        // v_reg is located in a clobbered register. We must
        // save it either in another register, or in memory.
        // Let's try to find another free register first:
        unsigned new_color = RF_UNDEF;
        if(this->pa_->has_preference(v_reg)) {
            unsigned pref_reg = this->pa_->get_preference(v_reg);
            if(
                Regs_Fer::is_register_available
                       (pref_reg, this->p_regs_X_, this->p_regs_Z_, *this->mf_)
            ) {
                new_color = pref_reg;
            }
        }
        if(new_color == RF_UNDEF) {
            new_color = Regs_Fer::get_free_register
                                          (v_reg, p_regs_X_, p_regs_Z_, * mf_);
        }
        if(new_color == RF_UNDEF) {
            // Ok, there is no way to get a free register. So, spill:
            new_color = this->spill_variable(v_reg, true);
        }
        // There is a possibility that v_reg is the best candidate to be
        // spilled.
        if(MRegisterInfo::isPhysicalRegister(new_color)) {
            assert(new_color != p_reg && "Inconsistent spill.");
            // a different variable has been spilled. Take its register:
            this->p_regs_X_[new_color] = v_reg;
            this->p_regs_Z_[new_color] = v_reg;
            this->rm_->allocate_color(v_reg, new_color);
            this->rs_->add_binding(new_color, v_reg, this->mi_);
        }
    }
}


//===--------------------------------------------------------------------------
// This method frees the colors used by the elements in "vec". Colors are
// freed by invoking the method free_color on the RegMapping object.
//===--------------------------------------------------------------------------
void P3_Fer::free_color(
    SmallVector<unsigned, NUM_USES> & vec,
    unsigned size
) const {
//std::cerr << "Free color\n";
    for(unsigned u = 0; u < size; u++) {
        unsigned v_reg = vec[u];
        unsigned p_reg = this->rm_->get_physical_location(v_reg);
        //assert(p_reg != RF_UNDEF && "Attempt to free an uncolored virtual");
        if(MRegisterInfo::isPhysicalRegister(p_reg)) {
            this->rm_->free_color(p_reg);
        }
    }
}


//===--------------------------------------------------------------------------
// This method colors nodes in the set Z. These nodes must not have been
// assigned colors before. If this happens, an exception is thrown.
//===--------------------------------------------------------------------------
void P3_Fer::color_set_Z(unsigned v_reg) {
//std::cout << "color_set_Z\n";
    // The instruction is not 3-address...
    unsigned p_reg = this->rm_->get_physical_location(v_reg);
    if(p_reg == RF_UNDEF) {
        // We must find an available color for v_reg;
        if(this->pa_->has_preference(v_reg)) {
            unsigned pref_reg = this->pa_->get_preference(v_reg);
            if(
                Regs_Fer::is_register_available
                (pref_reg, this->p_regs_X_, this->p_regs_X_, *this->mf_)
            ) {
                p_reg = pref_reg;
            }
        }
        if(p_reg == RF_UNDEF) {
            // preference is not available. Find another color:
            p_reg = Regs_Fer::get_free_register
                             (v_reg, p_regs_Z_, p_regs_Z_, *this->mf_);
        }
        if(p_reg == RF_UNDEF) {
            // There is no free color, and we must handle the spill.
            p_reg = this->spill_variable(v_reg, false);
        }
        this->p_regs_Z_[p_reg] = v_reg;
        this->rm_->allocate_color(v_reg, p_reg);
        this->rs_->add_binding(p_reg, v_reg, this->mi_);
    }
}


//===--------------------------------------------------------------------------
// This method inserts copies/loads/stores at the program point that this graph
// covers in order to preserve the semantics of the program.
//===--------------------------------------------------------------------------
void P3_Fer::transfer_colors() {
    std::map<unsigned, unsigned>::iterator i;
    std::map<unsigned, unsigned>::iterator e;

    // Insert copies for the variables in X
    for(i = this->in_X_.begin(), e = this->in_X_.end(); i != e; i++) {
        unsigned reg = i->first;
        unsigned old_loc = i->second;
        unsigned new_loc = this->rm_->get_physical_location(reg);
        // The variables in the X set need to be located in registers.
        assert(MRegisterInfo::isPhysicalRegister(new_loc) &&
                                                  "Use of operand in memory?");
        if(new_loc != old_loc) {
            if(old_loc != RF_SPILL) {
                // copy:
                this->rs_->add_cp(new_loc, reg, old_loc, reg, this->mi_);
//std::cout << "<> Adding copy (" << new_loc << ", " << reg << "), (" << old_loc << ", " << reg << ")\n";
                numMvs++;
            } else {
                // load:
                unsigned slot = this->rm_->get_stack_slot(reg);
//std::cout << "<> Adding load (" << new_loc << ", " << reg << "), (" << slot << ", " << reg << ")\n";
                this->rs_->add_ld(new_loc, reg, slot, reg, this->mi_);
            }
        }
    }

    // Insert copies for the variables in Y
    for(i = this->in_Y_.begin(), e = this->in_Y_.end(); i != e; i++) {
        unsigned reg = i->first;
        unsigned old_loc = i->second;
        unsigned new_loc = this->rm_->get_physical_location(reg);
        if(new_loc != old_loc) {
            if(new_loc == RF_SPILL) {
            //    std::cout << "SPILL " << reg << std::endl;
            } else if(old_loc != RF_SPILL) {
                // copy:
                this->rs_->add_cp(new_loc, reg, old_loc, reg, this->mi_);
//std::cout << "<> Adding copy (" << new_loc << ", " << reg << "), (" << old_loc << ", " << reg << ")\n";
                numMvs++;
            } else {
                // load:
                unsigned slot = this->rm_->get_stack_slot(reg);
//std::cout << "<> Adding load (" << new_loc << ", " << reg << "), (" << slot << ", " << reg << ")\n";
                this->rs_->add_ld(new_loc, reg, slot, reg, this->mi_);
            }
        } else if(new_loc==RF_SPILL && MRegisterInfo::isVirtualRegister(reg)) {
            // There is a trick situation here: the variable may be a fixed
            // register that came in memory. It was loaded, and then, due to
            // some variable in X, it was spilled again. We must insert this
            // load.
            if(this->fixed_regs_.find(reg) != fixed_regs_.end()) {
                unsigned slot = this->rm_->get_stack_slot(reg);
                unsigned new_loc = this->rm_->get_last_location(reg);
                this->rs_->add_ld(new_loc, reg, slot, reg, this->mi_);
//std::cout << "<> Adding load (" << new_loc << ", " << reg << "), (" << slot << ", " << reg << ")\n";
            }
        }
    }
}


//===--------------------------------------------------------------------------
// True if reg_file[p_reg] is not physical register or reg_file[p_reg] is not
// in the list of fixed registers.
//===--------------------------------------------------------------------------
bool is_spillable_aux(
    unsigned p_reg,
    const std::vector<unsigned> & reg_file,
    const std::set<unsigned> & fix_regs
) {
    if(MRegisterInfo::isPhysicalRegister(reg_file[p_reg])) {
        return false;
    } else {
        unsigned v_reg = reg_file[p_reg];
        if(fix_regs.find(v_reg) != fix_regs.end()) {
            return false;
        }
    }
    return true;
}


//===--------------------------------------------------------------------------
// This method return true if p_reg is neither a precolored, nor a fixed
// register.
//===--------------------------------------------------------------------------
bool P3_Fer::is_spillable(unsigned p_reg) const {
    switch(this->state_) {
        case P3_Fer::State::X: {
            return is_spillable_aux(p_reg, this->p_regs_X_, this->fixed_regs_);
        }
        case P3_Fer::State::Z: {
            return is_spillable_aux(p_reg, this->p_regs_Z_, this->fixed_regs_);
        }
        case P3_Fer::State::Y: {
            bool auxX = is_spillable_aux(p_reg, p_regs_X_, this->fixed_regs_);
            bool auxZ = is_spillable_aux(p_reg, p_regs_Z_, this->fixed_regs_);
            return auxX && auxZ;
        }
        default:
            assert(false && "Unknow state.");
    }
    return false;
}

//===--------------------------------------------------------------------------
// This method returns true if the contents of p_reg can be evicted to memory.
// The only situations when this cannot happen is if p_reg is precolored, or if
// p_reg is holding a virtual that is used in the current instruction.
//===--------------------------------------------------------------------------
bool P3_Fer::can_spill(unsigned p_reg) const {

    assert(MRegisterInfo::isPhysicalRegister(p_reg) &&
                                    "Only physical registers can be spilled.");

    // first, we check if p_reg is spillable:
    if(!is_spillable(p_reg)) {
        return false;
    } else {
        const TargetMachine & target_machine = this->mf_->getTarget();
        const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
        for (const unsigned *a = reg_info->getAliasSet(p_reg); *a; ++a) {
            if (!is_spillable(*a)) {
                return false;
            }
        }
    }

    return true;
}


//===--------------------------------------------------------------------------
// This method colors this clique substitution of P_3 with the free registers
// of this object.
//===--------------------------------------------------------------------------
MachineInstr * P3_Fer::color() {
    if(this->coalescible_move()) {
        return this->mi_;
    }
    // Reserve space for the vector:
    this->state_ = P3_Fer::State::X;

    // First, we color the nodes in X:
    this->color_set_X(this->Xs_, this->size_Xs_);
    this->color_set_X(this->XL_, this->size_XL_);
    this->color_set_X(this->othersX_, this->size_othersX_);

    this->state_ = P3_Fer::State::Y;

    // Color the fixed nodes in Y:
    if(this->fixed_regs_.size() > 0) {
        this->color_fixed_Y(this->Ys_);
        this->color_fixed_Y(this->YL_);
        this->color_fixed_Y(this->othersY_);
    }

    this->color_set_Y(this->Ys_);
    this->color_set_Y(this->YL_);
    this->color_set_Y(this->othersY_);

    // The nodes in X do not interfere with the nodes in Z, so their colors
    // can be made available for coloring:
    this->free_color(this->Xs_, this->size_Xs_);
    this->free_color(this->XL_, this->size_XL_);
    this->free_color(this->othersX_, this->size_othersX_);

    this->state_ = P3_Fer::State::Z;

    // Color the nodes in Z:
    if(X86PuzzleBoard_Fer::is_two_addr(this->mi_)) {
        // If the internal instruction is a case of two address instruction,
        // the definition and the use were merged together as a Y interval.
        unsigned v_def_reg = this->mi_->getOperand(0).getReg();
        this->rm_->set_defined_in_2addr_instr(v_def_reg);
    } else {
        if(this->Zs_) {
            this->color_set_Z(this->Zs_);
        } else if (this->ZL_) {
            this->color_set_Z(this->ZL_);
        } else if (this->othersZ_) {
            this->color_set_Z(this->othersZ_);
        }
    }

    if(this->numIter_ == 0) {
        zeroIter++;
    }

    return this->mi_;
}
