//===----------------- PassTemplate_Fer - Pass Template ------------------===//
//
//                     The LLVM Compiler Infrastructure
//
// This is a template to use when necessary to develope new passes.
//
//===---------------------------------------------------------------------===//
//
// PassTemplate is a file that should be used as a model for the implementation
// of new passes.
//
//===---------------------------------------------------------------------===//

#define DEBUG_TYPE "partition"

#include <iostream>                              // cerr
#include <queue>
#include <map>

#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/Target/TargetInstrInfo.h"

#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/CodeGen/LiveInterval.h"
#include "llvm/CodeGen/MachineFunctionPass.h"    // MachineBasicBlock
#include "llvm/CodeGen/LiveIntervalAnalysis.h"

#include "llvm/Support/Debug.h"

#include "llvm/Instructions.h"                   // MachineOperand, etc

#include "llvm/ADT/STLExtras.h"
#include "llvm/ADT/Statistic.h"

#include "llvm/fernando/P3_Fer.h"
#include "llvm/fernando/Regs_Fer.h"
#include "llvm/fernando/Line_Fer.h"
#include "llvm/fernando/PhiP3_Fer.h"
#include "llvm/fernando/RevVec_Fer.h"
#include "llvm/fernando/FacadeP3_Fer.h"
#include "llvm/fernando/RA_Result_Fer.h"
#include "llvm/fernando/NeatPrint_Fer.h"         // Prints formated output
#include "llvm/fernando/RegMapping_Fer.h"
#include "llvm/fernando/PhiAnalysis_Fer.h"
#include "llvm/fernando/DefUseSites_Fer.h"
#include "llvm/fernando/AllocOptions_Fer.h"
#include "llvm/fernando/MachineDomTree_Fer.h"

#include "llvm/fernando/X86PuzzleBoard_Fer.h"

#include "llvm/fernando/Partition_Fer.h"         // class definition

using namespace llvm;

char Partition_Fer::ID = 0;

namespace {
    STATISTIC(numIter, "Number of iterations due to spilling");
    STATISTIC(numEMPT, "Empty puzzles.");
    STATISTIC(numPCST, "Precol and only shorts nodes.");
    STATISTIC(numPCLG, "Precol and only long nodes.");
    STATISTIC(numPCSL, "Precol and short/long nodes.");
    STATISTIC(numUNST, "No precol and only shorts nodes.");
    STATISTIC(numUNLG, "No precol and only long nodes.");
    STATISTIC(numUNSL, "No precol and short/long nodes.");

    RegisterPass<Partition_Fer>
    X("partition", "Register assignment - Fernando");
}


//===--------------------------------------------------------------------------
// This method specifies which other passes are required, or are modified by
// this pass.
//===--------------------------------------------------------------------------
void Partition_Fer::getAnalysisUsage(AnalysisUsage & AU) const {
    AU.addRequired<MachineDomTree_Fer>();
    AU.addRequired<PhiAnalysis_Fer>();
    AU.addRequired<DefUseSites_Fer>();
    AU.addRequired<LiveIntervals>();
    AU.setPreservesAll();
}

//===--------------------------------------------------------------------------
// Free the memory, so the operating system can use it in other applications.
//===--------------------------------------------------------------------------
void Partition_Fer::releaseMemory() {
    if(SSA_based) {
        // I am not using one insertion like the one below because the last
        // intervals in the active may not be removed, and yet, they would have
        // been processed.
        // assert(this->active_.size() == 0 &&
        //           "There are still active intervals that must be handled.");
        assert(this->unhandled_.empty() &&
                              "The list of unhandled intervals is not empty.");
        this->active_.clear();
        this->bb2Pref_.clear();
        this->bb_phi_.clear();
        this->headers_.clear();
        this->terminators_.clear();
        this->rm_->releaseMemory();
    }
}


//===--------------------------------------------------------------------------
// This method fills the data structures with the data that they must have in
// order to start the processing of intervals.
//===--------------------------------------------------------------------------
void Partition_Fer::initialize_data_structures(MachineFunction & mf) {
    this->mf_ = & mf;

    const TargetMachine & target_machine = this->mf_->getTarget();
    this->ri_ = target_machine.getRegisterInfo();

    this->live_intervals_ = & getAnalysis<LiveIntervals>();

    this->du_ = & getAnalysis<DefUseSites_Fer>();

    // LiveIntervals is a structure like: std::map<unsigned, LiveInterval>
    for (LiveIntervals::iterator i = this->live_intervals_->begin(),
                               e = this->live_intervals_->end(); i != e; ++i) {
        LiveInterval & li = i->second;
        LiveInterval::iterator r_i = li.begin();
        LiveInterval::iterator r_end = li.end();
        Line_Fer *last_seen = NULL;
        while(r_i != r_end) {
            LiveRange * r = r_i;
            if(last_seen != NULL && last_seen->end == r->start) {
                last_seen->end = r->end;
            } else {
                last_seen = new Line_Fer(r->start, r->end, i->second.reg);
            }
            this->unhandled_.push(last_seen);
            r_i++;
        }
    }

    // Mark which instructios are headers of basic blocks:
    this->mark_delimiters();

    // Initialize the bb2Pref_ mapping:
    this->bb2Pref_.resize(this->mf_->size());

    // Initialize the bb_phi_ mapping:
    this->bb_phi_.resize(this->mf_->size());

    // Gets information about how virtuals are related to phi-functions.
    this->pa_ = & getAnalysis<PhiAnalysis_Fer>();

    this->dt_ = & getAnalysis<MachineDomTree_Fer>();

    // Initialite the register mapping:
    this->rm_ = new RegMapping_Fer(*this->mf_, this->pa_);

    this->rs_ = new RA_Result_Fer(*mf_, live_intervals_, rm_, this, du_);
}


//===--------------------------------------------------------------------------
// If the interval on top of unhandled_ is in the range of this->c_indx_, this
// method removes this interval, adds it to active_ and then returns it. If
// there is no such interval, NULL is returned instead.
//===--------------------------------------------------------------------------
void Partition_Fer::remove_top_intervals() {

    // This is the index of the limit of instruction mi. Any interval that
    // starts below this index and after 'this->c_indx_' is defined by mi.
    unsigned lim_index = this->c_indx_ + LiveIntervals::InstrSlots::STORE;

    while(!this->unhandled_.empty()) {
        Line_Fer *lf = this->unhandled_.top();
        if(lf->start <= lim_index && lf->start >= this->c_indx_) {
            this->activate_interval(lf);
        } else {
            break;
        }
    }
}


//===--------------------------------------------------------------------------
// If the instruction at index is the header of a basic block b, this
// method returns the index of b, otherwise it returns NO_DELIM.
//===--------------------------------------------------------------------------
unsigned Partition_Fer::getHeadedBB(unsigned index) const {
    unsigned cell = (int)(index / LiveIntervals::InstrSlots::NUM);
    return this->headers_[cell];
}


//===--------------------------------------------------------------------------
// If the instruction at index is the terminator of a basic block b, this
// method returns the index of b, otherwise it returns NO_DELIM.
//===--------------------------------------------------------------------------
unsigned Partition_Fer::getTerminatedBB(unsigned index) const {
    unsigned cell = (int)(index / LiveIntervals::InstrSlots::NUM);
    return this->terminators_[cell];
}



//===--------------------------------------------------------------------------
// The Partition_Fer class uses two vectors, headers_ and terminators_ to
// mark the state of the bank of registers at the beginning and at the end of
// each basic block.
// This method marks in the headers_ vector which instructions are the first
// instruction of some basic block.
// The method also marks in the terminators_ vector which instructions are
// last instruction in some basic blocks.
//===--------------------------------------------------------------------------
void Partition_Fer::mark_delimiters() {
    // The variable num is just to trim code lines:
    unsigned num = LiveIntervals::InstrSlots::NUM;

    unsigned max_index = (int)(this->live_intervals_->maxInstrIndex_ / num);
    this->headers_.resize(max_index, NO_DELIM);
    this->terminators_.resize(max_index, NO_DELIM);

    for (MachineFunction::iterator mbb = this->mf_->begin(),
                             mbbEnd = this->mf_->end(); mbb != mbbEnd; ++mbb) {
        unsigned bb_number = mbb->getNumber();
        // get the first instruction of the basic block:
        MachineBasicBlock::iterator mi = mbb->begin();
        assert(mi != mbb->end() && "Empty basic block being partitioned.");
        unsigned m_index = this->live_intervals_->getInstructionIndex(mi);
        unsigned cell = (int)(m_index / num);
        this->headers_[cell] = bb_number;

        // get the last instruction of the basic block:
        MachineBasicBlock::iterator ei = mbb->getFirstTerminator();
        if(ei == mbb->end()) {
            ei--;
        }
        m_index = this->live_intervals_->getInstructionIndex(ei);
        cell = (int)(m_index / num);
        this->terminators_[cell] = bb_number;
    }
}


//===--------------------------------------------------------------------------
// This method refills the preference vector with the registers in use at the
// end of basic block 'b'.
//===--------------------------------------------------------------------------
void Partition_Fer::reset_preference(unsigned b) {

    // The preferences for basic block mbb are defined by the set of variables
    // alive at the end of the basic block that dominates b.
    unsigned dom_num = dt_->get_idom(b);
    if(this->active_.size() > 0) {
        // Because there are active intervals at the beginning of b, some
        // predecessor of b has already been visited.
        while(this->bb2Pref_[dom_num].size() == 0) {
            dom_num = dt_->get_idom(dom_num);
        }
        // We must store the set of registers alive at the beginning of the
        // basic block. It will be used during the SSA-elimination phase.
        // But if the basic block has phi-functions, we must wait until those
        // instructions are visited, otherwise we can do it just now.
        this->rm_->reset_mapping(this->bb2Pref_[dom_num], this->active_);
        MachineBasicBlock * mbb = this->mf_->getBlockNumbered(b);
        if(mbb->begin() != mbb->end()) {
            if(mbb->begin()->getOpcode() != TargetInstrInfo::PHI) {
                rm_->store_v2p(this->bb_phi_[b], active_);
            }
        }
    } else {
        this->rm_->clear_mapping();
    }
}


//===--------------------------------------------------------------------------
// This method returns the physical register which contains the virtual with
// the lowest spilling cost among all the live virtuals.
// This method also fills sp_regs with the aliases of all the registers that
// were spilled to free the returned register.
// Parameters:
// v_parameter: the virtual that is in need of a register. The spilled virtual
//   must be in the same class as v_parameter.
// sp_regs: this vector will contain the aliases of the spilled virtuals.
//===--------------------------------------------------------------------------
unsigned Partition_Fer::choose_reg_and_spill(
    unsigned v_parameter,
    FacadeP3_Fer & p3_graph,
    std::vector<unsigned> & sp_regs,
    bool self_spill
) const {
    // Tell du_ in which index we are. The best spill candidate is the virtual
    // that will be used the farthest from c_indx_
    this->du_->advance_index(this->c_indx_);

    assert(this->active_.size() > 0 && "No register to spill.");

    std::set<UseVec_Fer * , UseVecComp_Fer >::iterator i = du_->end();
    std::set<UseVec_Fer * , UseVecComp_Fer >::iterator b = du_->begin();

    // The virtual that will be spilled:
    unsigned v_cur = RF_UNDEF;

    // The register that contains the virtual that will be spilled
    unsigned p_reg = RF_UNDEF;

    do {
        p_reg = RF_UNDEF;
        i = prior(i);
        UseVec_Fer * uv = *i;
        v_cur = uv->reg_;

        if(self_spill && v_parameter == v_cur) {
            this->rm_->set_spill(v_cur);
            break;
        }

        // p_reg is a physical register. We must get the alias of p_reg that
        // can be used to store v_parameter, if one exist.
        p_reg = this->rm_->get_physical_location(v_cur);
        if(MRegisterInfo::isPhysicalRegister(p_reg)) {
            unsigned p_reg_cl;
            p_reg_cl = Regs_Fer::get_reg_class(p_reg, v_parameter, * mf_);
            unsigned v_top = RF_UNDEF;
            if(p_reg_cl != RF_UNDEF && p3_graph.can_spill(p_reg_cl)) {
                if(this->rm_->number_of_occupied_aliases(p_reg_cl) <= 1) {
                    // v_top and v_cur are the same.
                    v_top = v_cur;
                } else {
                    v_top = rm_->next_nearest_alias_used(p_reg_cl, this->du_);
                }
            }

            if(v_top == v_cur) {
                assert(uv->end() > c_indx_ + LiveIntervals::InstrSlots::USE
                                     && "Only zero lenght intervals to spill");
                assert(MRegisterInfo::isVirtualRegister(v_cur)
                                      && "Attempt to spill physical register");
                this->rm_->free_register(p_reg_cl, sp_regs);
                p_reg = p_reg_cl;
                break;
            }
        }

    } while(i != b);
//std::cout << "\n*\nSpilling " << p_reg << "\n*\n";
    return p_reg;
}


//===--------------------------------------------------------------------------
// This method makes the interval lf to be active. An active interval has been
// scanned by the register allocator.
//===--------------------------------------------------------------------------
void Partition_Fer::activate_interval(Line_Fer *lf) {
    this->active_.insert(lf);
    this->unhandled_.pop(); 
    if(MRegisterInfo::isVirtualRegister(lf->reg)) {
        this->du_->set_active(lf->reg, lf->end);
    }
}


//===--------------------------------------------------------------------------
// This method process all the intervals that exist between two consecutive
// program points, or crossbars. This name, crossbar, was taken from David Koes
// PLDI paper, Incremental Register Allocation.
//===--------------------------------------------------------------------------
void Partition_Fer::process_intervals(P3_Fer & p3_graph) {
    // Iterators to range over active:
    std::set<Line_Fer *, LineCompareEnd>::iterator i;
    std::set<Line_Fer *, LineCompareEnd>::iterator e;

    // Handle the intervals that end at the very beginning of the
    // instruction. They must be removed from active_
    for(i = this->active_.begin(), e = this->active_.end(); i != e;) {
        Line_Fer *lf = *i;
        if(lf->end <= this->c_indx_) {
            this->active_.erase(i++);
        } else {
            break;
        }
    }

    // Handle the intervals that start at the very beginning of the 
    // instruction. They must be inserted into active_ and in the Y
    // component of p3_graph.
    while(!this->unhandled_.empty()) {
        Line_Fer *lf = this->unhandled_.top();
        if(lf->start == this->c_indx_) {
            this->activate_interval(lf);
        } else {          
            break;
        }
    }

    // If we are at the beginning of a basic block, the bank of registers must
    // be restarted.
    if(this->c_indx_ > 0) {
        unsigned delim = getHeadedBB(c_indx_);
        if(delim != NO_DELIM) {
            this->reset_preference(delim);
        }
    }
 
    // Add the intervals that end at USE to the X component of p3_graph.
    // This is the index where end intervals killed at mi. I am adding
    // "1" to this index because this is done in LiveIntervalAnalysis:
    unsigned use_index = this->c_indx_ + LiveIntervals::InstrSlots::USE + 1;
    for(i = this->active_.begin(), e = this->active_.end(); i != e;) {
        Line_Fer *lf = *i;
        if(lf->end <= use_index) {
            p3_graph.addX(lf->reg);
            this->active_.erase(i++);
        } else {
            break;
        }
    }

    // Add the intervals that are alive until the end of the instruction to
    // the Y component.
    for(i = this->active_.begin(), e = this->active_.end(); i != e; i++) {
        p3_graph.addY((*i)->reg);
    }

    // Add the intervals that start at USE to the Z component of the
    // instruction.
    unsigned store_index = this->c_indx_ + LiveIntervals::InstrSlots::STORE;
    while(!this->unhandled_.empty()) {
        Line_Fer *lf = this->unhandled_.top();
        if(lf->start <= store_index) {
            assert(lf->start > this->c_indx_ &&
                                         "Unhandled interval after the scan.");
            p3_graph.addZ(lf->reg);
            this->activate_interval(lf);
        } else {
            break;
        }
    }

    // Handle the intervals that end before the end of the instruction. They
    // must be removed from active_
    unsigned lim_index = this->c_indx_ + LiveIntervals::InstrSlots::NUM;
    for(i = this->active_.begin(), e = this->active_.end(); i != e;) {
        Line_Fer *lf = *i;
        if(lf->end < lim_index) {
            this->active_.erase(i++);
        } else {
            break;
        }
    }
}


//===--------------------------------------------------------------------------
// This method computes the type of the clique substitution of P3, based on the
// types of each of its nodes: short/long or precolored/uncolored.
// This method returns 0 if, and only if, the puzzle is empty.
//===--------------------------------------------------------------------------
unsigned Partition_Fer::compute_statistics(P3_Fer & p3_graph) {
    unsigned type = 0;
    if(p3_graph.contains_short()) {
        type += SHRT;
    }
    if(p3_graph.contains_long()) {
        type += LONG;
    }
    if(type == 0) {
        //the graph is empty.
        numEMPT++;
        return 0;
    }
    if(p3_graph.contains_precol()) {
        type += PRCL;
    } else {
        type += UNCL;
    }
    switch(type) {
        case (SHRT + PRCL): numPCST++;
            break;
        case (LONG + PRCL): numPCLG++;
            break;
        case (STLG + PRCL): numPCSL++;
            break;
        case (SHRT + UNCL): numUNST++;
            break;
        case (LONG + UNCL): numUNLG++;
            break;
        case (STLG + UNCL): numUNSL++;
            break;
        default: assert(false && "Unknow graph type");
    }
    return type;
}


//===--------------------------------------------------------------------------
// This function copies the contents of src into dst.
//===--------------------------------------------------------------------------
void copy_map(
    const std::map<unsigned, unsigned> & src,
    std::map<unsigned, unsigned> & dst
) {
    std::map<unsigned, unsigned>::const_iterator i;
    std::map<unsigned, unsigned>::const_iterator e;
    for(i = src.begin(), e = src.end(); i != e; i++) {
        dst[i->first] = i->second;
    }
}

#define UNDEF_BLOCK -1

//===--------------------------------------------------------------------------
// This method scans all the intervals, divide them into sets of clique
// substitutions of P_3 and assign registers to them.
//===--------------------------------------------------------------------------
void Partition_Fer::scan_intervals() {
    // This variable holds the number of a basic block. It will be used to
    // implement a small state machine that will track when to store the
    // liveness analysis information after the declaration of phi-functions.
    // IMPORTANT: this code assumes that a basic block never ends with phi
    // functions. Actually, this is guaranteed by the FillEmptyMBB_Fer pass.
    int phi_mbb_num = UNDEF_BLOCK;

    for(this->c_indx_ = 0; this->c_indx_ < live_intervals_->maxInstrIndex_;
                             this->c_indx_ += LiveIntervals::InstrSlots::NUM) {
        // Get a reference to the instruction at index 'this->c_indx_':
        MachineInstr *mi = live_intervals_->getInstructionFromIndex(c_indx_);

        if(mi == NULL) {
            // We are at the end of a basic block, or in a space without
            // instructions. Just add the intervals in this space to the
            // active set.
            this->remove_top_intervals();
            continue;
        } else {
            if(mi->getOpcode() == TargetInstrInfo::PHI) {
                if(phi_mbb_num == UNDEF_BLOCK) {
                    // We are visiting the first instruction of a basic block
                    // that starts with a phi-function.
                    phi_mbb_num = mi->getParent()->getNumber();
                }
            } else {
                if(phi_mbb_num != UNDEF_BLOCK) {
                    // The first instruction of the current basic block was a
                    // phi-function. We have already gone through every phi
                    // function, and now we have to store the preference for
                    // this basic block.
                    assert(phi_mbb_num == mi->getParent()->getNumber() &&
                                       "Basic block ended with phi-function!");
                    rm_->store_v2p(this->bb_phi_[phi_mbb_num], active_);
                    phi_mbb_num = UNDEF_BLOCK;
                }
            }
        }
        // This variable represents the interference graph defined by the live
        // ranges of live intervals that are alive in the area of instr mi.
        P3_Fer p3_graph(*mf_, mi, this, rm_, rs_, pa_);

        this->process_intervals(p3_graph);

        this->compute_statistics(p3_graph);

/* DEBUGGING CODE STARTS HERE ************************************************/
//std::cout << "Basic block " << mi->getParent()->getNumber() << " =======================================================\n";
//this->rm_->dump();
//std::cout << "-    -    -    -    -    -    -    -    -    -    -    -    -\n";
/* DEBUGGING CODE ENDS HERE **************************************************/

        // Color the KP3 that represents live ranges around this->c_indx_:
        mi = p3_graph.color();
        numIter += p3_graph.get_num_iterations();

        p3_graph.transfer_colors();

/* DEBUGGING CODE STARTS HERE ************************************************/
//p3_graph.print_in_set();
//std::cout << "-    -    -    -    -    -    -    -    -    -    -    -    -\n";
//p3_graph.dump_X_colors();
//std::cout << this->c_indx_ << ":\t" << NeatPrint_Fer::mi2string(*mi, *this->mf_) << "\n";
//p3_graph.dump_Z_colors();
//for(i = this->active_.begin(), e = this->active_.end(); i != e; i++) {
//    std::cout << (*i)->reg << ", (" << (*i)->start << ", " << (*i)->end << ")\n";
//}
//if(mi->getOpcode() != TargetInstrInfo::PHI) {
//    this->print_instruction(mi);
//}
/* DEBUGGING CODE ENDS HERE **************************************************/

        unsigned delim;
        if((delim = this->getTerminatedBB(this->c_indx_)) != NO_DELIM) {
            MachineBasicBlock * mbb = mi->getParent();
            if(mbb->succ_begin() != mbb->succ_end()) {
                this->rm_->store_v2p(bb2Pref_[mbb->getNumber()], active_);
            }
        }
    }
}

//===--------------------------------------------------------------------------
// This method implements instruciton folding. The value in slot must be loaded
// into register, or the value in register must be stored in the memory slot.
// Instead of performing the reload directly, this method attempts to replace
// the instruction old_mi by an equivalent instruction that accesses memory
// directly. The parameter op_num denotes the operand number of the register in
// which the instruction must be loaded.
//===--------------------------------------------------------------------------
MachineInstr * Partition_Fer::foldMemoryOperand(
    MachineInstr * old_mi,
    unsigned op_num,
    int slot
) {
    MachineInstr * new_mi = this->ri_->foldMemoryOperand(old_mi, op_num, slot);
    if(new_mi) {
        this->live_intervals_->replace_instruction(new_mi, old_mi);
    }
    return new_mi;
}


//===--------------------------------------------------------------------------
// This is an auxiliary method for the method print_instruction.
//===--------------------------------------------------------------------------
void print_reg(unsigned reg, RegMapping_Fer * rm) {
    unsigned p_reg = rm->get_physical_location(reg);
    if(!MRegisterInfo::isPhysicalRegister(p_reg)) {
        p_reg = rm->get_last_location(reg);
    }
    Regs_Fer::print_reg(p_reg, reg);
}


//===--------------------------------------------------------------------------
// This method prints all the variables defined/used in the instruction
// mi in a format suitable for debugging.
//===--------------------------------------------------------------------------
void Partition_Fer::print_instruction(MachineInstr * mi) const {
    const TargetMachine & target_machine = this->mf_->getTarget();
    const TargetInstrInfo & tii = * (target_machine.getInstrInfo());
    const TargetInstrDescriptor & tid = tii.get(mi->getOpcode());
    if (tii.isTerminatorInstr(mi->getOpcode())) {
        std::cout << "JMP ";
        MachineBasicBlock * mbb = mi->getParent();
        if(mbb->succ_begin() == mbb->succ_end()) {
            std::cout << "L" << this->mf_->getNumBlockIDs();
        } else {
            for(MachineBasicBlock::const_succ_iterator scc = mbb->succ_begin();
                                               scc != mbb->succ_end(); scc++) {
                std::cout << "L" << (*scc)->getNumber() << " ";
            }
        }
        std::cout << " ;\n\n";
    } else {
        // print the implict definitions:
        if(tid.ImplicitDefs) {
            for (const unsigned *implicit_def = tid.ImplicitDefs;
                                               *implicit_def; ++implicit_def) {
                print_reg(* implicit_def, this->rm_);
            }
        }
 
        // print the explict definitions:
        for(unsigned u = 0; u < mi->getNumOperands(); u++) {
            const MachineOperand & mo = mi->getOperand(u);
            if(mo.isRegister() && mo.getReg() && mo.isDef()) {
                print_reg(mo.getReg(), this->rm_);
            }
        }
 
        std::cout << " = ";
 
        // print the explicit uses:
        for(unsigned u = 0; u < mi->getNumOperands(); u++) {
            const MachineOperand & mo = mi->getOperand(u);
            if(mo.isRegister() && mo.getReg() && mo.isUse()) {
                print_reg(mo.getReg(), this->rm_);
            }
        }
 
        // print the implicit uses:
        if(tid.ImplicitUses) {
            for(const unsigned *implicit_use = tid.ImplicitUses;
                                               *implicit_use; implicit_use++) {
                print_reg(* implicit_use, this->rm_);
            }
        }

        std::cout << " ;\n";
    }
}


//===--------------------------------------------------------------------------
// This method is used to find liveness information for each basic block that
// has no instructions. For these basic blocks, the method scan_intervals will
// not assign a liveness mapping. This mapping is necessary to destroy
// phi-functions.
//===--------------------------------------------------------------------------
void Partition_Fer::store_liveness_for_null_blocks() {
    std::queue<MachineBasicBlock *> work_list;

    // We will go though all the basic blocks, and will insert those that have
    // no instructions into the worklist.
    MachineFunction::iterator mbbi, mbbe;
    for (mbbi = mf_->begin(), mbbe = mf_->end(); mbbi != mbbe; mbbi++) {
        MachineBasicBlock * mbb = mbbi;
        if(mbb->begin() == mbb->end()) {
            assert(mbb->succ_size() <= 1 && "Empty block with no successor.");
            if(mbb->succ_size() == 0) {
                std::cerr << "WARNING: " << mbb->getNumber() <<
                               " is empty basic block with zero successors.\n";
            } else {
                unsigned mbb_num = mbb->getNumber();
                assert(this->bb2Pref_[mbb_num].size() == 0 &&
                                      "Null basic block has preference list.");
                assert(this->bb_phi_[mbb_num].size() == 0 &&
                                             "Null basic block has phi list.");
                if(mbb->pred_size() == 1) {
                    unsigned pred_num = (*mbb->pred_begin())->getNumber();
                    copy_map(bb2Pref_[pred_num], bb2Pref_[mbb_num]);
                } else {
                    work_list.push(mbb);
                }
            }
        }
    }

    // The set visited will contain the basic blocks that have already been
    // handled.
    std::set<MachineBasicBlock *> visited;
    while(!work_list.empty()) {
        MachineBasicBlock * mbb = work_list.front();
        work_list.pop();
        MachineBasicBlock * succ_mbb = * mbb->succ_begin();
        if(
            succ_mbb->begin() != succ_mbb->end() ||
            visited.find(succ_mbb) != visited.end()
        ) {
            unsigned mbb_num = mbb->getNumber();
            if(
                succ_mbb->begin() == succ_mbb->end() ||
                succ_mbb->begin()->getOpcode() != TargetInstrInfo::PHI
            ) {
                // the successor block has no phi-function. Just copy its
                // phi-list to the phi-preference of this basic block:
                unsigned succ_num = succ_mbb->getNumber();
                std::map<unsigned, unsigned> & succ_map = bb_phi_[succ_num];
                copy_map(succ_map, bb2Pref_[mbb_num]);
            } else {
                this->build_preference_list(mbb, succ_mbb);
            }
            copy_map(bb2Pref_[mbb_num], bb_phi_[mbb_num]);
            visited.insert(mbb);
        } else {
            work_list.push(mbb);
        }
    }
}


//===--------------------------------------------------------------------------
// This method returns the operand of the phi function that corresponds to the
// basic block mbb_num.
//===--------------------------------------------------------------------------
unsigned get_phi_column(MachineInstr *phi, int mbb_num) {
    unsigned column = 0;
    for(unsigned u = phi->getNumOperands() - 1; u >= 2; u -= 2) {
        MachineBasicBlock * pred = phi->getOperand(u).getMachineBasicBlock();
        if(pred->getNumber() == mbb_num) {
            column = u;
            break;
        }
    }
    assert(column != 0 && "Basic block not found in phi-function.");
    return column;
}


//===--------------------------------------------------------------------------
// This method fills the vector phi_prm with the parameters in column of phi
// functions in mbb, and it fills phi_res with the results of the same phi-
// functions.
//===--------------------------------------------------------------------------
void fill_vector(
    MachineBasicBlock * mbb,
    unsigned column,
    std::vector<unsigned> & phi_prm,
    std::vector<unsigned> & phi_res
) {
    MachineBasicBlock::const_iterator mbb_i, mbb_e;
    for(mbb_i = mbb->begin(), mbb_e = mbb->end(); mbb_i != mbb_e &&
                         mbb_i->getOpcode() == TargetInstrInfo::PHI; mbb_i++) {
        const MachineInstr * phi = mbb_i;
        unsigned v_prm = phi->getOperand(column).getReg();
        phi_prm.push_back(v_prm);
        unsigned v_res = phi->getOperand(0).getReg();
        phi_res.push_back(v_res);
    }
}


//===--------------------------------------------------------------------------
// This method builds the preference list of a basic block with null
// instructions.
//===--------------------------------------------------------------------------
void Partition_Fer::build_preference_list(
    MachineBasicBlock * mbb,
    MachineBasicBlock * succ_mbb
) {
    unsigned mbb_num = mbb->getNumber();
    unsigned succ_num = succ_mbb->getNumber();

    // copy the bb_phi_ map of succ_mbb into the bb2Pref_ map of mbb:
    assert(bb2Pref_[mbb_num].empty() && "Preference map is not empty");
    copy_map(bb_phi_[succ_num], bb2Pref_[mbb_num]);

    MachineInstr * phi = succ_mbb->begin();
    unsigned column = get_phi_column(phi, mbb_num);
    std::vector<unsigned> phi_prm;
    std::vector<unsigned> phi_res;
    fill_vector(succ_mbb, column-1, phi_prm, phi_res);
//for(unsigned u = 0; u < phi_prm.size(); u++) {
//std::cout << phi_prm[u] << " ";
//}
//std::cout << std::endl;
//for(unsigned u = 0; u < phi_res.size(); u++) {
//std::cout << phi_res[u] << " ";
//}
//std::cout << std::endl;
    std::map<unsigned, unsigned> & mbb_map = this->bb2Pref_[mbb_num];
    for(unsigned u = 0; u < phi_res.size(); u++) {
        unsigned v_res = phi_res[u];
        std::map<unsigned, unsigned>::iterator i_res = mbb_map.find(v_res);
        assert(i_res != mbb_map.end() && "Parameter not found in map.");
        mbb_map[phi_prm[u]] = i_res->second;
        mbb_map.erase(i_res);
    }
}


//===--------------------------------------------------------------------------
// This method is used as an auxiliary function in print_bb_boundaries. This
// method prints a mapping that associate virtual registers to the physical
// locations where the value of the virtual is stored.
//===--------------------------------------------------------------------------
void print_reg_map(
    const std::map<unsigned, unsigned> & m,
    const MRegisterInfo * reg_info
) {
    std::map<unsigned, unsigned>::const_iterator i;
    std::map<unsigned, unsigned>::const_iterator e;
    for(i = m.begin(), e = m.end(); i != e; i++) {
        std::ostringstream str;
        if(i->second != RF_SPILL) {
            str << i->first << ":" << reg_info->getName(i->second);
        } else {
            str << i->first << ":mem";
        }
        std::cout.width(10);
        std::cout << str.str();
    }
    std::cout << std::endl;
}


//===--------------------------------------------------------------------------
// This method prints the mapping that must be joined due to the presence of
// phi-functions in the target control flow graph.
//===--------------------------------------------------------------------------
void Partition_Fer::print_join_points() const {
    for(unsigned u = 0; u < this->bb_phi_.size(); u++) {
        if(this->bb_phi_[u].size() > 0) {
            MachineBasicBlock * mbb = mf_->getBlockNumbered(u);
            MachineBasicBlock::const_pred_iterator pred_aux;
            std::cout << "Predecessors:\n";
            for(pred_aux = mbb->pred_begin();
                                     pred_aux != mbb->pred_end(); pred_aux++) {
                MachineBasicBlock * pred_mbb = * pred_aux;
                unsigned pred_index = pred_mbb->getNumber();
                std::cout << "Block " << pred_index << std::endl;
                print_reg_map(this->bb2Pref_[pred_index], this->ri_);
            }
            std::cout << "Phi-block " << u << std::endl;
            print_reg_map(this->bb_phi_[u], this->ri_);
            std::cout << "=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=\n";
        }
    }
}


//===--------------------------------------------------------------------------
// where: Partition_Fer.cpp
// This method prints the mappings of variables to locations that will be
// stored at the end of each basic block once the function has been traversed.
//===--------------------------------------------------------------------------
void Partition_Fer::print_bb_boundaries() const {
    for(unsigned u = 0; u < this->bb2Pref_.size(); u++) {
        std::cout << "Basic block " << u << "\n";
        const std::map<unsigned, unsigned> & m = this->bb2Pref_[u];
        std::map<unsigned, unsigned>::const_iterator i;
        std::map<unsigned, unsigned>::const_iterator e;
        for(i = m.begin(), e = m.end(); i != e; i++) {
            std::ostringstream str;
            if(i->second != RF_SPILL) {
                str << i->first << ":" << this->ri_->getName(i->second);
            } else {
                str << i->first << ":[MEM]";
            }
            std::cout.width(16);
            std::cout << str.str();
        }
        std::cout << std::endl;
    }
}


//===--------------------------------------------------------------------------
// Add any processing on the machine function here.
//===--------------------------------------------------------------------------
bool Partition_Fer::runOnMachineFunction(MachineFunction & mf) {
    if(SSA_based) {
//NeatPrint_Fer::print_cfg(mf);
        this->initialize_data_structures(mf);
//this->pa_->print_def_list(mf);
        this->scan_intervals();
//this->print_join_points();
    }
    return false;
}
