//===-------------- PhiAnalysis_Fer.cpp - Printing prototype --------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it groups virtual registers that are related by phi functions.
//
// Date: August 1th, 2006
//
//===----------------------------------------------------------------------===//
//
// This class groups together virtual registers that are related by phi
// functions in the control flow graph. Such virtuals describe the same
// variable in the original program. A good register allocator should try to
// assign the same physical register to these virtuals. For instance, if
// v1 := PHI (v2, v3, v3), then {v1, v2, v3} are phi related virtuals.
//
//===----------------------------------------------------------------------===//

#include <set>
#include <vector>
#include <sstream>                               // ostringstream
#include <iostream>

#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/CodeGen/MachineInstr.h"
#include "llvm/CodeGen/MachineFunctionPass.h"     // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"

// In order to declare cl::opt, it is necessary to import this file:
#include "llvm/CodeGen/RegAllocRegistry.h"

#include "llvm/fernando/X86PuzzleBoard_Fer.h"
#include "llvm/fernando/NeatPrint_Fer.h"
#include "llvm/fernando/AllocOptions_Fer.h"	// decl. of global_coalesce
#include "llvm/fernando/SplitPhiGroups_Fer.h"
#include "llvm/fernando/PhiMemCoalescer_Fer.h"

#include "llvm/fernando/PhiAnalysis_Fer.h"      // class definition

using namespace llvm;

namespace llvm {
    bool global_coalesce;
}

char PhiAnalysis_Fer::ID = 0;

namespace {

    cl::opt<bool, true> GlobalCoalesce(
        "gb-clsc",
        cl::desc("Performs global coalescing"),
        cl::location(global_coalesce),
        cl::init(true)
//        cl::init(false)
    );

    RegisterPass<PhiAnalysis_Fer>
    X("PhiAnalysis", "Groups phi-related virtuals - Fernando.", true);
}

//===--------------------------------------------------------------------------
// reporst that this pass does not change anything on the control flow graph.
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::getAnalysisUsage(AnalysisUsage & AU) const {
    AU.addRequired<SplitPhiGroups_Fer>();
    AU.setPreservesAll();
}


//===--------------------------------------------------------------------------
// This method returns true if the given virtual (or, for the same effect,
// any other phi-related virtual) has been mapped to a stack slot.
//===--------------------------------------------------------------------------
bool PhiAnalysis_Fer::has_stack_slot(unsigned v) {
    unsigned base_reg = v - MRegisterInfo::FirstVirtualRegister;
    if(base_reg >= this->v2s.size()) {
        return false;
    } else {
        return this->v2s[base_reg] != -1;
    }
}


//===--------------------------------------------------------------------------
// This method returns true if v_reg is defined in a phi-function.
//===--------------------------------------------------------------------------
bool PhiAnalysis_Fer::is_defined_in_phi(unsigned v_reg) const {
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    assert(base_reg < def_list.size() && "Invalid virtual register.");
    return this->def_list[base_reg] != NULL;
}


//===--------------------------------------------------------------------------
// This method maps the data from the PhiMemCoalesce analysis into this
// analysis:
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::get_new_data(PhiMemCoalescer_Fer * pmc) {
    // Clear the data structures, so that they can receive new data.
    this->v2phi.clear();
    this->phi_related_virtuals.clear();
    v2phi.resize(this->max_virtual + 1, X86PuzzleBoard_Fer::PzConst::UNDEF);
    unsigned u = 0;
    unsigned index = 0;
    while(u < pmc->get_num_eqv_classes()) {
        std::set<unsigned> * aux = pmc->get_virtual_set(u);
        if (aux != NULL) {
            this->phi_related_virtuals.push_back(aux);
            std::set<unsigned>::iterator aux_i = aux->begin();
            std::set<unsigned>::iterator aux_e = aux->end();
            while(aux_i != aux_e) {
                unsigned b_reg = (*aux_i) - MRegisterInfo::FirstVirtualRegister;
                this->v2phi[b_reg] = index;
                aux_i++;
            }
            index++;
        }
        u++;
    }

    // Init the vector with preferences, to help coalesce phi-related virtuals
    // during the register allocation.
    this->pref_list.resize(this->max_virtual + 1);

    // Reset the table of phi-definitions:
    for(unsigned u = 0; u < def_list.size(); u++) {
        this->def_list[u] = NULL;
    }
    // Update the definitions in phi-functions:
    MachineFunction::const_iterator mfi, mfe;
    for (mfi = this->machine_function->begin(),
                      mfe = this->machine_function->end(); mfi != mfe; mfi++) {
        const MachineBasicBlock * mbb = mfi;
        MachineBasicBlock::const_iterator mii = mbb->begin();
        while(mii != mbb->end()) {
            const MachineInstr * mi = mii++;
            if(mi->getOpcode() == TargetInstrInfo::PHI) {
                unsigned v_def = mi->getOperand(0).getReg();
                unsigned b_reg = v_def - MRegisterInfo::FirstVirtualRegister;
                this->def_list[b_reg] = mi;
            } else {
                break;
            }
        }
    }
}


//===--------------------------------------------------------------------------
// This method returns the phi-function that defines the virtual, and
// NULL if the virtual is not define in a phi-function
//===--------------------------------------------------------------------------
const MachineInstr * PhiAnalysis_Fer::get_phi_def(unsigned v_reg) {
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    assert(base_reg < def_list.size() && "Invalid virtual register.");
    return this->def_list[base_reg];
}

//===--------------------------------------------------------------------------
// This method returns the stack slot that has been assigned to virtual
// v, if it has been spilled to memory. All the phi-related virtuals have
// the same stack slot.
//===--------------------------------------------------------------------------
int PhiAnalysis_Fer::get_stack_slot(unsigned v) {
    unsigned base_reg = v - MRegisterInfo::FirstVirtualRegister;
    if(base_reg >= this->v2s.size()) {
        return -1;
    } else {
        return this->v2s[base_reg];
    }
}


//===--------------------------------------------------------------------------
// This method maps the virtual v to the stack slot s. It also has the
// effect of assigning s to all other virtuals that are phi-related to v.
// If v has already been mapped to a stack slot, the function will produce
// an error.
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::map_to_stack_slot(unsigned v, int s) {
    unsigned base_reg = v - MRegisterInfo::FirstVirtualRegister;
    assert(base_reg < this->v2s.size() &&
                      "Index of register out of bounds of stack slot vector.");
    assert(s >= 0 && "Invalid stack slot.");
    assert(this->v2s[base_reg] == -1 &&
                     "Error: virtual has already been assigned a stack slot.");
    this->v2s[base_reg] = s;
    // If the virtual is related to other variables, map them all to the same
    // stack slot:
    if(this->v2phi[base_reg] != X86PuzzleBoard_Fer::PzConst::UNDEF) {
        std::set<unsigned> * phi_set = phi_related_virtuals[v2phi[base_reg]];
        std::set<unsigned>::iterator pos;
        for (pos = phi_set->begin(); pos != phi_set->end(); ++pos) {
            unsigned r_src = * pos;
            unsigned b_reg = r_src - MRegisterInfo::FirstVirtualRegister;
            this->v2s[b_reg] = s;
        }
    }
}


//===--------------------------------------------------------------------------
// This method marks the virtuals v_def and v_reg as being related. This
// method must be called in such a way that v_def is the register defined in
// the phi-function, and v_reg is the register used in the same phi-function.
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::add_to_related_set(unsigned v_def, unsigned v_reg) {
    unsigned bv = v_reg - MRegisterInfo::FirstVirtualRegister;
    unsigned bd = v_def - MRegisterInfo::FirstVirtualRegister;
    if(this->v2phi[bd] == X86PuzzleBoard_Fer::PzConst::UNDEF) {
        assert(v_def == v_reg && "Lost entry in phi-collector.");
        std::set<unsigned> * set = new std::set<unsigned>();
        this->phi_related_virtuals.push_back(set);
        unsigned p_size = this->phi_related_virtuals.size();
        this->v2phi[bd] = p_size - 1;
        set->insert(v_def);
    } else {
        if(this->v2phi[bv] == X86PuzzleBoard_Fer::PzConst::UNDEF) {
            std::set<unsigned> * set = phi_related_virtuals[v2phi[bd]];
            set->insert(v_reg);
            this->v2phi[bv] = this->v2phi[bd];
        } else {
            // merge the sets
            std::set<unsigned> * set_old = phi_related_virtuals[v2phi[bv]];
            std::set<unsigned> * set_new = phi_related_virtuals[v2phi[bd]];
            std::set<unsigned>::iterator pos;
            for (pos = set_new->begin(); pos != set_new->end(); ++pos) {
                unsigned r_src = * pos;
                unsigned b_aux = r_src - MRegisterInfo::FirstVirtualRegister;
                this->v2phi[b_aux] = this->v2phi[bv];
                set_old->insert(r_src);
            }
            this->v2phi[bd] = this->v2phi[bv];
        }
    }
}


//===--------------------------------------------------------------------------
// This function merges the phi-group of v1 with the phi-group of v2, so that
// each virtual in any of these groups will be phi-related.
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::merge(unsigned v1, unsigned v2) {
    unsigned b1 = v1 - MRegisterInfo::FirstVirtualRegister;
    unsigned b2 = v2 - MRegisterInfo::FirstVirtualRegister;
    assert(
        (this->v2phi[b1] != X86PuzzleBoard_Fer::PzConst::UNDEF ||
         this->v2phi[b2] != X86PuzzleBoard_Fer::PzConst::UNDEF) &&
         "At least one of the virtuals must be phi-related."
    );
    if(this->v2phi[b1] == X86PuzzleBoard_Fer::PzConst::UNDEF) {
        std::set<unsigned> * set = phi_related_virtuals[v2phi[b2]];
        set->insert(v1);
        this->v2phi[b1] = this->v2phi[b2];
    } else if (this->v2phi[b2] == X86PuzzleBoard_Fer::PzConst::UNDEF) {
        std::set<unsigned> * set = phi_related_virtuals[v2phi[b1]];
        set->insert(v2);
        this->v2phi[b2] = this->v2phi[b1];
    } else {
        // merge the sets
        std::set<unsigned> * set1 = phi_related_virtuals[v2phi[b1]];
        std::set<unsigned> * set2 = phi_related_virtuals[v2phi[b2]];
        std::set<unsigned>::iterator pos;
        for (pos = set2->begin(); pos != set2->end(); ++pos) {
            unsigned r_src = * pos;
            unsigned b_aux = r_src - MRegisterInfo::FirstVirtualRegister;
            this->v2phi[b_aux] = this->v2phi[b1];
            set1->insert(r_src);
        }
    }
}


//===--------------------------------------------------------------------------
// Returns true if the virtual register is phi-related, and has preference for
// some particular physical register.
//===--------------------------------------------------------------------------
bool PhiAnalysis_Fer::has_preference(unsigned v_reg) const {
    if(global_coalesce) {
        if( ! is_phi_related(v_reg) ) {
            return false;
        }
        unsigned b_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
        const Preference & pref = this->pref_list[this->v2phi[b_reg]];
        return pref.preference != 0;
    } else {
        return false;
    }
}


//===--------------------------------------------------------------------------
// This method returns the physical register that should be assigned to v_reg
// if possible.
//===--------------------------------------------------------------------------
unsigned PhiAnalysis_Fer::get_preference(unsigned v_reg) const {
    if(global_coalesce && is_phi_related(v_reg)) {
        unsigned b_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
        const Preference & pref = this->pref_list[this->v2phi[b_reg]];
        return pref.preference;
    } else {
        return 0;
    }
}


//===--------------------------------------------------------------------------
// This method points that the physical p_reg has been assigned to the virtual
// v_reg. This contributes to set p_reg as the prefered register of any virtual
// that is phi-related to v_reg.
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::set_preference(unsigned v_reg, unsigned p_reg) {
    if( global_coalesce && is_phi_related(v_reg) ) {
        unsigned b_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
        Preference & pref = this->pref_list[this->v2phi[b_reg]];
        pref.add(p_reg);
    }
}


//===--------------------------------------------------------------------------
// This method returns the set of virtuals that are phi-related to v_reg.
//===--------------------------------------------------------------------------
std::set<unsigned> & PhiAnalysis_Fer::get_phi_group(unsigned v_reg) {
    unsigned b_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    assert(this->v2phi[b_reg] != X86PuzzleBoard_Fer::PzConst::UNDEF &&
                    "Attempt to grab the group of a non-phi-related virtual.");
    return * this->phi_related_virtuals[this->v2phi[b_reg]];
}


//===--------------------------------------------------------------------------
// This method reads a phi-function, and marks all the parameters of that
// phi-function, as well as its result, as phi-related. The relation among
// the phi-related variables partition the program into equivalence classes.
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::visit_phi_function(MachineInstr * phi) {
    assert(phi->getOpcode() == TargetInstrInfo::PHI &&
                             "Only phi instructions define related virtuals.");
    unsigned v_def = 0;

    // This set stores the already seen registers, so that they don't get
    // processed twice.
    std::set<unsigned> seen_regs;

    for(unsigned i = 0; i != phi->getNumOperands(); i++) {
        MachineOperand & mo = phi->getOperand(i);
        if (mo.isRegister() && mo.getReg()) {
            unsigned v_reg = mo.getReg() - MRegisterInfo::FirstVirtualRegister;
            if(v_reg >= this->use_list.size()) {
                this->use_list.resize(2 * (1 + v_reg));
                this->def_list.resize(2 * (1 + v_reg));
                v2phi.resize(2*(1+v_reg), X86PuzzleBoard_Fer::PzConst::UNDEF);
            }
            if(v_reg > this->max_virtual) {
                this->max_virtual = v_reg;
            }
            if(mo.isDef()) {
                v_def = mo.getReg();
                add_to_related_set(v_def, v_def);
                this->def_list[v_reg] = phi;
            } else {
                if(seen_regs.find(v_reg) != seen_regs.end()) {
                    continue;
                } else {
                    assert(v_def != 0 &&
                                     "Phi function does not define register?");
                    add_to_related_set(v_def, mo.getReg());
                    seen_regs.insert(v_reg);
                    this->use_list[v_reg].push_back(phi);
                }
            }
        }
    }
}


//===--------------------------------------------------------------------------
// This method returns true if v_reg is part of a phi function somewhere
// in the control flow graph, and false otherwise. 
//===--------------------------------------------------------------------------
bool PhiAnalysis_Fer::is_phi_related(unsigned v_reg) const {
    assert(MRegisterInfo::isVirtualRegister(v_reg) &&
                       "Only virtual registers can be part of phi-functions.");
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    if(base_reg > this->max_virtual) {
        return false;
    }
    return (use_list[base_reg].size() > 0) || (def_list[base_reg] != NULL);
}


//===--------------------------------------------------------------------------
// This method removes the phi-function associated to old_v, which must be
// unique, and makes it associated to new_v.
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::update_use_list(unsigned old_v, unsigned new_v) {
    assert(MRegisterInfo::isVirtualRegister(old_v) &&
                       "Register given as old virtual is not really virtual.");
    assert(MRegisterInfo::isVirtualRegister(new_v) &&
                       "Register given as new virtual is not really virtual.");
    unsigned old_b = old_v - MRegisterInfo::FirstVirtualRegister;
    assert(this->use_list[old_b].size() == 1 &&
                          "Old virtual is not part of only one phi-function.");
    MachineInstr * phi = this->use_list[old_b][0];
    this->use_list[old_b].clear();
    unsigned new_b = new_v - MRegisterInfo::FirstVirtualRegister;
    this->use_list[new_b].push_back(phi);
}


//===--------------------------------------------------------------------------
// This method removes the phi-function associated to old_v, which must be
// unique, and makes it associated to new_v. The only difference of this
// method to 'update_use_list' is that in this case old_v is expected to be
// defined by a phi-function.
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::update_def_list(unsigned old_v, unsigned new_v) {
    assert(MRegisterInfo::isVirtualRegister(old_v) &&
                       "Register given as old virtual is not really virtual.");
    assert(MRegisterInfo::isVirtualRegister(new_v) &&
                       "Register given as new virtual is not really virtual.");
    unsigned old_b = old_v - MRegisterInfo::FirstVirtualRegister;
    assert(this->def_list[old_b] != NULL &&
                              "Old virtual is not defined by a phi-function.");
    const MachineInstr * phi = this->def_list[old_b];
    this->def_list[old_b] = NULL;
    unsigned new_b = new_v - MRegisterInfo::FirstVirtualRegister;
    this->def_list[new_b] = phi;
}


//===--------------------------------------------------------------------------
// Prints the set of phi-related virtuals.
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::dump() {

    unsigned base = MRegisterInfo::FirstVirtualRegister;

    std::cerr << "Def list:\n";
    for(unsigned u = 0; u <= this->max_virtual; u++) {
        if(this->def_list[u] != NULL) {
            std::cerr << "t" << u << "(" <<
                               (u + base) << ") is defined in phi-function.\n";
        }
    }

    std::cerr << "Use list:\n";
    for(unsigned u = 0; u <= this->max_virtual; u++) {
        std::cerr << "t" << u << ") " << this->use_list[u].size() << "\n";
    }

    std::cerr << "Definition sites:\n";
    for(unsigned u = 0; u < this->def_sites.size(); u++) {
        if(this->def_sites[u].size() > 0) {
            std::set<unsigned> & s = def_sites[u];
            std::set<unsigned>::iterator pos;
            std::cerr << "t" << u << ")\n";
            for (pos = s.begin(); pos != s.end(); ++pos) {
                unsigned r_src = * pos;
                unsigned b_reg = r_src - MRegisterInfo::FirstVirtualRegister;
                std::cerr << "t" << b_reg << ", ";
            }
            std::cerr << "\n";
        }
    }

    this->print_phi_groups(true);

    std::cerr << "Phi Related Virtuals:\n";
    for(unsigned u = 0; u <= this->max_virtual; u++) {
        std::cerr << u << " -> " << this->v2phi[u] << "\n";
    }
}


//===--------------------------------------------------------------------------
// This method returns the number of phi-related variables in the source
// program. Any variable that is part of a phi-function is phi-related.
//===--------------------------------------------------------------------------
unsigned PhiAnalysis_Fer::get_number_phi_related_variables() const {
    unsigned num_phi_related = 0;
    for(unsigned u = 0; u < this->phi_related_virtuals.size(); u++) {
        const std::set<unsigned> * s = this->phi_related_virtuals[u];
        num_phi_related += s->size();
    }
    return num_phi_related;
}


//===--------------------------------------------------------------------------
// This method returns the number of phi-groups in the source program.
// Phi-groups constitute equivalence classes. This method only returns the
// number of non-singleton groups.
//===--------------------------------------------------------------------------
unsigned PhiAnalysis_Fer::get_number_phi_groups() const {
    return this->phi_related_virtuals.size();
}


//===--------------------------------------------------------------------------
// This method prints the groups of phi-related virtuals. If "sub_base" is set
// to true, the method prints virtuals as in t0, t1, etc. Otherwise, it prints
// the virtuals as in 1024, 1025, etc.
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::print_phi_groups(bool sub_base) const {
    std::cerr << "Phi Groups:\n";
    for(unsigned u = 0; u < this->phi_related_virtuals.size(); u++) {
        const std::set<unsigned> * s = this->phi_related_virtuals[u];
        std::set<unsigned>::iterator pos;
        std::cerr << "Group " << u << ": (";
        for (pos = s->begin(); pos != s->end(); ++pos) {
            unsigned r_src = * pos;
            if(sub_base) {
                r_src = r_src - MRegisterInfo::FirstVirtualRegister;
                std::cerr << "t" << r_src << "  ";
            } else {
                std::cerr << r_src << "  ";
            }
        }
        std::cerr << ")\n";
    }
}


//===--------------------------------------------------------------------------
// This method allocates memory for the data structures.
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::initialize_data_structures(MachineFunction & mf) {
    this->max_virtual = 0;
    this->machine_function = & mf;

    // Finds the number of physical registers:
    const TargetMachine & target_machine = mf.getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    this->num_p_regs = reg_info->getNumRegs();

    // Finds the number of virtual registers:
    SSARegMap * srm = mf.getSSARegMap();
    unsigned last_v_reg = srm->getLastVirtReg() + 1;
    this->max_virtual = last_v_reg - MRegisterInfo::FirstVirtualRegister;
    // This size must generally be enough for all the virtuals plus the regs
    // that can be created due to spills.
    this->use_list.resize(2 * (this->max_virtual + 1));
    this->def_list.resize(2 * (this->max_virtual + 1));
    this->v2phi.resize(2*(this->max_virtual + 1),
                                           X86PuzzleBoard_Fer::PzConst::UNDEF);

    // Init the vector with preferences, to help coalesce phi-related virtuals
    // during the register allocation.
    if(global_coalesce) {
        this->pref_list.resize(this->max_virtual + 1);
    }
}


//===--------------------------------------------------------------------------
// This method prints the assignment of virtuals to stack slots.
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::print_v2s_mapping() const {
    std::cout << "Stack configuration:\n";
    for(unsigned u = 0; u < this->v2s.size(); u++) {
        if(this->v2s[u] != -1) {
            unsigned v_reg = u + MRegisterInfo::FirstVirtualRegister;
            std::cout << v_reg << " => [" << this->v2s[u] << "]\n";
        }
    }
}


//===--------------------------------------------------------------------------
// Prints all phi-functions that have been found during this analysis.
//===--------------------------------------------------------------------------
void PhiAnalysis_Fer::print_def_list(MachineFunction & mf) const {
    for(unsigned u = 0; u < this->def_list.size(); u++) {
        if(this->def_list[u]) {
            unsigned v_reg = u + MRegisterInfo::FirstVirtualRegister;
            std::cout << v_reg << ": " <<
               NeatPrint_Fer::mi2string(* this->def_list[u], mf) << std::endl;
        }
    }
}


//===--------------------------------------------------------------------------
// runOnMachineFunction - Groups together the phi related virtuals.
//===--------------------------------------------------------------------------
bool PhiAnalysis_Fer::runOnMachineFunction(MachineFunction & mf) {
    this->initialize_data_structures(mf);
    for (MachineFunction::iterator mbbi = mf.begin(), mbbe = mf.end();
                                                        mbbi != mbbe; ++mbbi) {
        MachineBasicBlock * mbb = mbbi;
        MachineBasicBlock::iterator mii = mbb->begin();
        while (mii != mbb->end()) {
            MachineInstr * mi = mii;
            if(mi->getOpcode() == TargetInstrInfo::PHI) {
                visit_phi_function(mi);
            } else {
                // Always safe, because phi functions are always the first
                // instructions in the basic block.
                break;
            }
            mii++;
        }
    }

    // Reserve space for the mappings of virtuals to stack slots:
    this->v2s.resize(this->v2phi.size(), -1);
//this->print_def_list(mf);
//std::cout << "\n\n*******************************************\n\n";
    return false;
}
