//===------------- PhiMemCoalescer_Fer.cpp - Printing prototype -----------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This pass attempts to group together virtuals whose live range does not
// overlap, so that they can be inserted in the same memory spot.
//
// Date: March 3rd, 2007
//
//===----------------------------------------------------------------------===//
//
// The combinantion of PhiSplit and PhiAnalysis passes group together all the
// virtuals that are part of the same phi-function. This is fine for
// implementing the phi-deconstruction algorithm, but it is to conservative for
// assigning the same memory address to phi-related virtuals. This pass
// attempts to increase these groups by checking if the live ranges of virtuals
// do not overlap, and then assigning the virtuals to the same group.
//
//===----------------------------------------------------------------------===//

#define DEBUG_TYPE "phiCoalescer"

#include <iostream>
#include <fstream>
#include <sstream>
#include <string>

#include "llvm/CodeGen/LiveIntervalAnalysis.h"

#include "llvm/CodeGen/MachineFunctionPass.h"       // MachineBasicBlock
#include "llvm/CodeGen/LiveVariables.h"

#include "llvm/Instructions.h"                      // MachineOperand, etc

#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"

#include "llvm/Support/Debug.h"

#include "llvm/ADT/Statistic.h"

#include "llvm/fernando/NeatPrint_Fer.h"
#include "llvm/fernando/DefUseSites_Fer.h"
#include "llvm/fernando/PhiAnalysis_Fer.h"
#include "llvm/fernando/EqvPhiClass_Fer.h" 
#include "llvm/fernando/AllocOptions_Fer.h"
#include "llvm/fernando/SplitPhiGroups_Fer.h"
#include "llvm/fernando/RegisterCoalescer_Fer.h" 

#include "llvm/fernando/PhiMemCoalescer_Fer.h"      // class definition

using namespace llvm;

STATISTIC(delMoves, "Moves removed by phi-mem-coalescer.");
STATISTIC(phiElBef, "Number of phi-related variables before coalescing.");
STATISTIC(phiElAft, "Number of phi-related variables after coalescing.");
STATISTIC(phiGpBef, "Number of phi-groups before coalescing.");
STATISTIC(phiGpAft, "Number of phi-groups after coalescing.");

#define UNVISITED 2147483647

char PhiMemCoalescer_Fer::ID = 0;

namespace {
    RegisterPass<PhiMemCoalescer_Fer>
    X("phi_coal", "Coalescing of phi-related variables - Fernando", true);
}

//===--------------------------------------------------------------------------
// This method defines which passes are required by this analysis.
//===--------------------------------------------------------------------------
void PhiMemCoalescer_Fer::getAnalysisUsage(AnalysisUsage & AU) const {
    AU.addRequired<LiveVariables>();
    AU.addRequired<SplitPhiGroups_Fer>();
    AU.addRequired<PhiAnalysis_Fer>();
    AU.addRequired<LiveIntervals>();
    AU.setPreservesAll();
}

//===--------------------------------------------------------------------------
// This method initializes the data structures used during register allocation.
// Among the variables initilized, are ig, split_phi and def_use_sites.
//===--------------------------------------------------------------------------
void PhiMemCoalescer_Fer::initialize_data_structures(MachineFunction & mf) {
    // Get pointers to the required analysis:
    this->phi_analysis = & getAnalysis<PhiAnalysis_Fer>();
    this->split_phi = & getAnalysis<SplitPhiGroups_Fer>();
    this->interval_analysis = & getAnalysis<LiveIntervals>();

    // Initialize the phi-equivalente classes with the results of the
    // interval analysis:
    // Reserve space for all the virtual registers:
    this->v2pec.resize(this->phi_analysis->get_num_virtuals()+1, UNVISITED);
    // Handle each virtual register:
    for(unsigned u = 0; u < this->v2pec.size(); u++) {
        unsigned vr = u + MRegisterInfo::FirstVirtualRegister;
        // If the virtual has not yet been assigned a equivalence class:
        if((this->v2pec[u] == UNVISITED) && phi_analysis->is_phi_related(vr)) {
            // Create the equivalence class:
            EqvPhiClass_Fer * epc = new EqvPhiClass_Fer();
            unsigned index = this->phi_related_virtuals.size();
            phi_related_virtuals.push_back(epc);
            // Get all the virtuals related to vr:
            std::set<unsigned> & s = this->phi_analysis->get_phi_group(vr);
            std::set<unsigned>::iterator pos;
            for (pos = s.begin(); pos != s.end(); ++pos) {
                // Add the related virtual to the equivalence class:
                unsigned r_src = * pos;
                epc->add(this->interval_analysis->getInterval(r_src));
                unsigned b_aux = r_src - MRegisterInfo::FirstVirtualRegister;
                this->v2pec[b_aux] = index;
            }
        }
    }
    LiveVariables * lv = & getAnalysis<LiveVariables>();
    this->rc_ = new RegisterCoalescer_Fer(mf, interval_analysis, lv);
}


//===--------------------------------------------------------------------------
// This method return true if any of the virtuals in these sets are adjacent
// in the interference graph.
//===--------------------------------------------------------------------------
bool PhiMemCoalescer_Fer::overlap(unsigned use_reg, unsigned def_reg) {
 
    unsigned b_def = def_reg - MRegisterInfo::FirstVirtualRegister;
    unsigned b_use = use_reg - MRegisterInfo::FirstVirtualRegister;

    if(this->v2pec[b_def] == UNVISITED) {
        // It is the case:
        // b_use := phi(...);
        // b_def := b_use;
        assert(this->v2pec[b_use] != UNVISITED && "Use not phi-related.");
        EqvPhiClass_Fer * epc_use = phi_related_virtuals[this->v2pec[b_use]];
        LiveInterval & li = this->interval_analysis->getInterval(def_reg);
        return epc_use->overlap(li);
    } else {
        // It is a case:
        //          b_def := b_use;                ... -> go to MBB
        // MBB)         a := phi(b_def, ...);
        EqvPhiClass_Fer * epc_def = phi_related_virtuals[this->v2pec[b_def]];
 
        if(this->v2pec[b_use] == UNVISITED) {
            LiveInterval & li = this->interval_analysis->getInterval(use_reg);
            return epc_def->overlap(li);
        } else {
            EqvPhiClass_Fer * epc_use = phi_related_virtuals[v2pec[b_use]];
            return epc_def->overlap(* epc_use);
        }
    }

}


//===--------------------------------------------------------------------------
// Return the set of phi-related virtuals for a given virtual
//===--------------------------------------------------------------------------
std::set<unsigned> * PhiMemCoalescer_Fer::get_virtual_set(unsigned index) {
    EqvPhiClass_Fer * pec = this->phi_related_virtuals[index];
    std::set<unsigned> * set = pec->get_virtuals();
    unsigned b_reg = *(set->begin()) - MRegisterInfo::FirstVirtualRegister;
    if(this->v2pec[b_reg] == index) {
        return set;
    } else {
        return NULL;
    }
}


//===--------------------------------------------------------------------------
// This method merges the phi-equivalence classes of v1 and v2.
//===--------------------------------------------------------------------------
void PhiMemCoalescer_Fer::merge(unsigned v1, unsigned v2) {
    unsigned b1 = v1 - MRegisterInfo::FirstVirtualRegister;
    unsigned b2 = v2 - MRegisterInfo::FirstVirtualRegister;
    assert( (this->v2pec[b1] != UNVISITED || this->v2pec[b2] != UNVISITED) &&
                          "At least one of the virtuals must be phi-related.");
    // If v1 was not yet known to be phi-related, it does not have a phi
    // equivalence class. In this case, just add v1 to the equiv class of v2.
    if(this->v2pec[b1] == UNVISITED) {
        EqvPhiClass_Fer * epc = phi_related_virtuals[v2pec[b2]];
        epc->add(this->interval_analysis->getInterval(v1));
        this->v2pec[b1] = this->v2pec[b2];
    } else if (this->v2pec[b2] == UNVISITED) {
        // Same as before, but in this case v2 does not have yet phi-eqv-class.
        EqvPhiClass_Fer * epc = phi_related_virtuals[v2pec[b1]];
        epc->add(this->interval_analysis->getInterval(v2));
        this->v2pec[b2] = this->v2pec[b1];
    } else {
        // In this case, both, v1 and v2 are known te be phi-related; thus,
        // both have equivalence classes. Each equivalence class is a set of
        // virtuals, and we must merge the eqv-classes of v1 and v2:
        EqvPhiClass_Fer * epc1 = phi_related_virtuals[v2pec[b1]];
        EqvPhiClass_Fer * epc2 = phi_related_virtuals[v2pec[b2]];
        epc1->merge(*epc2);
        std::set<unsigned> * v_set = epc2->get_virtuals();
        std::set<unsigned>::iterator pos;
        unsigned epc1_index = this->v2pec[b1];
        for (pos = v_set->begin(); pos != v_set->end(); ++pos) {
            unsigned r_src = * pos;
            unsigned b_aux = r_src - MRegisterInfo::FirstVirtualRegister;
            this->v2pec[b_aux] = epc1_index;
        }
    }
}


//===--------------------------------------------------------------------------
// This method prints all the equivalence classes and virtuals in the source
// program.
//===--------------------------------------------------------------------------
void PhiMemCoalescer_Fer::dump() const {
    std::cerr << "Phi Eqv classes:\n";
    for(unsigned u = 0; u < this->phi_related_virtuals.size(); u++) {
        this->phi_related_virtuals[u]->dump();
    }

    std::cerr << "Virtuals:\n";
    for(unsigned u = 0; u < this->v2pec.size(); u++) {
        unsigned v_reg = u + MRegisterInfo::FirstVirtualRegister;
        std::cerr << u << ":" << v_reg << " = " << this->v2pec[u] << " , ";
    }

    std::cerr << "\n";
}


//===--------------------------------------------------------------------------
// This method removes the instruction mi from the basic block, and updates
// the liveness analysis information and the def/use sites. It must be used
// when removing a move instruction created to break the live range of a use
// in a phi-function. E.g:
// a = phi(...)
// b = a;
//===--------------------------------------------------------------------------
void PhiMemCoalescer_Fer::remove_instruction_rst(
    MachineInstr * mi,
    unsigned def_reg,
    unsigned use_reg
) {
    // Update the interval analysis:
    delMoves++;
    bool result = this->rc_->JoinCopy(mi, use_reg, def_reg, false);
    assert(result && "Intervals cannot be joined");
}


//===--------------------------------------------------------------------------
// This method removes the instruction mi from the basic block, and updates
// the liveness analysis information and the def/use sites. It must be used
// when removing a move instruction created to break the live range of a use
// in a phi-function. E.g:
// a := b
// c := phi(a, ...)
//===--------------------------------------------------------------------------
void PhiMemCoalescer_Fer::remove_instruction_prm(
    MachineInstr * mi,
    unsigned def_reg,
    unsigned use_reg
) {
    // Update the interval analysis:
    bool result = this->rc_->JoinCopy(mi, use_reg, def_reg, false);
    delMoves++;
    assert(result && "Intervals cannot be joined");
}


//===--------------------------------------------------------------------------
// Release memory back to the operating system.
//===--------------------------------------------------------------------------
void PhiMemCoalescer_Fer::releaseMemory() {
    this->phi_related_virtuals.clear();
    this->v2pec.clear();
    delete this->rc_;
}


// Implemented by Fernando on July 8th.
// This method sets the SSA-register allocation mode if the name of the
// function that is being compiled can be found in a IGNORE.txt file.
bool read_IGNORE(MachineFunction & mf) {
    std::ifstream file("IGNORE.txt");
    if(file.is_open()) {
        const Function * f = mf.getFunction();
        std::ostringstream f_str;
        f_str << f->getName();
        std::string f_name = f_str.str();
        while(!file.eof()) {
            std::string line;
            std::getline(file, line);
            if(line == f_name) {
                return true;
            }
        }
        file.close();
    }
    return false;
}

//===--------------------------------------------------------------------------
// runOnMachineFunction - Groups together the phi related virtuals.
//===--------------------------------------------------------------------------
bool PhiMemCoalescer_Fer::runOnMachineFunction(MachineFunction & mf) {

//    if(!SSA_based) {
//        return false;
//    }

    const TargetMachine & target_machine = mf.getTarget();
    const TargetInstrInfo & tii = * (target_machine.getInstrInfo());
    this->initialize_data_structures(mf);

//Only to disable it temporarily:
//return false;

    if(read_IGNORE(mf)) {
        std::cout << "Skipping " << mf.getFunction()->getName() << std::endl;
        return true;
    }

//this->interval_analysis->dump();
//this->phi_analysis->dump();
//this->phi_analysis->print_phi_groups(false);
    phiElBef += this->phi_analysis->get_number_phi_related_variables();
    phiGpBef += this->phi_analysis->get_number_phi_groups();
//NeatPrint_Fer::print_cfg(mf);

    // First, expand the equivalence classes of the virtuals used to
    // partition the parameters of the phi-function.
    const std::vector<MachineInstr *> & split_points_prm =
                                       this->split_phi->get_split_points_prm();
    for(unsigned u = 0; u < split_points_prm.size(); u++) {
        MachineInstr * mi = split_points_prm[u];
        unsigned use_reg, def_reg;
        if(tii.isMoveInstr(*mi, use_reg, def_reg)) {
            if( ! this->overlap(use_reg, def_reg) ) {
                this->merge(use_reg, def_reg);
                remove_instruction_prm(mi, def_reg, use_reg);
            }
        }
    }

    // Second, expand the eqv-classes of the virtuals used to partition the
    // live ranges of results of phi-functions.
    const std::vector<MachineInstr *> & split_points_rst =
                                   this->split_phi->get_split_points_rst();
    for(unsigned u = 0; u < split_points_rst.size(); u++) {
        MachineInstr * mi = split_points_rst[u];
        unsigned use_reg, def_reg;
        if(tii.isMoveInstr(*mi, use_reg, def_reg)) {
            if( ! this->overlap(use_reg, def_reg) ) {
                this->merge(use_reg, def_reg);
                remove_instruction_rst(mi, def_reg, use_reg);
            }
        }
    }

    this->rc_->rewrite_instructions();

    this->phi_analysis->get_new_data(this);

//NeatPrint_Fer::print_cfg(mf);

/*
    if(!read_IGNORE(mf)) {
        NeatPrint_Fer::print_cfg(mf);
    }
*/
//this->interval_analysis->dump();
//this->phi_analysis->print_phi_groups(false);
    phiElAft += this->phi_analysis->get_number_phi_related_variables();
    phiGpAft += this->phi_analysis->get_number_phi_groups();
//this->phi_analysis->dump();

    return false;
}
