//===-------------- PhiP3_Fer.h - Clique substitution of P_3 --------------===//
//
//            Register Allocation via Coloring of Chordal Graphs 
//
// This class represents a clique substitution of P_3
//
// Date: Oct 8th, 2007
//
//===----------------------------------------------------------------------===//
//
// This class represents the clique substitution of P3 that forms the head of
// a basic block that contains phi-functions. It is interestly to treat this
// graph differently than the other clique substitutions because it allows to
// design better spilling heuristics.
//
//===----------------------------------------------------------------------===//

#define DEBUG_TYPE "phi3graph"

#include <algorithm>                             // std::sort(...)
#include <iostream>                              // cerr
#include <sstream>                               // ostringstream (RegClique)
#include <iomanip>                               // setw
#include <vector>
#include <map>

#include "llvm/Instructions.h"                   // MachineOperand, etc
#include "llvm/CodeGen/Passes.h"                 // RegAllocChord_Fer
#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/CodeGen/LiveInterval.h"
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/CodeGen/MachineFunctionPass.h"    // MachineBasicBlock
#include "llvm/CodeGen/LiveIntervalAnalysis.h"

#include "llvm/Support/Debug.h"

#include "llvm/ADT/SmallVector.h"
#include "llvm/ADT/STLExtras.h"
#include "llvm/ADT/Statistic.h"
#include "llvm/ADT/DepthFirstIterator.h"         // df_ext_iterator

#include "llvm/fernando/Regs_Fer.h"
#include "llvm/fernando/Partition_Fer.h"
#include "llvm/fernando/RA_Result_Fer.h"
#include "llvm/fernando/RegMapping_Fer.h"
#include "llvm/fernando/DefUseSites_Fer.h"
#include "llvm/fernando/PhiAnalysis_Fer.h"
#include "llvm/fernando/X86PuzzleBoard_Fer.h"

#include "llvm/fernando/PhiP3_Fer.h"

using namespace llvm;

STATISTIC(zeroIter, "Puzzles solved without iterations.");
STATISTIC(maxIterations, "Maximum number of iterations due to spilling.");

//===--------------------------------------------------------------------------
// Constructor method.
//===--------------------------------------------------------------------------
PhiP3_Fer::PhiP3_Fer(
    MachineInstr * phi,
    MachineFunction & mf,
    Partition_Fer * partition,
    RegMapping_Fer *rm,
    RA_Result_Fer *rs,
    PhiAnalysis_Fer * pa,
    DefUseSites_Fer * du
) : 
    mf_(& mf),
    partition_(partition),
    rm_(rm),
    rs_(rs),
    pa_(pa),
    du_(du),
    numIter_(0)
{
    assert(phi->getOpcode() == TargetInstrInfo::PHI &&
                                              "Non-phi-function in PhiP3_Fer");
    for(unsigned u = phi->getNumOperands() - 1; u >= 2; u -= 2) {
        MachineBasicBlock & pred = * phi->getOperand(u).getMachineBasicBlock();
        this->bb_.push_back(pred.getNumber());
    }

    // Discover the number of physical registers that must be handled:
    const TargetMachine & target_machine = mf.getTarget(); 
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    unsigned num_p_regs = reg_info->getNumRegs();
    this->p_regs_X_.resize(num_p_regs, RF_UNDEF);
    this->p_regs_Z_.resize(num_p_regs, RF_UNDEF);
}


//===--------------------------------------------------------------------------
// Add 'v_reg' as a node of component Z
//===--------------------------------------------------------------------------
void PhiP3_Fer::addY(unsigned v_reg) {
//std::cout << "Add Y " << v_reg << "\n";
    if(this->phi_virtuals_.find(v_reg) == this->phi_virtuals_.end()) {
        assert(MRegisterInfo::isVirtualRegister(v_reg) &&
                                    "Register in phi-matrix must be virtual.");
        unsigned p_reg = this->rm_->get_physical_location(v_reg);
        if(MRegisterInfo::isPhysicalRegister(p_reg)) {
            assert(p_regs_X_[p_reg] == RF_UNDEF && "Inconsistent coloring");
            assert(p_regs_Z_[p_reg] == RF_UNDEF && "Inconsistent coloring");
            this->p_regs_X_[p_reg] = v_reg;
            this->p_regs_Z_[p_reg] = v_reg;
        }
        this->Y_.push_back(v_reg);
    }
}


//===--------------------------------------------------------------------------
// Adds the phi function to this graph. The virtual defined by the phi-function
// will contribute to the Z group of this graph.
//===--------------------------------------------------------------------------
void PhiP3_Fer::add_phi(MachineInstr * phi) {
    unsigned v_reg = phi->getOperand(0).getReg();
    this->addZ(v_reg);
    this->v2phi_[v_reg] = phi;
}


//===--------------------------------------------------------------------------
// Add 'v_reg' as a node of component Z
//===--------------------------------------------------------------------------
void PhiP3_Fer::addZ(unsigned v_reg) {
//std::cout << "Add Z " << v_reg << "\n";
    assert(MRegisterInfo::isVirtualRegister(v_reg) &&
                                    "Register in phi-matrix must be virtual.");
    unsigned p_reg = this->rm_->get_physical_location(v_reg);
    assert(p_reg == RF_UNDEF && "Phi def virtual has already been allocated");
    this->Z_.push_back(v_reg);
    this->phi_virtuals_.insert(v_reg);
}


//===--------------------------------------------------------------------------
// This method prints a vector of unsigned elements.
//===--------------------------------------------------------------------------
void print_vec(
    const std::vector<unsigned> & v,
    const MRegisterInfo * reg_info
) {
    for(unsigned u = 0; u < v.size(); u++) {
        std::ostringstream str;
        str << v[u];
        std::cout.width(8);
        std::cout << str.str();
    }
}


//===--------------------------------------------------------------------------
// This method prints the location of each variable that is alive at the
// beginning of the program point that this graph represents.
//===--------------------------------------------------------------------------
void PhiP3_Fer::print_in_set() const {
    const TargetMachine & target_machine = this->mf_->getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();

    // Print the predecessor basic blocks that have been already visited:
    for(unsigned u = 0; u < this->bb_.size(); u++) {
        const std::map<unsigned, unsigned> & bb_map =
                                  this->partition_->get_tail_map(this->bb_[u]);
        if(bb_map.size()) {
            std::cout << "Source block: " << this->bb_[u] << std::endl;
            std::map<unsigned, unsigned>::const_iterator i;
            std::map<unsigned, unsigned>::const_iterator e;
            for(i = bb_map.begin(), e = bb_map.end(); i != e; i++) {
                std::ostringstream str;
                if(i->second != RF_SPILL) {
                    str << i->first << ":" << reg_info->getName(i->second);
                } else {
                    str << i->first << ":[MEM]";
                }
                std::cout.width(16);
                std::cout << str.str();
            }
            std::cout << std::endl;
        }
    }

    // Print the registers alive at the beginning of the basic block:
    print_vec(this->Z_, reg_info);

    // Print the registers alive in the Y set:
    print_vec(this->Y_, reg_info);

    std::cout << std::endl;
}


//===--------------------------------------------------------------------------
// This auxiliary method prints the pairs stored in the mapping m.
//===--------------------------------------------------------------------------
void print_reg_map(const std::map<unsigned, unsigned> & m) {
    std::map<unsigned, unsigned>::const_iterator i;
    std::map<unsigned, unsigned>::const_iterator e;
    for(i = m.begin(), e = m.end(); i != e; i++) {
        std::cout.width(10);
        std::cout << i->first << ":" << i->second;
    }
    std::cout << std::endl;
}


//===--------------------------------------------------------------------------
// This method returns an index of a basic block that has already been visited
// by the register allocator. This basic block contains a non-empty vector of
// locations.
//===--------------------------------------------------------------------------
const std::map<unsigned, unsigned> * PhiP3_Fer::get_incoming_map(
    unsigned & mbb_number
) const {
    // Find a reference to any phi_function that exists in the basic block
    // represented by this data structure:
    MachineInstr * phi = (*this->v2phi_.begin()).second;

    // Go through the elements of this phi-function to find a non-empty mapping
    // of virtuals to locations:
    for(unsigned u = phi->getNumOperands() - 1; u >= 2; u -= 2) {
        MachineBasicBlock & pred = * phi->getOperand(u).getMachineBasicBlock();
        const std::map<unsigned, unsigned> & ans =
                             this->partition_->get_tail_map(pred.getNumber());
        if(ans.size() > 0) {
            mbb_number = u;
            return & ans;
        }
    }
    return NULL;
}


//===--------------------------------------------------------------------------
// This method colors this phi block.
//===--------------------------------------------------------------------------
MachineInstr * PhiP3_Fer::color() {

    // Find a basic block that has been colored and Get the vector B of
    // preferences for that basic block:
    unsigned mbb_number = RF_UNDEF;
    const std::map<unsigned, unsigned> & map = * get_incoming_map(mbb_number);

    // This vector will hold the virtuals that do not reach the current
    // basic block in registers.
    std::vector<unsigned> uncolored_regs;

    // Find colors for the variables defined in phi-functions. Just repeat
    // the colors of the incoming virtuals.
    std::map<unsigned, MachineInstr *>::const_iterator i, e;
    for(i = v2phi_.begin(), e = v2phi_.end(); i != e; i++) {
        const MachineInstr * phi = (*i).second;
        // v_def is the virtual defined by the phi-function:
        unsigned v_def = phi->getOperand(0).getReg();
        // v_use is the virtual used in the phi-function that feeds v_def:
        unsigned v_use = phi->getOperand(mbb_number - 1).getReg();
        std::map<unsigned, unsigned>::const_iterator i_use = map.find(v_use);
        const unsigned p_use = (*i_use).second;
        if(MRegisterInfo::isPhysicalRegister(p_use)) {
            // As p_use has already been colored, we keep the same color to
            // v_def:
            this->p_regs_X_[p_use] = v_def;
            this->p_regs_Z_[p_use] = v_def;
            this->rm_->allocate_color(v_def, p_use);
        } else {
            uncolored_regs.push_back(v_def);
        }
    }

    // The virtuals defined in phi-functions have already been colored. Now
    // we must try to find colors for the virtuals that are alive at the
    // beginning of the current basic block.
    for(unsigned u = 0; u < this->Y_.size(); u++) {
        unsigned v_reg = this->Y_[u];
        std::map<unsigned, unsigned>::const_iterator i_reg = map.find(v_reg);
        const unsigned p_use = (*i_reg).second;
        if(MRegisterInfo::isPhysicalRegister(p_use)) {
            // As p_use has already been colored, we keep the same color to
            // v_reg:
            this->p_regs_X_[p_use] = v_reg;
            this->p_regs_Z_[p_use] = v_reg;
            this->rm_->allocate_color(v_reg, p_use);
        } else {
            uncolored_regs.push_back(v_reg);
        }
    }

    // Now we must find locations to the registers in the uncolored vector.
    for(unsigned u = 0; u < uncolored_regs.size(); u++) {
        unsigned v_reg = uncolored_regs[u];
        unsigned new_color = RF_UNDEF;
        if(this->pa_->has_preference(v_reg)) {
            // The register if phi-related and has some preference. We will
            // try to accommodate this preference:
            unsigned pref_reg = this->pa_->get_preference(v_reg);
            if(
                Regs_Fer::is_register_available
                       (pref_reg, this->p_regs_X_, this->p_regs_Z_, *this->mf_)
            ) {
                new_color = pref_reg;
            }
        }
        if(new_color == RF_UNDEF) {
            // preference is not available. Find another color:
            new_color = Regs_Fer::get_free_register
                                     (v_reg, p_regs_X_, p_regs_Z_, *this->mf_);
        }
        std::vector<unsigned> sp_regs;
        if(new_color == RF_UNDEF) {
            // We could not find a location for v_reg. So spill something.
            new_color = partition_->choose_reg_and_spill
                                                (v_reg, * this, sp_regs, true);
            this->numIter_++;
            if(this->numIter_ > maxIterations) {
                maxIterations = numIter_;
            }
        }
        // There is a possibility that v_reg is the best candidate to be
        // spilled.
        if(MRegisterInfo::isPhysicalRegister(new_color)) {
            // clean the spilled registers if there is any:
            for(unsigned u = 0; u < sp_regs.size(); u++) {
                this->p_regs_X_[sp_regs[u]] = RF_UNDEF;
                this->p_regs_Z_[sp_regs[u]] = RF_UNDEF;
            }
            // a different variable has been spilled. Take its register:
            this->p_regs_X_[new_color] = v_reg;
            this->p_regs_Z_[new_color] = v_reg;
            this->rm_->allocate_color(v_reg, new_color);
        }
    }

    // We must go over the registers defined in phi-functions, and for those
    // that got allocated into some register, we must define this binding.
    std::map<unsigned, MachineInstr *>::const_iterator i_phi, e_phi;
    for(
        i_phi = this->v2phi_.begin(), e_phi = this->v2phi_.end();
        i_phi != e_phi;
        i_phi++
    ) {
        MachineInstr *phi = i_phi->second;
        unsigned v_def = i_phi->first;
        unsigned p_reg = this->rm_->get_physical_location(v_def);
        if(MRegisterInfo::isPhysicalRegister(p_reg)) {
            this->rs_->add_binding(p_reg, v_def, phi);
        } else {
            this->rs_->add_binding(RF_SPILL, v_def, phi);
        }
    }

    if(this->numIter_ == 0) {
        zeroIter++;
    }

    return NULL;
}
