//===------------- RA_Result_Fer.h - Bank of Machine Registers ------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This class holds the results of register allocation.
//
// Date: June 27th, 2007
//
//===----------------------------------------------------------------------===//
//
// This class holds the results of register allocation. These results consist
// in:
// Instructions -> {Virtual X Register}
// Instructions -> {Event}, where an event is a copy, load, store or swap
// Basic block -> {Event}
//
//===----------------------------------------------------------------------===//

#define DEBUG_TYPE "RA_results"

#include <iostream>                              // cerr

#include "llvm/Function.h"

#include "llvm/Target/TargetInstrInfo.h"

#include "llvm/ADT/STLExtras.h"
#include "llvm/ADT/Statistic.h"

#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/CodeGen/MachineInstr.h"
#include "llvm/CodeGen/MachineFunction.h"
#include "llvm/CodeGen/MachineBasicBlock.h"

#include "llvm/Support/Debug.h"

#include "llvm/CodeGen/LiveIntervalAnalysis.h"

#include "llvm/fernando/Regs_Fer.h"
#include "llvm/fernando/RevVec_Fer.h"
#include "llvm/fernando/NeatPrint_Fer.h"
#include "llvm/fernando/Partition_Fer.h"
#include "llvm/fernando/RegMapping_Fer.h"
#include "llvm/fernando/DefUseSites_Fer.h"

#include "llvm/fernando/RA_Result_Fer.h"

using namespace llvm;

STATISTIC(totalLoads, "Number of loads inserted due to spilling.");
STATISTIC(savedLoads, "Number of loads saved due to LR reduction.");
STATISTIC(savedCopies, "Number of copies saved due to LR reduction.");

// Instructions inserted to destroy phi-functions:
STATISTIC(numLdPH, "Number of Loads - PHI.");
STATISTIC(numMvPH, "Number of Moves - PHI.");
STATISTIC(numSwPH, "Number of Swaps - PHI.");
STATISTIC(numStPH, "Number of Stores - PHI.");

// Instructions inserted to destroy pi-functions:
STATISTIC(numLdPI, "Number of Loads - PI.");
STATISTIC(numMvPI, "Number of Moves - PI.");
STATISTIC(numSwPI, "Number of Swaps - PI.");
STATISTIC(numStPI, "Number of Stores - PI.");

// Instructions inserted to destroy parallel-copies:
STATISTIC(numLdPC, "Number of Loads - PC.");
STATISTIC(numMvPC, "Number of Moves - PC.");
STATISTIC(numSwPC, "Number of Swaps - PC.");
STATISTIC(numStPC, "Number of Stores - PC.");

//===--------------------------------------------------------------------------
// Constructor method.
//===--------------------------------------------------------------------------
RA_Result_Fer::RA_Result_Fer(
    MachineFunction & mf,
    const LiveIntervals * li,
    const RegMapping_Fer * rm,
    Partition_Fer *pf,
    DefUseSites_Fer *du
) :
    li_(li),
    rm_(rm),
    mf_(& mf),
    pf_(pf),
    du_(du)
{
    // Number of basic blocks:
    unsigned num_mbb = mf.size();

    this->mbb_event_list_.resize(num_mbb);
    this->mock_event_list_.resize(num_mbb);
    this->mbb_beginning_list_.resize(num_mbb);

    // Number of machine instructions:
    unsigned num = LiveIntervals::InstrSlots::NUM;
    unsigned num_mi = (int)(li->maxInstrIndex_ / num);

    this->instr_event_list_.resize(num_mi);
    this->binding_list_.resize(num_mi);

    // Get the number of virtual registers:
    SSARegMap * srm = mf.getSSARegMap();
    unsigned last_v_reg = srm->getLastVirtReg() + 1;
    unsigned num_v_regs = last_v_reg - MRegisterInfo::FirstVirtualRegister;

    this->num_loads_.resize(num_v_regs);
}


#define STACK_MARK 1000000


//===--------------------------------------------------------------------------
// Print the instructions that preecede the instruction, such as loads and
// copies.
//===--------------------------------------------------------------------------
void RA_Result_Fer::print_event_list(const std::vector<Event_Fer> & vec) const {
    for(unsigned u = 0; u < vec.size(); u++) {
        const Event_Fer & e = vec[u];
        switch(e.type) {
            case CP:
                // copy:
                Regs_Fer::print_def(*this->mf_, e.l_dst, e.v_dst);
                std::cout << " = ";
                Regs_Fer::print_reg(e.l_src, e.v_src);
                std::cout << " ;\n";
                break;
            case ST:
                // store:
                Regs_Fer::print_reg(e.l_dst + STACK_MARK, e.v_dst);
                std::cout << " = ";
                Regs_Fer::print_reg(e.l_src, e.v_src);
                std::cout << " ;\n";
                break;
            case LD:
                // load:
                if(e.v_src != e.v_dst || !load_local(e.v_src)) {
                    Regs_Fer::print_def(*this->mf_, e.l_dst, e.v_dst);
                    std::cout << " = ";
                    Regs_Fer::print_reg(e.l_src + STACK_MARK, e.v_src);
                    std::cout << " ;\n";
                }
                break;
            case SW:
                // swap:
                std::cout << "Insert swap = " << e.l_dst << " " << e.l_src;
                std::cout << " ; // swap \n";
                break;
            default:
                assert(false && "Unknown event type.");
        }
    }
    // print the swaps:
}


//===--------------------------------------------------------------------------
// This method returns true if mi has been already visited by the register
// allocator and physical registers have been assigned to its variables.
//===--------------------------------------------------------------------------
bool RA_Result_Fer::has_binding(MachineInstr * mi) const {
    unsigned num = LiveIntervals::InstrSlots::NUM;
    unsigned m_index = this->li_->getInstructionIndex(mi);
    unsigned i = (int)(m_index / num);

    return this->binding_list_[i].size() > 0;
}


//===--------------------------------------------------------------------------
// This map searches the vector 'vec' for a binding containing the register
// v_reg and returns a pointer to this binding object.
//===--------------------------------------------------------------------------
const Binding_Fer * get_event
                       (unsigned v_reg, const std::vector<Binding_Fer> & vec) {
    const Binding_Fer * bf = NULL;
    for(unsigned u = 0; u < vec.size(); u++) {
        if(vec[u].v_reg == v_reg) {
            bf = & vec[u];
        }
    }
    if(bf == NULL) {
        std::cerr << "Null binding: " << v_reg << std::endl;
    }
    assert(bf != NULL && "Unbound virtual.");
    return bf;
}


//===--------------------------------------------------------------------------
// This method prints the register binding for a phi-function. Only the
// definition is printed.
//===--------------------------------------------------------------------------
void RA_Result_Fer::print_phi_function(
    const MachineInstr * phi,
    unsigned i
) const {
    unsigned def_reg = phi->getOperand(0).getReg();
    const Binding_Fer & bf = * get_event(def_reg, binding_list_[i]);
    unsigned p_reg = bf.p_reg;
    unsigned v_reg = bf.v_reg;

    if(this->rm_->is_spill(v_reg)) {
        // Check if v_reg reaches the basic block in memory:
        const MachineBasicBlock * mbb = phi->getParent();
        const std::map<unsigned, unsigned> & map =
                                     this->pf_->get_head_map(mbb->getNumber());
        std::map<unsigned, unsigned>::const_iterator loc = map.find(v_reg);
        assert(loc != map.end() && "Invalid location in phi-deconstruction.");
        if(loc->second != RF_SPILL) {
            unsigned slot = this->rm_->get_stack_slot(v_reg);
            Regs_Fer::print_reg(slot + STACK_MARK, v_reg);
            std::cout << " = ";
            Regs_Fer::print_reg(p_reg, v_reg);
            std::cout << " ;\n";
        }
    }
}


//===--------------------------------------------------------------------------
// Prints the binddings on this instructions.
//===--------------------------------------------------------------------------
void RA_Result_Fer::print_bindings(
    const MachineInstr * mi,
    unsigned i,
    const TargetInstrInfo & tii
) const {

    const TargetInstrDescriptor & tid = tii.get(mi->getOpcode());

    // print the local loads:
    for(unsigned u = 0; u < mi->getNumOperands(); u++) {
        const MachineOperand & mo = mi->getOperand(u);
        if(mo.isRegister() && mo.getReg() && mo.isUse()) {
            unsigned p_reg, reg;
            if(MRegisterInfo::isVirtualRegister(mo.getReg())) {
                const Binding_Fer & bf =
                                    * get_event(mo.getReg(), binding_list_[i]);
                p_reg = bf.p_reg;
                reg = bf.v_reg;
                if(
                   this->rm_->is_spill(reg) &&
                   load_local(reg) &&
                   !X86PuzzleBoard_Fer::is_def_two_addr(mi, reg)
                ) {
                    unsigned slot = this->rm_->get_stack_slot(reg);
                    Regs_Fer::print_reg(bf.p_reg, bf.v_reg);
                    std::cout << " = ";
                    Regs_Fer::print_reg(slot + STACK_MARK, bf.v_reg);
                    std::cout << " ; // load local\n";
                }
            }
        }
    }

    // print the implict definitions:
    if(tid.ImplicitDefs) {
        for (const unsigned *implicit_def = tid.ImplicitDefs;
                                               *implicit_def; ++implicit_def) {
            Regs_Fer::print_reg(* implicit_def, * implicit_def);
        }
    }

    // We save the variables that must be stored in this vector.
    std::vector<Binding_Fer> store_biddings;

    // print the explict definitions:
    for(unsigned u = 0; u < mi->getNumOperands(); u++) {
        const MachineOperand & mo = mi->getOperand(u);
        if(mo.isRegister() && mo.getReg() && mo.isDef()) {
            unsigned p_reg, reg;
            if(MRegisterInfo::isVirtualRegister(mo.getReg())) {
                const Binding_Fer & bf =
                                    * get_event(mo.getReg(), binding_list_[i]);
                p_reg = bf.p_reg;
                reg = bf.v_reg;
                if(this->rm_->is_spill(reg)) {
                    if(
                        !this->rm_->is_defined_in_2addr_instr(reg) ||
                        X86PuzzleBoard_Fer::is_two_addr(mi)
                    ) {
                        store_biddings.push_back(bf);
                    }
                }
            } else {
                p_reg = mo.getReg();
                reg = mo.getReg();
            }
            Regs_Fer::print_def(*this->mf_, p_reg, reg);
        }
    }

    std::cout << " = ";

    // print the explicit uses:
    for(unsigned u = 0; u < mi->getNumOperands(); u++) {
        const MachineOperand & mo = mi->getOperand(u);
        if(mo.isRegister() && mo.getReg() && mo.isUse()) {
            unsigned p_reg, reg;
            if(MRegisterInfo::isVirtualRegister(mo.getReg())) {
                const Binding_Fer & bf =
                                    * get_event(mo.getReg(), binding_list_[i]);
                p_reg = bf.p_reg;
                reg = bf.v_reg;
            } else {
                p_reg = mo.getReg();
                reg = mo.getReg();
            }
            Regs_Fer::print_reg(p_reg, reg);
        }
    }

    // print the implicit uses:
    if(tid.ImplicitUses) {
        for(const unsigned *implicit_use = tid.ImplicitUses;
                                               *implicit_use; implicit_use++) {
            Regs_Fer::print_reg(* implicit_use, * implicit_use);
        }
    }

    std::cout << " ;\n";

    // print the stores, if any is necessary:
    for(unsigned u = 0; u < store_biddings.size(); u++) {
        Binding_Fer & bf = store_biddings[u];
        unsigned slot = this->rm_->get_stack_slot(bf.v_reg);
        Regs_Fer::print_reg(slot + STACK_MARK, bf.v_reg);
        std::cout << " = ";
        Regs_Fer::print_reg(bf.p_reg, bf.v_reg);
        std::cout << " ;\n";
    }
}


//===--------------------------------------------------------------------------
// This method prints all the instructions necessary at the end of the basic
// block. It includes jumps, and also the instructions used to destroy
// phi-functions.
//===--------------------------------------------------------------------------
void RA_Result_Fer::print_tail_code(
    MachineFunction & mf,
    const MachineBasicBlock * mbb
) const {

    // Prints the code to transfer values to the next block:
    this->print_event_list(this->mbb_event_list_[mbb->getNumber()]);

    // prints the ending JUMP
    std::cout << "JMP ";
    if(mbb->succ_begin() == mbb->succ_end()) {
        std::cout << "L" << mf.getNumBlockIDs();
    } else {
        for(MachineBasicBlock::const_succ_iterator scc =
                            mbb->succ_begin(); scc != mbb->succ_end(); scc++) {
            std::cout << "L" << (*scc)->getNumber() << " ";
        }
    }

    std::cout << " ;\n\n";
}


//===--------------------------------------------------------------------------
// Prints the name of the function between comments.
//===--------------------------------------------------------------------------
void print_function_header(const MachineFunction & mf) {
    const Function * f = mf.getFunction();
    std::cout << "/*\n";
    std::cout << " * Function name: " << f->getName() << std::endl;
    std::cout << " * Number of basic blocks: " << mf.size() << std::endl;
    std::cout << " */\n\n";
}


//===--------------------------------------------------------------------------
// Prints a representation of this object.
//===--------------------------------------------------------------------------
void RA_Result_Fer::dump(MachineFunction & mf) const {

    print_function_header(mf);

    const TargetMachine & target_machine = mf.getTarget();
    const TargetInstrInfo & tii = * (target_machine.getInstrInfo());

    // Go through all the basic blocks in the function:
    MachineFunction::iterator mfi, mfe;
    for (mfi = mf.begin(), mfe = mf.end(); mfi != mfe; mfi++) {
        MachineBasicBlock * mbb = mfi;
        // Printing the header information:
        unsigned mbb_num = mbb->getNumber();
        std::cout << "L" << mbb_num << "\n";
        this->print_event_list(this->mbb_beginning_list_[mbb_num]);

        // Go through all the instructions in the basic block:
        MachineBasicBlock::iterator mbbi, mbbe;
        for(mbbi = mbb->begin(), mbbe = mbb->end(); mbbi != mbbe; mbbi++) {
            MachineInstr * mi = mbbi;
            // Get the index of the instruction:
            unsigned num = LiveIntervals::InstrSlots::NUM;
            unsigned m_index = this->li_->getInstructionIndex(mi);
            unsigned i = (int)(m_index / num);
            if(mi->getOpcode() == TargetInstrInfo::PHI) {
                this->print_phi_function(mi, i); 
            } else {
                this->print_event_list(this->instr_event_list_[i]);
                this->print_bindings(mi, i, tii);
            }   
        }

        this->print_tail_code(mf, mbb);
    }

    unsigned last_label = mf.size();
    std::cout << "\nL" << last_label << std::endl;
    std::cout << "JMP L" << last_label << " ;\n";
}


//===--------------------------------------------------------------------------
// Adds the binding (loc, v_reg) at instructin i.
//===--------------------------------------------------------------------------
void RA_Result_Fer::add_binding(
    unsigned loc,
    unsigned v_reg,
    MachineInstr *mi
) {
    // Get the index of the instruction:
    unsigned num = LiveIntervals::InstrSlots::NUM;
    unsigned m_index = this->li_->getInstructionIndex(mi);
    unsigned i = (int)(m_index / num);
//std::cout << "Adding binding (" << loc << ", " << v_reg << ") at " << i << "\n";

    assert(i < this->binding_list_.size() && "Invalid instruction");
    Binding_Fer bf(v_reg, loc);
    this->binding_list_[i].push_back(bf);
}

//===--------------------------------------------------------------------------
// Add the copy (l_dst, v_dst) = (l_src, v_src) before i
//===--------------------------------------------------------------------------
void RA_Result_Fer::add_cp(
    unsigned l_dst,
    unsigned v_dst,
    unsigned l_src,
    unsigned v_src,
    MachineInstr * mi
) {
    // Get the index of the instruction:
    unsigned num = LiveIntervals::InstrSlots::NUM;
    unsigned m_index = this->li_->getInstructionIndex(mi);
    unsigned i = (int)(m_index / num);
//std::cout << "<> Adding copy (" << l_dst << ", " << v_dst << "), (" << l_src << ", " << v_src << ") at " << i << "\n";

    assert(i < this->instr_event_list_.size() && "Invalid instruction");
    Event_Fer e(CP, l_dst, v_dst, l_src, v_src);
    this->instr_event_list_[i].push_back(e);
}

//===--------------------------------------------------------------------------
// Add the load (l_dst, v_dst) = (l_src, v_src) before i
//===--------------------------------------------------------------------------
void RA_Result_Fer::add_ld(
    unsigned l_dst,
    unsigned v_dst,
    unsigned l_src,
    unsigned v_src,
    MachineInstr * mi
) {
    // Get the index of the instruction:
    unsigned num = LiveIntervals::InstrSlots::NUM;
    unsigned m_index = this->li_->getInstructionIndex(mi);
    unsigned i = (int)(m_index / num);
//std::cout << "<> Adding load (" << l_dst << ", " << v_dst << "), (" << l_src << ", " << v_src << ") at " << i << "\n";

    assert(i < this->instr_event_list_.size() && "Invalid instruction");
    Event_Fer e(LD, l_dst, v_dst, l_src, v_src);
    this->instr_event_list_[i].push_back(e);
    this->inc_load(v_src);
}

//===--------------------------------------------------------------------------
// Add the store (l_dst, v_dst) = (l_src, v_src) after i
//===--------------------------------------------------------------------------
void RA_Result_Fer::add_st(
    unsigned l_dst,
    unsigned v_dst,
    unsigned l_src,
    unsigned v_src,
    MachineInstr * mi
) {
    // Get the index of the instruction:
    unsigned num = LiveIntervals::InstrSlots::NUM;
    unsigned m_index = this->li_->getInstructionIndex(mi);
    unsigned i = (int)(m_index / num);

//std::cout << "<> Adding store (" << l_dst << ", " << v_dst << "), (" << l_src << ", " << v_src << ") at " << i << "\n";

    assert(i < this->instr_event_list_.size() && "Invalid instruction");
    Event_Fer e(ST, l_dst, v_dst, l_src, v_src);
    this->instr_event_list_[i].push_back(e);
}

//===--------------------------------------------------------------------------
// Add the copy (l_dst, v_dst) = (l_src, v_src) at the end of machine basic
// block b.
//===--------------------------------------------------------------------------
void RA_Result_Fer::add_cp2mbb(
    unsigned l_dst,
    unsigned v_dst,
    unsigned l_src,
    unsigned v_src,
    unsigned b,
    bool add_at_end
) {
    if(add_at_end) {
//std::cout << "<> Adding copy (" << l_dst << ", " << v_dst << "), (" << l_src << ", " << v_src << ") at the end of " << b << "\n";
        assert(b < this->mbb_event_list_.size() && "Invalid basic block");
        Event_Fer e(CP, l_dst, v_dst, l_src, v_src);
        this->mbb_event_list_[b].push_back(e);
    } else {
//std::cout << "<> Adding copy (" << l_dst << ", " << v_dst << "), (" << l_src << ", " << v_src << ") at the beginning of " << b << "\n";
        assert(b < this->mbb_beginning_list_.size() && "Invalid basic block");
        Event_Fer e(CP, l_dst, v_dst, l_src, v_src);
        this->mbb_beginning_list_[b].push_back(e);
    }
}

//===--------------------------------------------------------------------------
// This method adds a 'mock' load at the end of basic block b. This load will
// be implemented only if v_src is marked as a 'local load' by the load
// reduction heuristics.
// Add the load (l_dst, v_dst) = (l_src, v_src) at the end of machine basic
// block b.
//===--------------------------------------------------------------------------
void RA_Result_Fer::add_mock_load(
    unsigned l_dst,
    unsigned v_dst,
    unsigned l_src,
    unsigned v_src,
    unsigned b
) {
//std::cout << "<> Adding mock load (" << l_dst << ", " << v_dst << "), (" << l_src << ", " << v_src << ") at the end of " << b << "\n";
    assert(b < this->mbb_event_list_.size() && "Invalid basic block");
    Event_Fer e(LD, l_dst, v_dst, l_src, v_src);
    this->mock_event_list_[b].push_back(e);
}

//===--------------------------------------------------------------------------
// Add the load (l_dst, v_dst) = (l_src, v_src) at the end of machine basic
// block b.
//===--------------------------------------------------------------------------
void RA_Result_Fer::add_ld2mbb(
    unsigned l_dst,
    unsigned v_dst,
    unsigned l_src,
    unsigned v_src,
    unsigned b,
    bool add_at_end
) {
    if(add_at_end) {
//std::cout << "<> Adding load (" << l_dst << ", " << v_dst << "), (" << l_src << ", " << v_src << ") at the end of " << b << "\n";
        assert(b < this->mbb_event_list_.size() && "Invalid basic block");
        Event_Fer e(LD, l_dst, v_dst, l_src, v_src);
        this->mbb_event_list_[b].push_back(e);
    } else {
//std::cout << "<> Adding load (" << l_dst << ", " << v_dst << "), (" << l_src << ", " << v_src << ") at the beginning of " << b << "\n";
        assert(b < this->mbb_beginning_list_.size() && "Invalid basic block");
        Event_Fer e(LD, l_dst, v_dst, l_src, v_src);
        this->mbb_beginning_list_[b].push_back(e);
    }
}

//===--------------------------------------------------------------------------
// Add the store (l_dst, v_dst) = (l_src, v_src) at the end of machine basic
// block b.
//===--------------------------------------------------------------------------
void RA_Result_Fer::add_st2mbb(
    unsigned l_dst,
    unsigned v_dst,
    unsigned l_src,
    unsigned v_src,
    unsigned b,
    bool add_at_end
) {
    if(add_at_end) {
        assert(b < this->mbb_event_list_.size() && "Invalid basic block");
        Event_Fer e(ST, l_dst, v_dst, l_src, v_src);
        this->mbb_event_list_[b].push_back(e);
//std::cout << "<> Adding store (" << l_dst << ", " << v_dst << "), (" << l_src << ", " << v_src << ") at the end of block " << b << "\n";
    } else {
        assert(b < this->mbb_beginning_list_.size() && "Invalid basic block");
        Event_Fer e(ST, l_dst, v_dst, l_src, v_src);
        this->mbb_beginning_list_[b].push_back(e);
//std::cout << "<> Adding store (" << l_dst << ", " << v_dst << "), (" << l_src << ", " << v_src << ") at the beginning of block " << b << "\n";
    }
}


//===--------------------------------------------------------------------------
// SW: (l_dst, v_src) (l_src, v_dst) = (l_dst, v_dst) (l_src, v_src) at
// the end of basic block b
//===--------------------------------------------------------------------------
void RA_Result_Fer::add_sw2mbb(
    unsigned l_dst,
    unsigned v_dst,
    unsigned l_src,
    unsigned v_src,
    unsigned b
) {
//std::cout << "<> Adding swap (" << l_dst << ", " << v_dst << "), (" << l_src << ", " << v_src << ") at the end of " << b << std::endl;
    assert(b < this->mbb_event_list_.size() && "Invalid basic block");
    Event_Fer e(SW, l_dst, v_dst, l_src, v_src);
    this->mbb_event_list_[b].push_back(e);
}


//===--------------------------------------------------------------------------
// This method determines where we should insert loads for spilled variables.
// The puzzle solver tends to split the live range of variables into many
// smaller pieces, and inserts loads to keep the splitting consistent. One
// typical example happens at the beginning of basic blocks where loads must be
// inserted to join the live ranges of variables that reach the basic block at
// both memory and registers. However, if the variable is used very few times,
// say once, it is not worth to insert all these loads. In this case we could
// replace all these loads with a single load. This method determines when we
// should do this.
//===--------------------------------------------------------------------------
bool RA_Result_Fer::load_local(unsigned v_reg) const {
/*
    // The results of this heuristics are in pz5
    unsigned num_uses = this->du_->get_num_uses(v_reg);
    bool used_once = num_uses == 1;
    bool two_addr_used_twice = num_uses == 2 &&
                                   this->rm_->is_defined_in_2addr_instr(v_reg);
    return used_once || two_addr_used_twice;
*/
/*
    // The results of this heuristics are in pz6
    unsigned num_uses = this->du_->get_num_uses(v_reg);
    if(this->rm_->is_defined_in_2addr_instr(v_reg)) {
        num_uses--;
    }

    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister; 
    if(this->num_loads_[base_reg] >= (2 * num_uses)) {
        return true;
    } else {
        return false;
    }
*/
    // The results of this heuristics are in pz7
    unsigned num_uses = this->du_->get_num_uses(v_reg);
    if(this->rm_->is_defined_in_2addr_instr(v_reg)) {
        num_uses--;
    }

    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister; 
    if(
        this->num_loads_[base_reg] >= (2 * num_uses) ||
        this->num_loads_[base_reg] >= (2 + num_uses)
    ) {
        return true;
    } else {
        return false;
    }
/*
    // The results of this heuristics are in pz8
    unsigned num_uses = this->du_->get_num_uses(v_reg);
    if(this->rm_->is_defined_in_2addr_instr(v_reg)) {
        num_uses--;
    }

    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister; 
    if(this->num_loads_[base_reg] > num_uses) {
        return true;
    } else {
        return false;
    }
*/
}


//===--------------------------------------------------------------------------
// This method increments by one the number of loads inserted to split the
// live range of v_reg.
//===--------------------------------------------------------------------------
void RA_Result_Fer::inc_load(unsigned v_reg) {
    assert(MRegisterInfo::isVirtualRegister(v_reg) &&
                                                   "Reloading a pre-colored?");
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister; 
    this->num_loads_[base_reg]++;
    totalLoads++;
}


//===--------------------------------------------------------------------------
// This method converts the mapping from virtual register to physical register
// by the register allocator to a real assignment of  physical registers in
// each operand of each instruction of the given machine function.
//===--------------------------------------------------------------------------
void RA_Result_Fer::hard_code_register_mapping() {
    const TargetMachine & target_machine = this->mf_->getTarget();
    const MRegisterInfo & reg_info = * target_machine.getRegisterInfo();

//this->dump(*this->mf_);

    // Go through all the basic blocks in the function substituting virtuals
    // by physical registers.
    MachineFunction::iterator mfi, mfe;
    for (mfi = this->mf_->begin(), mfe = this->mf_->end(); mfi != mfe; mfi++) {
        MachineBasicBlock * mbb = mfi;
        // Go through all the instructions in the basic block:
        this->stat_st = RA_Result_Fer::StatsState::PC;
        this->hard_code_bindings(* mbb, reg_info); 
        this->destroy_phi_functions(* mbb, reg_info);
    }

    // Go through all the basic blocks inserting code to implement the parallel
    // copies that existed before register allocation:
    for (mfi = this->mf_->begin(), mfe = this->mf_->end(); mfi != mfe; mfi++) {
        MachineBasicBlock * mbb = mfi;
        unsigned mbb_num = mbb->getNumber();

        // Insert code at the beginning of the basic block. This code is
        // necessary if the basic block is the successor of another basic
        // block that has many successors.
        this->stat_st = RA_Result_Fer::StatsState::PI;
        MachineBasicBlock::iterator mbb_begin = mbb->begin();
        this->hard_code_event_list
                  (mbb_beginning_list_[mbb_num], * mbb, & mbb_begin, reg_info);

        // Insert code at the end of the basic block. This type of code is
        // necessary to correctly implement a basic block if it is the
        // predecessor of a basic block with many predecessors.
        this->stat_st = RA_Result_Fer::StatsState::PHI;
        this->hard_code_mock_loads(mock_event_list_[mbb_num], *mbb, reg_info);
        MachineBasicBlock::iterator ft = mbb->getFirstTerminator();
        this->hard_code_event_list
                         (this->mbb_event_list_[mbb_num], *mbb, &ft, reg_info);
    }

    // Again, go through all the instructions, and remove the moves that have
    // been assigned the same physical registers.
    const TargetInstrInfo & tii = * target_machine.getInstrInfo();
    std::vector<MachineInstr *> to_remove;
    for (mfi = this->mf_->begin(), mfe = this->mf_->end(); mfi != mfe; mfi++) {
        MachineBasicBlock * mbb = mfi;
        MachineBasicBlock::iterator mbbi, mbbe;
        for(mbbi = mbb->begin(), mbbe = mbb->end(); mbbi != mbbe; mbbi++) {
            MachineInstr * mi = mbbi;
            unsigned use, def;
            if(tii.isMoveInstr(* mi, use, def) && use == def) {
                to_remove.push_back(mi);
            }
        }
    }
    for(unsigned u = 0; u < to_remove.size(); u++) {
        MachineInstr * mi = to_remove[u];
        mi->eraseFromParent();
    }
}

//===--------------------------------------------------------------------------
// This method verifies if p_reg can be used to allocated v_reg. This will be
// the case if both registers belong to the same class.
//===--------------------------------------------------------------------------
bool compatible_class(MachineFunction & mf, unsigned v_reg, unsigned p_reg) {
    assert(MRegisterInfo::isPhysicalRegister(p_reg) &&
                                           "Target register must be physical");
    const TargetRegisterClass *trc = mf.getSSARegMap()->getRegClass(v_reg);
    return trc->contains(p_reg);
}


//===--------------------------------------------------------------------------
// This method removes all the occurrences of phi-functions from the given
// basic block. Before this method is called, instructions must be inserted in
// the control flow graph to reproduce the behavior of phi functions.
// This method also inserts stores for the values defined by phi-functions in
// registers.
//===--------------------------------------------------------------------------
MachineBasicBlock::iterator RA_Result_Fer::destroy_phi_functions(
    MachineBasicBlock & mbb,
    const MRegisterInfo & reg_info
) {
    std::vector<Binding_Fer> vec;

    while(!mbb.empty()) {
        if(mbb.front().getOpcode() == TargetInstrInfo::PHI) {
            MachineInstr * mphi = mbb.remove(mbb.begin());

            // Variable defined by the phi-function.
            unsigned def_reg = mphi->getOperand(0).getReg();

            // get the instruction index:
            unsigned num = LiveIntervals::InstrSlots::NUM;
            unsigned m_index = this->li_->getInstructionIndex(mphi);
            unsigned idx = (int)(m_index / num);

            // Save the binding, if the register defined by the phi-function
            // is a spill.
            const Binding_Fer & bf = * get_event(def_reg, binding_list_[idx]);
            if(this->rm_->is_spill(bf.v_reg)) {
                vec.push_back(bf);
            }

            delete mphi;
        } else {
            break;
        }
    }

    // Go through the list of bindings and add stores for those that have been
    // spilled and reached the mbb block in registers.
    MachineBasicBlock::iterator rt_i = mbb.begin();
    const std::map<unsigned, unsigned> & map =
                                      this->pf_->get_head_map(mbb.getNumber());
    for(unsigned u = 0; u < vec.size(); u++) {
        Binding_Fer & bf = vec[u];
        std::map<unsigned, unsigned>::const_iterator loc = map.find(bf.v_reg);
        assert(loc != map.end() && "Invalid location in phi-deconstruction.");
        if(loc->second != RF_SPILL) {
            // bf.v_reg is in memory, but it does not reach mbb in memory.
            // Thus, it must be stored.
            unsigned slot = this->rm_->get_stack_slot(bf.v_reg);
            const TargetRegisterClass * rc =
                              this->mf_->getSSARegMap()->getRegClass(bf.v_reg);
//std::cout << "Storing (" << bf.v_reg << ", " << bf.p_reg << ") at the very beginning of block " << mbb.getNumber() << std::endl; 
            reg_info.storeRegToStackSlot(mbb, mbb.begin(), bf.p_reg, slot, rc);
        }
    }
    return rt_i;
}


//===--------------------------------------------------------------------------
// This method implements the mock loads. Only loads of variables marked as
// 'load local' will be inserted at the end of the basic block mbb.
//===--------------------------------------------------------------------------
void RA_Result_Fer::hard_code_mock_loads(
    const std::vector<Event_Fer> & vec,
    MachineBasicBlock & mbb,
    const MRegisterInfo & reg_info
) {
    MachineBasicBlock::iterator ft = mbb.getFirstTerminator();
    for(unsigned u = 0; u < vec.size(); u++) {
        const Event_Fer & e = vec[u];
        if(load_local(e.v_src)) {
            const TargetRegisterClass * rc =
                               this->mf_->getSSARegMap()->getRegClass(e.v_src);
            int slot = this->rm_->get_stack_slot(e.v_src);
            reg_info.loadRegFromStackSlot(mbb, ft, e.l_dst, slot, rc);
//std::cout << "Inserting LOAD LOCAL at mem_" << slot << " (" << e.v_src << ", " << e.v_dst << ") at MBB_" << mbb.getNumber() << std::endl;
        }
    }
}

//===--------------------------------------------------------------------------
// This method gets a list of events and an iterator and inserts code before
// the iterator in order to implement these events.
//===--------------------------------------------------------------------------
void RA_Result_Fer::hard_code_event_list(
    const std::vector<Event_Fer> & vec,
    MachineBasicBlock & mbb,
    MachineBasicBlock::iterator * mbbi,
    const MRegisterInfo & reg_info
) {
    for(unsigned u = 0; u < vec.size(); u++) {
        const Event_Fer & e = vec[u];
        int slot = 0;
        const TargetRegisterClass *rc;
        unsigned stats_tp = RA_Result_Fer::StatsType::COPY;
        switch(e.type) {
            case CP:
                // copy:
                if(
                    e.v_src == e.v_dst &&
                    this->rm_->is_spill(e.v_dst) &&
                    load_local(e.v_src)
                ) {
//std::cout << "Omitting copy for " << " (" << e.v_src << ", " << e.v_dst << ") at MBB_" << mbb.getNumber() << std::endl;
                    savedCopies++;
                } else if(e.l_dst != e.l_src) {
                    rc = mf_->getSSARegMap()->getRegClass(e.v_dst);
                    reg_info.copyRegToReg(mbb, *mbbi, e.l_dst, e.l_src, rc);
//std::cout << "Inserting copy: " << e.v_dst << "(" << e.l_dst << ") <= " << e.v_src << "(" << e.l_src << ") at MBB_" << mbb.getNumber() << std::endl;
                    stats_tp = RA_Result_Fer::StatsType::COPY;
                }
                break;
            case ST:
                // store:
                assert(false && "Store in event list.");
                break;
            case LD:
                // load:
                if(e.v_src != e.v_dst || !load_local(e.v_src)) {
                    rc = mf_->getSSARegMap()->getRegClass(e.v_src);
                    slot = this->rm_->get_stack_slot(e.v_src);
                    reg_info.loadRegFromStackSlot(mbb,*mbbi,e.l_dst,slot,rc);
//std::cout << "Inserting load at mem_" << slot << " (" << e.v_src << ", " << e.v_dst << ") at MBB_" << mbb.getNumber() << std::endl;
                    stats_tp = RA_Result_Fer::StatsType::LOAD;
                } else {
//std::cout << "Omitting load for " << " (" << e.v_src << ", " << e.v_dst << ") at MBB_" << mbb.getNumber() << std::endl;
                    savedLoads++;
                }
                break;
            case SW:
                // swap:
                rc = mf_->getSSARegMap()->getRegClass(e.v_src);
                reg_info.swapRegs(mbb, *mbbi, e.l_dst, e.l_src, rc);
                stats_tp = RA_Result_Fer::StatsType::SWAP;
                break;
            default:
                assert(false && "Unknown event type.");
        }
        this->compute_statistics(stats_tp);
    }
}


//===--------------------------------------------------------------------------
// This method computes statistics about the number of moves, stores, loads and
// swaps inserted due to phi-functions are parallel copies.
//===--------------------------------------------------------------------------
void RA_Result_Fer::compute_statistics(unsigned stats_tp) {
    switch(stats_tp) {
        case RA_Result_Fer::StatsType::COPY:
            switch(stat_st) {
                case RA_Result_Fer::StatsState::PHI:
                    numMvPH++;
                    return;
                case RA_Result_Fer::StatsState::PI:
                    numMvPI++;
                    return;
                case RA_Result_Fer::StatsState::PC:
                    numMvPC++;
                    return;
            }
            break;
        case RA_Result_Fer::StatsType::LOAD:
            switch(stat_st) {
                case RA_Result_Fer::StatsState::PHI:
                    numLdPH++;
                    return;
                case RA_Result_Fer::StatsState::PI:
                    numLdPI++;
                    return;
                case RA_Result_Fer::StatsState::PC:
                    numLdPC++;
                    return;
            }
            break;
        case RA_Result_Fer::StatsType::STORE:
            switch(stat_st) {
                case RA_Result_Fer::StatsState::PHI:
                    numStPH++;
                    return;
                case RA_Result_Fer::StatsState::PI:
                    numStPI++;
                    return;
                case RA_Result_Fer::StatsState::PC:
                    numStPC++;
                    return;
            }
            break;
        case RA_Result_Fer::StatsType::SWAP:
            switch(stat_st) {
                case RA_Result_Fer::StatsState::PHI:
                    numSwPH++;
                    return;
                case RA_Result_Fer::StatsState::PI:
                    numSwPI++;
                    return;
                case RA_Result_Fer::StatsState::PC:
                    numSwPC++;
                    return;
            }
            break;
    }
}


//===--------------------------------------------------------------------------
// Return the register defined by the instruction mi, if any.
//===--------------------------------------------------------------------------
unsigned get_definition(const MachineInstr * mi) {
    for (unsigned i = 0, e = mi->getNumOperands(); i != e; ++i) {
        const MachineOperand & mo = mi->getOperand(i);
        if (mo.isRegister() && mo.isDef()) {
            return mo.getReg();
        }
    }
    return 0;
}

//===--------------------------------------------------------------------------
// This method go through all the instructions in mbb that hold virtual
// variables, and changed them into physical registers. If mbbi define
// variables that have been spilled, then the method inserts a store.
//===--------------------------------------------------------------------------
void RA_Result_Fer::hard_code_bindings(
    MachineBasicBlock & mbb,
    const MRegisterInfo & ri
) {
    MachineBasicBlock::iterator mbbi, mbbe;
    for(mbbi = mbb.begin(), mbbe = mbb.end(); mbbi != mbbe; mbbi++) {
        MachineInstr * mi = mbbi;
        if(mi->getOpcode() == TargetInstrInfo::PHI) {
            continue;
        }

        // get the instruction index:
        unsigned num = LiveIntervals::InstrSlots::NUM;
        unsigned m_index = this->li_->getInstructionIndex(mi);
        unsigned idx = (int)(m_index / num);
 
        // Implement the list of events for this instruction, if any is given:
        this->hard_code_event_list
                         (this->instr_event_list_[idx], mbb, & mbbi, ri);

        // Insert loads and stores, if necessary:
        bool addv = false;
        unsigned v_def = get_definition(mi);

        // I am using 'last_reload' to hold the last register reloaded due
        // to a load lowering optimization. It is not necessary to use a
        // vector because most of the instructions use at most two
        // register operands.
        unsigned last_reload = 0;

        for(unsigned u = 0; u < mi->getNumOperands(); u++) {
            MachineOperand & mo = mi->getOperand(u);
            if(
                mo.isRegister() &&
                mo.getReg() &&
                MRegisterInfo::isVirtualRegister(mo.getReg())
            ) {
                unsigned v_reg = mo.getReg();
                const Binding_Fer & bf =
                                  * get_event(v_reg, this->binding_list_[idx]);
                unsigned p_reg = bf.p_reg;
                mo.setReg(p_reg);
                if(this->rm_->is_spill(v_reg)) {
                    const TargetRegisterClass * rc =
                                 this->mf_->getSSARegMap()->getRegClass(v_reg);
                    unsigned slot = this->rm_->get_stack_slot(v_reg);
                    if(
                        mo.isDef() &&
                        (! this->rm_->is_defined_in_2addr_instr(v_reg) ||
                        X86PuzzleBoard_Fer::is_two_addr(mi))
                    ) {
                        // Because the convertion of three-address into
                        // two-address some variables are defined twice, e.g:
                        // a=b, a = a + c In this case, if 'a' is spilled, it
                        // must be stored only after the second instruction.
//std::cout << "Storing (" << v_reg << ", " << p_reg << ") at block " << mbb.getNumber() << " right after " << NeatPrint_Fer::mi2string(*mi) << std::endl; 
                        ri.storeRegToStackSlot
                                        (mbb, next(mbbi), p_reg, slot, rc);
                        // Because the store is inserted after mbbi, it is
                        // necessary to increment, otherwise the next increment
                        // will point to the store just created.
                        addv = true;
                    }
                    if(
                        mo.isUse() &&
                        load_local(v_reg) &&
                        v_reg != v_def &&
                        v_reg != last_reload
                    ) {
                        ri.loadRegFromStackSlot(mbb, mbbi, p_reg, slot, rc);
                        last_reload = v_reg;
//std::cout << "Loading (" << v_reg << ", " << p_reg << ") at block " << mbb.getNumber() << " right before " << NeatPrint_Fer::mi2string(*mi) << std::endl; 
                    }
                }
            }
        }
        if(addv) {
            mbbi++;
        }
    }
}
