//===---- RegAllocPuzzles.cpp - Register Allocation by Puzzle Solving -----===//
//
//                     The LLVM Compiler Infrastructure
//
// This file was developed by the LLVM research group and is distributed under
// the University of Illinois Open Source License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This file implements Register Allocation by Puzzle Solving.
//
//===----------------------------------------------------------------------===//

#define DEBUG_TYPE "regalloc"

#include "llvm/CodeGen/Passes.h"
#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/CodeGen/MachineInstr.h"
#include "llvm/CodeGen/MachineFrameInfo.h"
#include "llvm/CodeGen/RegAllocRegistry.h"
#include "llvm/CodeGen/MachineFunctionPass.h"
#include "llvm/CodeGen/LiveIntervalAnalysis.h"

#include "llvm/Target/TargetMachine.h"
#include "llvm/Target/TargetInstrInfo.h"

//#include "llvm/Support/Debug.h"
#include "llvm/Support/Compiler.h"

#include "llvm/ADT/STLExtras.h"

// Files included from the c++ library
#include <iostream>

// Files included from the fernando directory
#include "llvm/fernando/CriticalEdgeRemoval_Fer.h"
#include "llvm/fernando/SplitPhiGroups_Fer.h"
#include "llvm/fernando/FillEmptyMBB_Fer.h"
#include "llvm/fernando/NeatPrint_Fer.h"
#include "llvm/fernando/RA_Result_Fer.h"
#include "llvm/fernando/Partition_Fer.h"
#include "llvm/fernando/SSAElim_Fer.h"

using namespace llvm;

namespace llvm {

  /// This variable has an extern declaration in AllocOptions_Fer
  bool SSA_based;

}

namespace {
  cl::opt<bool, true> JoinIntervals("SSA-based",
    cl::desc("Remove phi-functions after register allocation."),
    cl::location(SSA_based), cl::init(true));

  static RegisterRegAlloc
    puzzleRegAlloc("puzzles", "  Register allocation by puzzle solving",
                                               createPuzzlesRegisterAllocator);

  class VISIBILITY_HIDDEN RegAllocPuzzles : public MachineFunctionPass {
  public:
    static char ID;
    RegAllocPuzzles() : MachineFunctionPass((intptr_t)&ID) {}

    virtual const char *getPassName() const {
      return "Register Allocator by Puzzle Solving";
    }

    /// runOnMachineFunction - Register allocate the whole function
    bool runOnMachineFunction(MachineFunction & mf);

    virtual void getAnalysisUsage(AnalysisUsage &AU) const {
      AU.addRequiredTransitive<LiveIntervals>();
      AU.addRequired<Partition_Fer>();
      AU.addRequired<SSAElim_Fer>();
      AU.addPreserved<FillEmptyMBB_Fer>();
      AU.addPreserved<SplitPhiGroups_Fer>();
      AU.addPreserved<CriticalEdgeRemoval_Fer>();
      MachineFunctionPass::getAnalysisUsage(AU);
    }
  };
  char RegAllocPuzzles::ID = 0;
}


/// runOnMachineFunction - Register allocate the whole function
///
bool RegAllocPuzzles::runOnMachineFunction(MachineFunction & mf) {
    assert(SSA_based && "Must enable SSA-based=true to use puzzle solver.");
//NeatPrint_Fer::print_cfg(mf);
    SSAElim_Fer * ssa_ra =  & getAnalysis<SSAElim_Fer>();
    RA_Result_Fer * ra_map = ssa_ra->get_RA_result();
    ra_map->hard_code_register_mapping();
//NeatPrint_Fer::print_cfg(mf);
    return true;
}

FunctionPass *llvm::createPuzzlesRegisterAllocator() {
  return new RegAllocPuzzles();
}
