//===----------- RegMapping_Fer.cpp - Bank of Machine Registers -----------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This class represents the bank of machine registers for register allocation
// purposes.
//
// Date: May 27th, 2006
//
//===----------------------------------------------------------------------===//
//
// This class represents the bank of machine registers for register allocation
// purposes. Essentially, it marks which physical registers are being used,
// and which ones are free to use. For each physical register in use, it keeps
// track of which virtual registers they use.
//
//===----------------------------------------------------------------------===//

#define DEBUG_TYPE "reg_map"

#include "llvm/Instructions.h"                   // MachineOperand, etc
#include "llvm/CodeGen/MachineFrameInfo.h"
#include "llvm/CodeGen/MachineFunctionPass.h"    // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/CodeGen/SSARegMap.h"

#include "llvm/ADT/Statistic.h"

#include "llvm/Support/Debug.h"

#include "llvm/fernando/Line_Fer.h"
#include "llvm/fernando/Regs_Fer.h"
#include "llvm/fernando/NeatPrint_Fer.h"
#include "llvm/fernando/DefUseSites_Fer.h"
#include "llvm/fernando/PhiAnalysis_Fer.h"

#include "llvm/fernando/RegMapping_Fer.h"

using namespace llvm;

STATISTIC(StackSize, "Stack size.");

//===--------------------------------------------------------------------------
// Constructor method. Initializes the data structures. The machine function
// parameter is only used to provide initializing information, such as the
// number of virtual and physical registers.
//===--------------------------------------------------------------------------
RegMapping_Fer::RegMapping_Fer(
    MachineFunction & mf,
    PhiAnalysis_Fer * pa
) :
    machine_function_(& mf),
    phi_analysis_(pa)
{
    // The same color should be assigned to virtuals in the same family
    // everytime that is possible. With this purpose, we use the preference
    // array, whose use is to transfer coloring information from one KP3 to
    // the next:
    const SSARegMap * srm = mf.getSSARegMap();
    unsigned last_v_reg = srm->getLastVirtReg();
    unsigned num_v_regs = last_v_reg - MRegisterInfo::FirstVirtualRegister + 1;

    this->v2p_.resize(num_v_regs, RF_UNDEF);

    this->last_location_.resize(num_v_regs, RF_UNDEF);

    this->v2s_.resize(num_v_regs, RF_UNDEF);

    this->is_spill_.resize(num_v_regs, false);

    this->is_defined_in_2addr_.resize(num_v_regs, false);

    // Initialize the register bank:
    const TargetMachine & target_machine = mf.getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    unsigned num_p_regs = reg_info->getNumRegs();
    this->p2v_.resize(num_p_regs, RF_UNDEF);
}


//===--------------------------------------------------------------------------
// Destructor method.
//===--------------------------------------------------------------------------
RegMapping_Fer::~RegMapping_Fer() {
    this->p2v_.clear();
    this->v2p_.clear();
    this->is_spill_.clear();
}


//===--------------------------------------------------------------------------
// This method returns true if v_reg is defined in a two address instruction,
// and false otherwise.
//===--------------------------------------------------------------------------
bool RegMapping_Fer::is_defined_in_2addr_instr(unsigned v_reg) const {
    assert(MRegisterInfo::isVirtualRegister(v_reg) &&
         "Only virtual registers can be defined in two address instructions.");
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    return this->is_defined_in_2addr_[base_reg];
}


//===--------------------------------------------------------------------------
// This method marks v_reg as defined in a two address instruction. Therefore,
// v_reg is defined twice: v_reg = x; v_reg = v_reg + y;
//===--------------------------------------------------------------------------
void RegMapping_Fer::set_defined_in_2addr_instr(unsigned v_reg) {
    assert(MRegisterInfo::isVirtualRegister(v_reg) &&
         "Only virtual registers can be defined in two address instructions.");
    unsigned b_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    this->is_defined_in_2addr_[b_reg] = true;
}


//===--------------------------------------------------------------------------
// Returns true if the given virtual has been spilled.
//===--------------------------------------------------------------------------
bool RegMapping_Fer::is_spill(unsigned v_reg) const {
    assert(MRegisterInfo::isVirtualRegister(v_reg) &&
                                     "Only virtual registers can be spilled.");
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    return this->v2s_[base_reg] != RF_UNDEF;
}


//===--------------------------------------------------------------------------
// This method makes the physical register p_reg available for allocating
// virtuals. If p_reg is not a physical register, an error is issued.
//===--------------------------------------------------------------------------
void RegMapping_Fer::free_color(unsigned reg) {
//std::cerr << "Liberating " << reg << "\n";
    if(MRegisterInfo::isVirtualRegister(reg)) {
        unsigned base_reg = reg - MRegisterInfo::FirstVirtualRegister;
        unsigned p_reg = this->v2p_[base_reg];
        if(p_reg != RF_UNDEF) {
            if(this->p2v_[p_reg] == reg) {
                this->p2v_[p_reg] = RF_UNDEF;
            }
        }
    } else {
        this->p2v_[reg] = RF_UNDEF;
    }
}


//===--------------------------------------------------------------------------
// This method marks v_reg as a spilled variable. A stack slot will be assigned
// to v_reg.
//===--------------------------------------------------------------------------
void RegMapping_Fer::set_spill(unsigned v_reg) {
    // Only virtual variables can be spilled:
    assert(MRegisterInfo::isVirtualRegister(v_reg) &&
                                     "Only virtual registers can be spilled.");
    unsigned b_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    bool is_spill = this->is_spill_[b_reg];
    this->is_spill_[b_reg] = true;
    this->v2p_[b_reg] = RF_SPILL;

    if(!is_spill) {
        // Find and assign a stack slot to v_reg.
        int slot;
        if(this->phi_analysis_->has_stack_slot(v_reg)) {
            slot = this->phi_analysis_->get_stack_slot(v_reg);
        } else {
            const TargetRegisterClass * rc =
                   this->machine_function_->getSSARegMap()->getRegClass(v_reg);
            slot = this->machine_function_->getFrameInfo()->
                          CreateStackObject(rc->getSize(), rc->getAlignment());
            StackSize++;
            this->phi_analysis_->map_to_stack_slot(v_reg, slot);
        }
        this->v2s_[b_reg] = slot;
    }
}


//===--------------------------------------------------------------------------
// This method spill the physical register associated to v_reg. If there are
// different virtuals sharing aliases of the register that must be freed, then
// spill all those virtuals.
//===--------------------------------------------------------------------------
void RegMapping_Fer::free_register(
    unsigned p_reg,
    std::vector<unsigned> & sp_regs
) {
//std::cout << "Liberating " << p_reg << std::endl;
    assert(MRegisterInfo::isPhysicalRegister(p_reg) &&
                        "Only physical registers can be freed from virtuals.");
    // Free the virtual register that is directly assigned to p_reg, if any:
    unsigned v_reg = this->p2v_[p_reg];
    if(v_reg != RF_UNDEF && v_reg != RF_SPILL) {
        free_color(v_reg);
        set_spill(v_reg);
        sp_regs.push_back(p_reg);
    }
    // Go through the aliases:
    const TargetMachine & target_machine = machine_function_->getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    for (const unsigned *a = reg_info->getAliasSet(p_reg); *a; ++a) {
        v_reg = this->p2v_[*a];
        if(v_reg != RF_UNDEF && v_reg != RF_SPILL) {
            free_color(v_reg);
            set_spill(v_reg);
            sp_regs.push_back(*a);
        }
    }
}


//===--------------------------------------------------------------------------
// This method returns the number of aliases of p_reg that are being in use.
//===--------------------------------------------------------------------------
unsigned RegMapping_Fer::number_of_occupied_aliases(unsigned p_reg) const {
    const TargetMachine & target_machine = machine_function_->getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    unsigned counter = 0;
    unsigned v_reg = this->p2v_[p_reg];
    if(v_reg != RF_UNDEF && v_reg != RF_SPILL) {
        counter++;
    }
    for (const unsigned *a = reg_info->getAliasSet(p_reg); *a; ++a) {
        v_reg = this->p2v_[*a];
        if(v_reg != RF_UNDEF && v_reg != RF_SPILL) {
            counter++;
        }
    }
    return counter;
}


//===--------------------------------------------------------------------------
// This method modifies the vector aliases. For each physical register p_reg
// in regs, it will insert the value 'p_reg' in every cell of aliases that
// represents a physical register that is an alias of p_reg and is being
// used.
//===--------------------------------------------------------------------------
void RegMapping_Fer::fill_aliases(
    std::vector<unsigned> & regs,
    std::vector<unsigned> & aliases
) const {
    aliases.clear();
    aliases.resize(this->p2v_.size(), RF_UNDEF);

    const TargetMachine & target_machine = machine_function_->getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    for(unsigned u = 0; u < regs.size(); u++) {
        unsigned p_reg = regs[u];
        unsigned v_reg = this->p2v_[p_reg];
        if(v_reg != RF_UNDEF && v_reg != RF_SPILL) {
            aliases[p_reg] = p_reg;
        }
        for (const unsigned *a = reg_info->getAliasSet(p_reg); *a; ++a) {
            v_reg = this->p2v_[*a];
            if(v_reg != RF_UNDEF && v_reg != RF_SPILL) {
                aliases[*a] = p_reg;
            }
        }
    }
}


//===--------------------------------------------------------------------------
// This method assigns RF_UNDEF to the virtual v_reg.
//===--------------------------------------------------------------------------
void RegMapping_Fer::reset_color(unsigned v_reg) {
//std::cout << "RegMapping_Fer::reset_color(" << v_reg << ")\n";
    assert(MRegisterInfo::isVirtualRegister(v_reg) &&
                                      "Target variable is not valid virtual.");
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    unsigned p_reg = this->v2p_[base_reg];
    this->v2p_[base_reg] = RF_UNDEF;
    if(MRegisterInfo::isPhysicalRegister(p_reg) && p2v_[p_reg] == v_reg) {
        this->p2v_[p_reg] = RF_UNDEF;
    }
}


//===--------------------------------------------------------------------------
// This method stores the values in the 'preference' vector as the last
// registers in use at the end of the basic block 'b'.
//===--------------------------------------------------------------------------
void RegMapping_Fer::store_v2p(
    std::map<unsigned, unsigned> & map,
    std::set<Line_Fer * , LineCompareEnd > & live_set
) const {
    std::set<Line_Fer *, LineCompareEnd>::iterator e;
    std::set<Line_Fer *, LineCompareEnd>::iterator i;
    for(i = live_set.begin(), e = live_set.end(); i != e; i++) {
        Line_Fer *lf = *i;
        unsigned v_reg = lf->reg;
        assert(MRegisterInfo::isVirtualRegister(v_reg) &&
                   "Only virtual registers can be alive at block boundaries.");
        unsigned loc = this->get_physical_location(v_reg);
        if(loc != RF_UNDEF) {
            map[v_reg] = loc;
        }
    }
}


//===--------------------------------------------------------------------------
// This method clears the mapping between virtuals and physical registers.
// It does not clear the last physical assigned to each virtual, neither the
// memory slot of each virtual.
//===--------------------------------------------------------------------------
void RegMapping_Fer::clear_mapping() {
    // Clean the mapping from virtuals to physicals
    unsigned num_v_regs = this->v2p_.size();
    this->v2p_.assign(num_v_regs, RF_UNDEF);

    // Clean the mapping from virtuals to memory locations
    this->is_spill_.assign(num_v_regs, false);

    // Clean the mapping from physicals to virtuals
    unsigned num_p_regs = this->p2v_.size();
    this->p2v_.assign(num_p_regs, RF_UNDEF);
}


//===--------------------------------------------------------------------------
// This method cleans the current register mapping in this object, and fills
// it with the mapping given in the "map" parameter.
//===--------------------------------------------------------------------------
void RegMapping_Fer::reset_mapping(
    const std::map<unsigned, unsigned> & map,
    std::set<Line_Fer * , LineCompareEnd > & live_set
) {
    // Clean the mapping from virtuals to physicals
    this->clear_mapping();

    // Go through the live set and recover the live intervals and retrive the
    // color of them:
    std::set<Line_Fer *, LineCompareEnd>::iterator e;
    std::set<Line_Fer *, LineCompareEnd>::iterator i;
    for(i = live_set.begin(), e = live_set.end(); i != e; i++) {
        Line_Fer *lf = *i;
        unsigned reg = lf->reg;
        assert(MRegisterInfo::isVirtualRegister(reg) &&
                   "Only virtual registers can be alive at block boundaries.");
        std::map<unsigned, unsigned>::const_iterator iloc = map.find(reg);
        unsigned loc = iloc->second;
        unsigned b_reg = reg - MRegisterInfo::FirstVirtualRegister;
        if(loc == RF_SPILL) {
            this->set_spill(reg);
        } else {
if(!MRegisterInfo::isPhysicalRegister(loc)) {
    std::cout << "The Location "  << loc << " assigned to " << reg << " is neither memory nor physical register\n";
}
            assert(MRegisterInfo::isPhysicalRegister(loc) &&
                           "Location is neither memory nor physical register");
            this->v2p_[b_reg] = loc;
            this->p2v_[loc] = b_reg + MRegisterInfo::FirstVirtualRegister;
            this->machine_function_->setPhysRegUsed(loc);
        }
    }
/*
std::cout << "RESETING PREFERENCES:\n";
for(unsigned u = 0; u < this->p2v_.size(); u++) {
    if(this->p2v_[u] != RF_UNDEF) {
        std::cout << this->p2v_[u] << ":" << u << " ";
    }
}
std::cout << "\n";
*/
}


//===--------------------------------------------------------------------------
// This method returns a string in the format "t1(r1), t2(r2), ... tm(rm),
// where virtual registers ti are allocated in machine registers ri.
//===--------------------------------------------------------------------------
std::string RegMapping_Fer::virtual_to_string() const {
    std::ostringstream aux;
    bool is_first = true;
    const TargetMachine & target_machine = this->machine_function_->getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
   for(unsigned i = 0; i < this->v2p_.size(); i++) {
        if(this->v2p_[i] != RF_UNDEF) {
            const char * phys_reg_name;
            if(this->v2p_[i] != RF_SPILL) {
                phys_reg_name = reg_info->getName(this->v2p_[i]);
            } else {
                phys_reg_name = "[MEM]";
            }
            if(is_first) {
                aux << "t" << i << "(" << phys_reg_name << ")";
                is_first = false;
            } else {
                aux << ", " << "t" << i << "(" << phys_reg_name << ")";
            }
        }
    }
    return aux.str();
}


//===--------------------------------------------------------------------------
// This method returns a string in the format "r1(t1), r2(t2), ... rm(tm),
// where physical registers ri is holding the value of virtual register ti.
//===--------------------------------------------------------------------------
std::string RegMapping_Fer::physical_to_string() const {
    std::ostringstream aux;
    bool is_first = true;
    for(unsigned i = 0; i < this->p2v_.size(); i++) {
        if(this->p2v_[i] != RF_UNDEF) {
            unsigned reg = this->p2v_[i];
            std::ostringstream reg_name;
            reg_name << NeatPrint_Fer::op2string(reg, *this->machine_function_);
            std::ostringstream phys_name;
            phys_name << NeatPrint_Fer::op2string(i, *this->machine_function_);
            std::ostringstream num_name;
            if(is_first) {
                aux << phys_name.str() << "[" << reg_name.str() << "]";
                is_first = false;
            } else {
                aux << ", " << phys_name.str() << "[" << reg_name.str() << "]";
            }
        }
    }
    return aux.str();
}


//===--------------------------------------------------------------------------
// This method sets p_reg as the physical location of reg, which can be
// either a virtual variable or a physical register itself, due to
// precoloring. If reg is a physical register, we require that reg==p_reg
// or an exception would be thrown.
//===--------------------------------------------------------------------------
void RegMapping_Fer::allocate_color(unsigned reg, unsigned p_reg) {
//std::cout << "allocating " << reg << " into " << p_reg << "\n";

    // Remove any previous color that could exist in the aliases of p_reg
    const TargetMachine & target_machine = this->machine_function_->getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    for (const unsigned *a = reg_info->getAliasSet(p_reg); *a; ++a) {
        this->p2v_[*a] = RF_UNDEF;
    }

    // Binds p_reg to reg:
    assert(MRegisterInfo::isPhysicalRegister(p_reg) &&
                                           "Target register is not physical.");
    if(MRegisterInfo::isVirtualRegister(reg)) {
        unsigned base_reg = reg - MRegisterInfo::FirstVirtualRegister;
        this->v2p_[base_reg] = p_reg;
        this->last_location_[base_reg] = p_reg;
        this->p2v_[p_reg] = reg;
        this->phi_analysis_->set_preference(reg, p_reg);
    } else {
        assert(reg == p_reg &&
                    "Attempt to assign incorrect color to physical register.");
        this->p2v_[p_reg] = reg;
    }

    // Records that p_reg has ever been used to allocate a register. This info
    // is necessary in order to insert correct epilogue/prologue into the code.
    this->machine_function_->setPhysRegUsed(p_reg);
}


//===--------------------------------------------------------------------------
// This method returns the last physical register that has been allocated to
// v_reg.
//===--------------------------------------------------------------------------
unsigned RegMapping_Fer::get_last_location(unsigned v_reg) const {
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    assert(MRegisterInfo::isPhysicalRegister(this->last_location_[base_reg]) &&
                           "Last location of v_reg is not physical register.");
    return this->last_location_[base_reg];
}


//===--------------------------------------------------------------------------
// This method returns the physical register that was assigned to the virtual
// register. If the virtual register has not been allocated to any register, it
// will return null. It is also possible that the given parameter describes a
// physical register. In this case, it is a pre-colored register, and its
// location must be the physical register itself.
//===--------------------------------------------------------------------------
unsigned RegMapping_Fer::get_physical_location(unsigned reg) const {
    if(MRegisterInfo::isVirtualRegister(reg)) {
        unsigned base_reg = reg - MRegisterInfo::FirstVirtualRegister;
        assert(base_reg < this->v2p_.size() && "Invalid virtual register.");
        return this->v2p_[base_reg];
    } else {
        assert(MRegisterInfo::isPhysicalRegister(reg) && "Unknown value.");
        return reg;
    }
}


//===--------------------------------------------------------------------------
// This method returns the virtual register that was assigned to the given
// physical register. If the physical register does not contain any virtual
// register, the method will return RF_UNDEF.
//===--------------------------------------------------------------------------
unsigned RegMapping_Fer::get_assigned_data(unsigned p_reg) const {
    assert(p_reg < this->p2v_.size() && "Invalid physical register.");
    unsigned data = this->p2v_[p_reg];
    if(data == RF_UNDEF) {
        const TargetMachine & target_machine = machine_function_->getTarget();
        const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
        for (const unsigned *a = reg_info->getAliasSet(p_reg); *a; ++a) {
            if (this->p2v_[*a] != RF_UNDEF) {
                data = this->p2v_[*a];
                break;
            }
        }
    }   
    return data;
}


//===--------------------------------------------------------------------------
// This method scans the vector of available registers and returns a register
// that can be assigned to 'v_reg'. It will return RF_UNDEF if there is no
// register available.
//===--------------------------------------------------------------------------
unsigned RegMapping_Fer::get_free_register(unsigned v_reg) const {
    const TargetRegisterClass *trc =
                   this->machine_function_->getSSARegMap()->getRegClass(v_reg);
    TargetRegisterClass::iterator trci =
                         trc->allocation_order_begin(*this->machine_function_);
    TargetRegisterClass::iterator trce =
                           trc->allocation_order_end(*this->machine_function_);

    for (; trci != trce; ++trci) {
        if (is_register_available(*trci)) {
            return *trci;
        }
    }

    return RF_UNDEF;
}


//===--------------------------------------------------------------------------
// This method determines if p_reg is not currently allocated to any virtual
// register. In order to verify this, it does not suffice to check if the
// cell of p_reg in the physical register table contains the RF_UNDEF value,
// because physical registers contain alias. For instance, in the X86
// architecture, the physical registers AL, AX and EAX are all alias, and if
// one of them is in use, the others are not free. If neither p_reg nor any of
// its aliases is been used to allocate virtuals, the method returns true. It
// returns false otherwise.
//===--------------------------------------------------------------------------
bool RegMapping_Fer::is_register_available(unsigned p_reg) const {
    assert(MRegisterInfo::isPhysicalRegister(p_reg) &&
                                           "Target register must be physical");

    if (this->p2v_[p_reg] != RF_UNDEF) {
        return false;
    }

    const TargetMachine & target_machine = this->machine_function_->getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();

    for (const unsigned *a = reg_info->getAliasSet(p_reg); *a; ++a) {
        if (this->p2v_[*a] != RF_UNDEF) {
            return false;
        }
    }

    return true;
}


//===--------------------------------------------------------------------------
// Clean the arrays used in this object.
//===--------------------------------------------------------------------------
void RegMapping_Fer::releaseMemory() {
    this->p2v_.clear();
    this->v2p_.clear();
    this->is_spill_.clear();
    this->v2s_.clear();
}


//===--------------------------------------------------------------------------
// Return the stack slot that has been assigned to the virtual
//===--------------------------------------------------------------------------
unsigned RegMapping_Fer::get_stack_slot(unsigned v_reg) const {
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    assert(this->v2s_[base_reg] != RF_UNDEF &&
                       "Attempt to get stack location of non-spilled virtual");
    return this->v2s_[base_reg];
}


//===--------------------------------------------------------------------------
// This method prints the internal state of this register mapping, which
// includes the bank of physical registers plus the memory.
//===--------------------------------------------------------------------------
void RegMapping_Fer::dump() const {
    // First, print the variables that are currently in the register bank:
    const TargetMachine & target_machine = this->machine_function_->getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    std::cout << "RegMap: ";
    for(unsigned u = 0; u < this->p2v_.size(); u++) {
        unsigned reg = this->p2v_[u];
        if(reg != RF_UNDEF) {
            std::ostringstream str1;
            str1 << reg << ":" << reg_info->getName(u);
            std::cout.width(10);
            std::cout << str1.str();
        }
    }

    // Print the spilled variables:
    for(unsigned u = 0; u < this->is_spill_.size(); u++) {
        if(this->is_spill_[u]) {
            unsigned reg = u + MRegisterInfo::FirstVirtualRegister;
            std::ostringstream tr2;
            tr2 << reg << ":MEM";
            std::cout.width(10);
            std::cout << tr2.str();
        }
    }

    std::cout << std::endl;
}


//===--------------------------------------------------------------------------
// This method returns the virtual register that is the next to be used among
// the virtual registers stored in aliases of p_reg.
//===--------------------------------------------------------------------------
unsigned RegMapping_Fer::next_nearest_alias_used(
    unsigned p_reg,
    DefUseSites_Fer * du
) const {
    assert(p_reg < this->p2v_.size() && "Invalid physical register.");

    unsigned best_cndt = RF_UNDEF;
    unsigned min_index = ~0U;

    // Get the index of p_reg
    unsigned data = this->p2v_[p_reg];
    if(data != RF_UNDEF) {
        min_index = du->next_usage(data);
    }

    const TargetMachine & target_machine = machine_function_->getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    for (const unsigned *a = reg_info->getAliasSet(p_reg); *a; ++a) {
        if (this->p2v_[*a] != RF_UNDEF) {
            data = this->p2v_[*a];
            unsigned aux_index = du->next_usage(data);
            if(aux_index < min_index) {
                min_index = aux_index;
                best_cndt = data;
            }
        }
    }
    return best_cndt;
}

//===--------------------------------------------------------------------------
// This method prints in the standard output all the virtuals that are
// spilled, plus their stack position.
//===--------------------------------------------------------------------------
void RegMapping_Fer::print_stack_allocation() const {
    std::cout << "Stack configuration:\n";
    for(unsigned u = 0; u < this->is_spill_.size(); u++) {
        if(this->v2s_[u] != RF_UNDEF) {
            unsigned v_reg = u + MRegisterInfo::FirstVirtualRegister;
            std::cout << v_reg << " => [" << this->v2s_[u] << "]\n";
        }
    }
}


//===--------------------------------------------------------------------------
// This method prints a list with all the variables that have been spilled.
//===--------------------------------------------------------------------------
void RegMapping_Fer::print_store_list() const {
    std::cout << "List of stores to be inserted into the code:\n";

    for(unsigned u = 0; u < this->is_spill_.size(); u++) {
        if(this->v2s_[u] != RF_UNDEF) {
            unsigned v_reg = u + MRegisterInfo::FirstVirtualRegister;
            Regs_Fer::print_reg(this->v2s_[u] + 100000, v_reg);
            std::cout << " = ( ???, P" << v_reg << ")";
            std::cout << " ;\n";
        }
    }
}
