//===-------------------- RevVec_Fer.cpp - PHI Matrix ---------------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This class implements the reverse vector data structure.
//
// Date: Jun 22, 2007
//
//===----------------------------------------------------------------------===//
//
// A reverse vector provides a "contains" operation that is O(1). The only
// requirement is that the maximum element stored in the vector must be known.
// This class only stores unsigned ints.
//
//===----------------------------------------------------------------------===//

#include <map>
#include <set>
#include <vector>
#include <sstream>                               // ostringstream (RegClique)
#include <iostream>

#include "llvm/Target/MRegisterInfo.h"

#include "llvm/fernando/Line_Fer.h"
#include "llvm/fernando/RegMapping_Fer.h"

#include "llvm/fernando/RevVec_Fer.h"

using namespace llvm;


//===--------------------------------------------------------------------------
// Constructor method.
// This method receives two parameters. The first is a mapping of virtual
// registers to locations. The virtual registers will be used to order the
// vector containing the physical registers.
//===--------------------------------------------------------------------------
RevVec_Fer::RevVec_Fer(
    unsigned size_phi_row,
    unsigned max_elem,
    const RegMapping_Fer * rm
) : rm_(rm)
{
    // Reserve memory for data structures:
    this->phi_rows_.resize(size_phi_row, RV_UNDEF);
    this->vrt_regs_.resize(size_phi_row, RV_UNDEF);
    this->phy_regs_.resize(max_elem, RV_UNDEF);
}


//===--------------------------------------------------------------------------
// This method returns true if this vector contains, either the physical
// register loc, or one of its alias. If loc is not a physical register, or if
// neither loc nor any of its alias is in the reverse vector, then the method
// returns false;
//===--------------------------------------------------------------------------
bool RevVec_Fer::contains_reg_or_alias(
    unsigned loc,
    const MRegisterInfo & reg_info
) const {
    if(loc < this->phy_regs_.size()) {
        // First, check if loc itself is used.
        if(this->phy_regs_[loc] != RV_UNDEF) {
            return true;
        }
        // loc is not being used, but maybe one of its aliases is:
        for (const unsigned *a = reg_info.getAliasSet(loc); *a; ++a) {
            if (this->phy_regs_[*a] != RV_UNDEF) {
                return true;
            }
        }
    }
    return false;
}


//===--------------------------------------------------------------------------
// Swaps the contents of registers r1 and r2.
//===--------------------------------------------------------------------------
void RevVec_Fer::swap(unsigned r1, unsigned r2) {
    assert(MRegisterInfo::isPhysicalRegister(r1) &&
                                      "Only physical registers can be swaped");
    assert(MRegisterInfo::isPhysicalRegister(r2) &&
                                      "Only physical registers can be swaped");
    unsigned index1 = this->phy_regs_[r1];
    unsigned index2 = this->phy_regs_[r2];
    assert(index1 != RV_UNDEF && index2 != RV_UNDEF &&
                          "Registers have no location in the reverse vector.");
    // swap the phi_rows:
    this->phi_rows_[index1] = r2;
    this->phi_rows_[index2] = r1;

    // swap the reverse index:
    this->phy_regs_[r1] = index2;
    this->phy_regs_[r2] = index1;
}


//===--------------------------------------------------------------------------
// This method replaces the old physical register with the new one.
//===--------------------------------------------------------------------------
void RevVec_Fer::replace_phys(unsigned old_reg, unsigned new_reg) {
    assert(old_reg < phy_regs_.size() && new_reg < phy_regs_.size());
    unsigned index = phy_regs_[old_reg];
    assert(index < this->vrt_regs_.size() && "Invalid index");
    this->phy_regs_[new_reg] = index;
    this->phi_rows_[index] = new_reg;
    this->phy_regs_[old_reg] = RV_UNDEF;
}


//===--------------------------------------------------------------------------
// Removes the location at the given index from this object, and returns the
// location stored at the removed index.
//===--------------------------------------------------------------------------
unsigned RevVec_Fer::remove(unsigned index) {
    assert(index < this->phi_rows_.size() && "Invalid index");
    unsigned old_data = phi_rows_[index];
    phi_rows_[index] = RV_UNDEF;
    vrt_regs_[index] = RV_UNDEF;
    if(MRegisterInfo::isPhysicalRegister(old_data)) {
        phy_regs_[old_data] = RV_UNDEF;
    }
    return old_data;
}


//===--------------------------------------------------------------------------
// Inserts loc at the "index"  position of the phi-vector. "loc" is the
// location of v_reg.
//===--------------------------------------------------------------------------
void RevVec_Fer::push(unsigned v_reg, unsigned loc, unsigned index) {
//std::cout << "push: " << v_reg << ", " << loc << ", " << index << std::endl;
    assert(index < this->phi_rows_.size() && "Invalid index");
    if(MRegisterInfo::isPhysicalRegister(loc)) {
        this->phy_regs_[loc] = index;
        this->phi_rows_[index] = loc;
    } else {
        // I am adding FVR only to make it easy to distinguish these locations
        // from virtual registers.
        this->phi_rows_[index] = this->rm_->get_stack_slot(v_reg)
                                         + MRegisterInfo::FirstVirtualRegister;
    }
    this->vrt_regs_[index] = v_reg;
}


void print_vec(const std::vector<unsigned> & v) {
    for(unsigned u = 0; u < v.size(); u++) {
        std::cout.width(16);
        std::cout << v[u];
    }
    std::cout << std::endl;
}


//===--------------------------------------------------------------------------
// Prints the information of this object.
//===--------------------------------------------------------------------------
void RevVec_Fer::dump(const MRegisterInfo * reg_info) const {
    for(unsigned u = 0; u < this->phi_rows_.size(); u++) {
        std::ostringstream str;
        unsigned loc = this->phi_rows_[u];
        if(MRegisterInfo::isPhysicalRegister(loc)) {
            str << this->vrt_regs_[u] << ":" << reg_info->getName(loc) << ":"
                                                                        << loc;
        } else {
            str << this->vrt_regs_[u] << ":" << "[" <<
                            (loc - MRegisterInfo::FirstVirtualRegister) << "]";
        }
        std::cout.width(16);
        std::cout << str.str();
    }
    std::cout << std::endl;
}
