//===------------------ SSAElim.h - Phi-elimination Pass ------------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it eliminate phi functions after register allocation has been
// performed.
//
// Date: June 30th, 2006
//
//===----------------------------------------------------------------------===//
//
// Traditional algorithms used to destroy phi functions normally replace them
// by copy instructions. Although the classical algorithms preserve the
// semantics of the SSA-form program, they may increase the number of registers
// found during the register assignment phase. To avoid this problem, we
// replace phi functions by parallel copies. Parallel copies are normally
// implemented with xor instructions (Remember the trick of swaping two
// variables without an auxiliary using xor?). The register swap is a costly
// operation, because it demands three instructions to be implemented;
// therefore, it should be avoided whenever possible. One way of avoiding
// swaps is by means of move instructions. Basically, swap operations are not
// necessary in every time that phi instructions are destroyed. This happens
// when the definition vector contains machine registers not used in the
// parameter vector. This pass tries to use such optimization  before inserting
//  swap operations. It performs phi elimination in time linear in the size of
// the definition vector.
//
//===----------------------------------------------------------------------===//

#include <set>
#include <map>

#include "llvm/Target/TargetMachine.h"
#include "llvm/Target/MRegisterInfo.h"

#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/CodeGen/MachineFunction.h"
#include "llvm/CodeGen/MachineBasicBlock.h"
#include "llvm/CodeGen/MachineFunctionPass.h"

#include "llvm/ADT/STLExtras.h"
#include "llvm/ADT/DepthFirstIterator.h"            // df_ext_iterator

#include "llvm/fernando/Regs_Fer.h"
#include "llvm/fernando/RevVec_Fer.h"
#include "llvm/fernando/NeatPrint_Fer.h"
#include "llvm/fernando/RA_Result_Fer.h"
#include "llvm/fernando/Partition_Fer.h"
#include "llvm/fernando/RegMapping_Fer.h"
#include "llvm/fernando/PhiAnalysis_Fer.h"
#include "llvm/fernando/AllocOptions_Fer.h"
#include "llvm/fernando/X86PuzzleBoard_Fer.h"

#include "llvm/fernando/SSAElim_Fer.h"              // class definition

using namespace llvm;

char SSAElim_Fer::ID = 0;

namespace {

    RegisterPass<SSAElim_Fer>
    X("SSAElim", "SSA elimination phase - Fernando");
}


//===--------------------------------------------------------------------------
// Prints the contents of v.
//===--------------------------------------------------------------------------
void print_vec(std::vector<unsigned> & v) {
    for(unsigned u = 0; u < v.size(); u++) {
        std::cout.width(8);
        std::cout << v[u];
    }
    std::cout << std::endl;
}


//===--------------------------------------------------------------------------
// This method specifies which other passes are required, or are modified
// by this pass.
//===--------------------------------------------------------------------------
void SSAElim_Fer::getAnalysisUsage(AnalysisUsage & AU) const {
    AU.addRequired<PhiAnalysis_Fer>();
    AU.addRequired<Partition_Fer>();
    AU.setPreservesAll();
}


//===--------------------------------------------------------------------------
// Returns true if the element in the c-th column of phi is alive at the
// end of basic block pred.
//===--------------------------------------------------------------------------
bool SSAElim_Fer::is_column_alive(
    const MachineInstr * phi,
    MachineBasicBlock & src_mbb,
    unsigned c
) const {
    // Get the values alive at the end of src_mbb
    const std::map<unsigned, unsigned> & src_map =
                                  this->pf_->get_tail_map(src_mbb.getNumber());
    unsigned v_reg = phi->getOperand(c).getReg();
    std::map<unsigned, unsigned>::const_iterator loc = src_map.find(v_reg);
    return loc != src_map.end();
}


//===--------------------------------------------------------------------------
// This method inserts instructions at all the predecessors of mbb in order
// to transfer colors.
//===--------------------------------------------------------------------------
void SSAElim_Fer::visit(MachineBasicBlock * mbb) {
//std::cout << "Visiting machine basic block " << mbb->getNumber() << std::endl;
    // First, handle basic blocks that can have many predecessors. In this
    // case, the transfering code will be inserted at the end of each
    // predecessor.
    if(mbb->pred_size() >= 2) {
        if(
            mbb->begin() != mbb->end() &&
            mbb->begin()->getOpcode() == TargetInstrInfo::PHI
        ) {
            const MachineInstr * phi = mbb->begin();
            for(unsigned u = phi->getNumOperands() - 1; u >= 2; u -= 2) {
                MachineBasicBlock & pred =
                                   * phi->getOperand(u).getMachineBasicBlock();
                // Due to some LLVM crazyness, it is possible that the same
                // basic block feeds different entries of the phi-matrix. In
                // this case, the phiMemCoalesce transformation will insert
                // some copies that are dead code. We can avoid these dead code
                // using the mapping at the end of the source block.
                if(is_column_alive(phi, pred, u - 1)) {
                    this->insert_transfer_at_end(mbb, &pred, u - 1);
                }
            }
        } else {
            MachineBasicBlock::const_pred_iterator pred_i, pred_e;
            for(pred_i = mbb->pred_begin(), pred_e = mbb->pred_end();
                                                  pred_i != pred_e; pred_i++) {
                MachineBasicBlock * pred_mbb = * pred_i;
                this->insert_transfer_at_end(mbb, pred_mbb, 0);
            }
        }
    }
}


//===--------------------------------------------------------------------------
// This method fills the reverse vector with the elements in the given column
// of the phi-matrix that head the basic block mbb. loc_map contains the
// location of each parameter.
//===--------------------------------------------------------------------------
void insert_phi_parameters(
    MachineBasicBlock * mbb,
    RevVec_Fer & rv,
    unsigned phi_column,
    const std::map<unsigned, unsigned> & loc_map,
    std::set<unsigned> & phi_regs
) {
    unsigned index = 0;
    MachineBasicBlock::const_iterator mbb_i, mbb_e;
    for(mbb_i = mbb->begin(), mbb_e = mbb->end(); mbb_i != mbb_e &&
                         mbb_i->getOpcode() == TargetInstrInfo::PHI; mbb_i++) {
        const MachineInstr * phi = mbb_i;
        unsigned v_reg = phi->getOperand(phi_column).getReg();
        std::map<unsigned, unsigned>::const_iterator loc = loc_map.find(v_reg);
        assert(loc != loc_map.end() && "Parameter of phi-function not alive");
        rv.push(v_reg, loc->second, index);
        phi_regs.insert(v_reg);
        index++;
    }
}


//===--------------------------------------------------------------------------
// This pass inserts loads for the variables that are marked as load local
// by RA_Results. This is a heuristics that aims at reducing the number of
// loads in the target program.
//===--------------------------------------------------------------------------
void SSAElim_Fer::reload_locals(
    RevVec_Fer & src,
    RevVec_Fer & dst,
    MachineBasicBlock * mbb
) {
    for(unsigned u = 0; u < dst.size(); u++) {
        unsigned v_reg_dst = dst.get_virtual(u);
        unsigned v_reg_src = src.get_virtual(u);
        if(v_reg_dst != v_reg_src && this->rm_->is_spill(v_reg_src)) {
            unsigned src_loc = src.get_loc(u);
            unsigned dst_loc = dst.get_loc(u);
            if(
                MRegisterInfo::isPhysicalRegister(src_loc) &&
                MRegisterInfo::isPhysicalRegister(dst_loc)
            ) {
                unsigned slot = this->rm_->get_stack_slot(v_reg_src);
//std::cout << "Inserting MOCK load at mem_" << slot << " (" << v_reg_src << ", " << dst_loc << ") at MBB_" << mbb->getNumber() << std::endl;
                this->rs_->add_mock_load
                       (src_loc, v_reg_dst, slot, v_reg_src, mbb->getNumber());
            }
        }
    }
}


//===--------------------------------------------------------------------------
// This pass removes instances from the parallel copy where the source and the
// destination have the same location.
//===--------------------------------------------------------------------------
void SSAElim_Fer::implement_coalescing(
    RevVec_Fer & src,
    RevVec_Fer & dst,
    MachineBasicBlock * mbb
) {
    for(unsigned u = 0; u < dst.size(); u++) {
        unsigned dst_loc = dst.get_loc(u);
        unsigned src_loc = src.get_loc(u);
        if(src_loc == dst_loc) {
            if(MRegisterInfo::isPhysicalRegister(dst_loc)) {
                // That is the case for inserting a copy that later will be
                // coalesced:
                unsigned v_reg_dst = dst.get_virtual(u);
                unsigned v_reg_src = src.get_virtual(u);
                this->rs_->add_cp2mbb
                  (dst_loc,v_reg_dst,src_loc,v_reg_src,mbb->getNumber(), true);
            }
            src.remove(u);
            dst.remove(u);
        }
    }
}


//===--------------------------------------------------------------------------
// When performing the tranfer of values, it is always safe to perform a store,
// if the destination location is a memory address.
//===--------------------------------------------------------------------------
void SSAElim_Fer::implement_stores(
    RevVec_Fer & src,
    RevVec_Fer & dst,
    MachineBasicBlock * mbb
) {
    for(unsigned u = 0; u < dst.size(); u++) {
        unsigned dst_loc = dst.get_loc(u);
        unsigned src_loc = src.get_loc(u);
        if(!MRegisterInfo::isPhysicalRegister(dst_loc)) {
            if(src_loc != dst_loc) {
                // That is the case for inserting a store:
                unsigned v_reg_src = src.get_virtual(u);
                if(!this->rm_->is_spill(v_reg_src)) {
                    this->rm_->set_spill(v_reg_src);
                }
            }
            src.remove(u);
            dst.remove(u);
        }
    }
}


//===--------------------------------------------------------------------------
// This method fills the vector diff with indexes of dst. Each of these indexes
// correspond to a safe operation that can be performed to transfer values from
// the source vector to the destination vector.
//===--------------------------------------------------------------------------
void SSAElim_Fer::find_safe_operations(
    RevVec_Fer & src,
    RevVec_Fer & dst,
    std::vector<unsigned> & diff
) const {
    const TargetMachine & target_machine = this->mf_->getTarget();
    const MRegisterInfo & reg_info = * target_machine.getRegisterInfo();

    for(unsigned u = 0; u < dst.size(); u++) {
        unsigned dst_loc = dst.get_loc(u);
        unsigned src_loc = src.get_loc(u);
        if(dst_loc == RV_UNDEF) {
            assert(src_loc == RV_UNDEF && "Inconsistent state.");
            continue;
        }
        if(dst_loc == src_loc) {
// BEGIN DEBUG: ***************************************************
/*
unsigned v_reg_dst = dst.get_virtual(u);
unsigned v_reg_src = src.get_virtual(u);
if(v_reg_dst != v_reg_src) {
      Regs_Fer::print_def(*this->mf_, dst_loc, v_reg_dst);
      std::cout << " = ";
      Regs_Fer::print_reg(src_loc, v_reg_src);
      std::cout << " ; // copy (swap) \n";
}
*/
// END DEBUG: *****************************************************
            // This case is an instance of register coalescing.
            src.remove(u);
            dst.remove(u);
        } else {
            assert(MRegisterInfo::isPhysicalRegister(dst_loc) &&
                               "Stores should have been inserted previously.");
            if(! src.contains_reg_or_alias(dst_loc, reg_info)) {
                diff.push_back(u);
            }
        }
    }
}


//===--------------------------------------------------------------------------
// This method inserts copy instructions at the end of the target basic block
// in order to implement safe operations.
//===--------------------------------------------------------------------------
void SSAElim_Fer::implement_safe_operations(
    RevVec_Fer & src,
    RevVec_Fer & dst,
    std::vector<unsigned> & diff,
    MachineBasicBlock * mbb
) {
    while(diff.size() > 0) {
        unsigned index = diff.back();
        diff.pop_back();
        unsigned src_loc = src.get_loc(index);
        unsigned dst_loc = dst.get_loc(index);
        if(dst_loc != RV_UNDEF && dst_loc != src_loc) {
            assert(src_loc != RV_UNDEF && "Inconsistent state.");
            assert(MRegisterInfo::isPhysicalRegister(dst_loc) &&
                                "Destination location must not be in memory.");
            // locations are different, and we have to insert a copy.
            unsigned v_reg_dst = dst.get_virtual(index);
            unsigned v_reg_src = src.get_virtual(index);
            if(MRegisterInfo::isPhysicalRegister(src_loc)) {
                // Insert a register copy
                this->rs_->add_cp2mbb
                   (dst_loc,v_reg_dst,src_loc,v_reg_src,mbb->getNumber(),true);
            } else {
                // Insert a load
                unsigned slot = src_loc - MRegisterInfo::FirstVirtualRegister;
                this->rs_->inc_load(v_reg_src);
                this->rs_->add_ld2mbb
                 (dst_loc, v_reg_dst, slot, v_reg_src, mbb->getNumber(), true);
            }
            src.remove(index);
            dst.remove(index);
            this->find_safe_operations(src, dst, diff);
        }
    }
}


//===--------------------------------------------------------------------------
// This method replaces each alias of p_dst in vec with the corresponding
// alias of p_src.
//===--------------------------------------------------------------------------
void SSAElim_Fer::replace_regs(
    RevVec_Fer & vec,
    unsigned p_src,
    unsigned p_dst,
    const MRegisterInfo & reg_info
) const {
    // if p_dst is already in vec, we only have to swap p_src and p_dst:
    if(vec.contains_phys(p_dst)) {
// BEGIN DEBUG: *******************************************************
/*
unsigned v_swap = vec.get_virtual_assign_to_phys(p_dst);
Regs_Fer::print_def(*this->mf_, p_src, v_swap);
std::cout << " = ";
Regs_Fer::print_reg(p_dst, v_swap);
*/
// END DEBUG: *********************************************************
        vec.swap(p_src, p_dst);
    } else {
        // The family of a register is either: A, B, C, D, BP, DI
        unsigned src_fm = X86PuzzleBoard_Fer::reg2family(p_src);
        unsigned dst_fm = X86PuzzleBoard_Fer::reg2family(p_dst);

        // It is possible that p_dst is a 32 bit register, and the equivalent
        // 16 bit register is in vec. In this case, we can just swap them.
        if(X86PuzzleBoard_Fer::reg2size(p_dst)==X86PuzzleBoard_Fer::RegSz::E) {
            unsigned lit_dst = X86PuzzleBoard_Fer::find_replacement
                                        (X86PuzzleBoard_Fer::RegSz::X, dst_fm);
            if(vec.contains_phys(lit_dst)) {
                vec.swap(p_src, lit_dst);
                unsigned lit_src =X86PuzzleBoard_Fer::find_replacement
                                        (X86PuzzleBoard_Fer::RegSz::X, src_fm);
                vec.replace_phys(p_src, lit_src);
                vec.replace_phys(lit_dst, p_dst);
                return;
            }
        }

        // The  opposite is also possible: p_dst is 16 bit, and it has the
        // corresponding 32 bits in vec.
        if(X86PuzzleBoard_Fer::reg2size(p_dst)==X86PuzzleBoard_Fer::RegSz::X) {
            unsigned big_dst = X86PuzzleBoard_Fer::find_replacement
                                        (X86PuzzleBoard_Fer::RegSz::E, dst_fm);
            if(vec.contains_phys(big_dst)) {
                vec.swap(p_src, big_dst);
                unsigned big_src =X86PuzzleBoard_Fer::find_replacement
                                        (X86PuzzleBoard_Fer::RegSz::E, src_fm);
                vec.replace_phys(big_dst, p_dst);
                vec.replace_phys(p_src, big_src);
                return;
            }
        }

// BEGIN DEBUG: *******************************************************
//std::vector<Binding_Fer> uses;
// END DEBUG: *********************************************************
        for (const unsigned *a = reg_info.getAliasSet(p_dst); *a; ++a) {
            unsigned index = vec.get_index_phy(*a);
            if(index != RV_UNDEF) {
                // 64, 32, 16, 8 bits
                unsigned sz = X86PuzzleBoard_Fer::reg2size(*a);
                assert(sz != RV_UNDEF && "Register with no size?");
                unsigned new_reg =
                              X86PuzzleBoard_Fer::find_replacement(sz, src_fm);
                assert(new_reg != RV_UNDEF && "Unknown register.");
                vec.update_phys(new_reg, index);
// BEGIN DEBUG: ***********************************************
/*
unsigned v_swap = vec.get_virtual_assign_to_phys(*a);
Regs_Fer::print_def(*this->mf_, new_reg, v_swap);
Binding_Fer bf(v_swap, *a);
uses.push_back(bf);
*/
// END DEBUG: *************************************************
            }
        }
// BEGIN DEBUG: *******************************************************
/*
std::cout << " = ";
for(unsigned u = 0; u < uses.size(); u++) {
    Regs_Fer::print_reg(uses[u].p_reg, uses[u].v_reg);
}
*/
// END DEBUG: *********************************************************
    }
}


//===--------------------------------------------------------------------------
// This function uses register swaps to implement the transfers that remain
// between src and dst, once safe operations no longer apply.
//===--------------------------------------------------------------------------
void SSAElim_Fer::implement_swaps(
    RevVec_Fer & src,
    RevVec_Fer & dst,
    MachineBasicBlock * mbb
) {
// BEGIN DEBUG: ***********************************************************
//std::cout << "Inserting swaps at the end of " << mbb->getNumber() << "\n";
// END DEBUG: *************************************************************
    bool has_changed = false;

    do {
        has_changed = swap_big_registers(src, dst, mbb);

        // The sequence of swaps may have left many copies to be implemented. We
        // execute one round of safe operations to solve them:
        std::vector<unsigned> diff_aux;
        find_safe_operations(src, dst, diff_aux);
        if(diff_aux.size() > 0) {
            has_changed = true;
            implement_safe_operations(src, dst, diff_aux, mbb);
        }

        has_changed = has_changed || swap_small_registers(src,dst, mbb);
    } while(has_changed);
}


//===--------------------------------------------------------------------------
// This method returns true if it is possible to swap p_src with p_dst. The
// swapping can be prevented if the smallest alias of p_dst is smaller than
// the smallest alias of p_src.
//===--------------------------------------------------------------------------
bool can_swap(
    const MRegisterInfo & reg_info,
    RevVec_Fer & vec,
    unsigned p_dst,
    unsigned p_src
) {
    unsigned src_min_size = X86PuzzleBoard_Fer::minimum_size(p_src);
    for (const unsigned *a = reg_info.getAliasSet(p_dst); *a; ++a) {
        unsigned index = vec.get_index_phy(*a);
        if(index != RV_UNDEF) {
            if(X86PuzzleBoard_Fer::minimum_size(*a) < src_min_size) {
                return false;
            }
        }
    }
    return true;
}


//===--------------------------------------------------------------------------
// This function swaps registers of 32 and 16 bits, but do nothing about the
// little registers of 8 bits.
//===--------------------------------------------------------------------------
bool SSAElim_Fer::swap_big_registers(
    RevVec_Fer & src,
    RevVec_Fer & dst,
    MachineBasicBlock * mbb
) {
    const TargetMachine & target_machine = this->mf_->getTarget();
    const MRegisterInfo & reg_info = * target_machine.getRegisterInfo();
    bool has_changed = false;

    for(unsigned u = 0; u < src.size(); u++) {
        unsigned src_loc = src.get_loc(u);
        unsigned dst_loc = dst.get_loc(u);
        if(src_loc == RV_UNDEF) {
            assert(dst_loc == RV_UNDEF && "Inconsistent mapping.");
        } else if(
            MRegisterInfo::isPhysicalRegister(src_loc) &&
            // Skip the 8-bit registers:
            X86PuzzleBoard_Fer::reg2size(src_loc) !=
                                                X86PuzzleBoard_Fer::RegSz::H &&
            X86PuzzleBoard_Fer::reg2size(src_loc) !=
                                                X86PuzzleBoard_Fer::RegSz::L &&
            can_swap(reg_info, src, dst_loc, src_loc)
        ) {
            assert((MRegisterInfo::isPhysicalRegister(dst_loc)) &&
                                  "Target of swap must be physical register.");
            unsigned v_reg_src = src.get_virtual(u);
            unsigned v_reg_dst = dst.get_virtual(u);
            if(src_loc != dst_loc) {
                unsigned src_eax = src_loc;
                unsigned dst_eax = dst_loc;
                // We cannot swap 16-bit registers, because this would cause
                // values stores in the corresponding 32-bit registers to be
                // wrong. If dst_loc and src_loc are 16-bit, we must swap the
                // equivalente 32-bit registers.
                if(X86PuzzleBoard_Fer::size(dst_loc) == 16) {
                    assert(X86PuzzleBoard_Fer::size(src_loc == 16) &&
                                                        "Swaps of wrong size");
                    dst_eax = X86PuzzleBoard_Fer::get_ex_reg(dst_loc);
                    src_eax = X86PuzzleBoard_Fer::get_ex_reg(src_loc);
                    assert(src_eax != RG_UNDEF && "Undefined swap source");
                    assert(dst_eax != RG_UNDEF && "Undefined swap destination");
                }
                this->rs_->add_sw2mbb(dst_eax, v_reg_dst, src_eax, v_reg_src,
                                                             mbb->getNumber());
// BEGIN DEBUG: ***********************************************
//Regs_Fer::print_def(*this->mf_, dst_loc, v_reg_dst);
// END DEBUG: *************************************************
                replace_regs(src, src_loc, dst_loc, reg_info);
// BEGIN DEBUG: ***********************************************
//Regs_Fer::print_reg(src_loc, v_reg_src);
//std::cout << " ; // swap \n";
// END DEBUG: *************************************************
                has_changed = true;
            } else {
                this->rs_->add_cp2mbb
                  (dst_loc,v_reg_dst,src_loc,v_reg_src,mbb->getNumber(), true);
            }
            src.remove(u);
            dst.remove(u);
        }
    }

    return has_changed;
}


//===--------------------------------------------------------------------------
// This function swaps the eight bit registers.
//===--------------------------------------------------------------------------
bool SSAElim_Fer::swap_small_registers(
    RevVec_Fer & src,
    RevVec_Fer & dst,
    MachineBasicBlock * mbb
) {
    bool has_changed = false;

    for(unsigned u = 0; u < src.size(); u++) {
        unsigned src_loc = src.get_loc(u);
        unsigned dst_loc = dst.get_loc(u);
        if(src_loc == RV_UNDEF) {
            assert(dst_loc == RV_UNDEF && "Inconsistent mapping.");
        } else if(MRegisterInfo::isPhysicalRegister(src_loc)) {
            assert(
                (X86PuzzleBoard_Fer::reg2size(src_loc) ==
                                                X86PuzzleBoard_Fer::RegSz::H ||
                 X86PuzzleBoard_Fer::reg2size(src_loc) ==
                                               X86PuzzleBoard_Fer::RegSz::L) &&
                "Big register in small register swaping"
            );
            assert((MRegisterInfo::isPhysicalRegister(dst_loc)) &&
                                  "Target of swap must be physical register.");
            unsigned v_reg_src = src.get_virtual(u);
            unsigned v_reg_dst = dst.get_virtual(u);
            if(src_loc != dst_loc) {
                this->rs_->add_sw2mbb(dst_loc, v_reg_dst, src_loc, v_reg_src,
                                                             mbb->getNumber());
// BEGIN DEBUG: ***********************************************
/*
unsigned v_swap_src = src.get_virtual_assign_to_phys(dst_loc);
Regs_Fer::print_def(*this->mf_, dst_loc, v_reg_dst);
Regs_Fer::print_def(*this->mf_, src_loc, v_swap_src);
std::cout << " = ";
Regs_Fer::print_reg(dst_loc, v_swap_src);
Regs_Fer::print_reg(src_loc, v_reg_src);
std::cout << " ; // swap \n";
*/
// END DEBUG: *************************************************
                src.swap(dst_loc, src_loc);
                has_changed = true;
            } else {
                this->rs_->add_cp2mbb
                  (dst_loc,v_reg_dst,src_loc,v_reg_src,mbb->getNumber(), true);
            }
            src.remove(u);
            dst.remove(u);
        }
    }

    return has_changed;
}


//===--------------------------------------------------------------------------
// This method performs the color transfer for basic blocks that do not contain
// phi-functions.
//===--------------------------------------------------------------------------
void SSAElim_Fer::insert_transfer_at_end(
    MachineBasicBlock * dst_mbb,
    MachineBasicBlock * src_mbb,
    unsigned phi_column
) {
// BEGIN DEBUG: ***********************************************************
//std::cout << "Inserting transfers from " << src_mbb->getNumber() << " to "
//                                   << dst_mbb->getNumber() << std::endl;
// END DEBUG: *************************************************************
    const std::map<unsigned, unsigned> & dst_map =
                                 this->pf_->get_head_map(dst_mbb->getNumber());
    // Fills the destination vector with the virtuals in phi-functions first:
    RevVec_Fer rv_dst(dst_map.size(), this->num_p_regs_, this->rm_);
    std::set<unsigned> phi_regs;
    insert_phi_parameters(dst_mbb, rv_dst, 0, dst_map, phi_regs);

    // Now, go through the map, and add each virtual not yet seen into rv_dst:
    unsigned index = phi_regs.size();
    std::map<unsigned, unsigned>::const_iterator dst_i, dst_e;
    for(dst_e=dst_map.end(), dst_i=dst_map.begin(); dst_i != dst_e; dst_i++) {
        unsigned dst_reg = dst_i->first;
        if(phi_regs.find(dst_reg) == phi_regs.end()) {
            unsigned dst_loc = dst_i->second;
            rv_dst.push(dst_reg, dst_loc, index++);
        }
    }

    index = phi_regs.size();
    phi_regs.clear();

    // Add the parameters of the phi-function to the source map.
    const std::map<unsigned, unsigned> & src_map =
                                 this->pf_->get_tail_map(src_mbb->getNumber());
    RevVec_Fer rv_src(dst_map.size(), this->num_p_regs_, this->rm_);
    insert_phi_parameters(dst_mbb, rv_src, phi_column, src_map, phi_regs);

    // Add the registers that are not in phi-functions:
    for( ; index < rv_dst.size(); index++) {
        unsigned src_dst_v_reg = rv_dst.get_virtual(index);
        std::map<unsigned, unsigned>::const_iterator src_loc =
                                                   src_map.find(src_dst_v_reg);
        assert(src_loc != src_map.end() &&
           "Variable in destination block has no correspondent in src block.");
        rv_src.push(src_dst_v_reg, src_loc->second, index);
    }
/*
const TargetMachine & target_machine = this->mf_->getTarget();
const MRegisterInfo & reg_info = * target_machine.getRegisterInfo();
std::cout << "Source vector from block " << src_mbb->getNumber() << std::endl;
rv_src.dump(&reg_info);
std::cout << "Destination vector:" << dst_mbb->getNumber() << std::endl;
rv_dst.dump(&reg_info);
*/

    // Reload the variables that are marked as load_local by RA_Results.
    reload_locals(rv_src, rv_dst, src_mbb);

    // Perform coalescing:
    implement_coalescing(rv_src, rv_dst, src_mbb);

    // Implement the stores, which are always safe:
    implement_stores(rv_src, rv_dst, src_mbb);

    // Diff will contain the indexes of src/dst that can be implemented via
    // safe operations.
    std::vector<unsigned> diff;
    find_safe_operations(rv_src, rv_dst, diff);
    implement_safe_operations(rv_src, rv_dst, diff, src_mbb);

    // Apply swaps to fix ordering:
    implement_swaps(rv_src, rv_dst, src_mbb);

    // Ok, we have added all the swaps that should be added, but it is possible
    // that there are some loads that must be inserted. We have to use a new
    // round of safe operations:
    assert(diff.size() == 0 && "Diff vector is not empty.");
    find_safe_operations(rv_src, rv_dst, diff);
    implement_safe_operations(rv_src, rv_dst, diff, src_mbb);

    // Verifies if the vectors are now empty:
    bool empty = true;
    for(unsigned u = 0; u < rv_src.size(); u++) {
        unsigned src_loc = rv_src.get_loc(u);
        unsigned dst_loc = rv_dst.get_loc(u);
        if(src_loc != RV_UNDEF || dst_loc != RV_UNDEF) {
            empty = false;
            break;
        }
    }
    assert(empty && "There are copies yet to be removed.");
}


//===--------------------------------------------------------------------------
// This method fills the data structures with the data that they must have in
// order to start the processing of intervals.
//===--------------------------------------------------------------------------
void SSAElim_Fer::initialize_data_structures(MachineFunction & mf) {
    this->mf_ = & mf;
    this->pf_ = & getAnalysis<Partition_Fer>();
    this->pa_ = & getAnalysis<PhiAnalysis_Fer>();
    this->rs_ = this->pf_->get_RA_result();
    this->rm_ = this->pf_->get_register_mapping();

    // Get the number of machine registers:
    const TargetMachine & target_machine = mf.getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    this->num_p_regs_ = reg_info->getNumRegs();
}


//===--------------------------------------------------------------------------
// Visits the basic blocks and lists of intervals assigning registers to
// virtual variables.
//===--------------------------------------------------------------------------
bool SSAElim_Fer::runOnMachineFunction(MachineFunction & mf) {
    if(SSA_based) {
        this->initialize_data_structures(mf);
        MachineBasicBlock * entry = mf.begin();
        std::set<MachineBasicBlock*> visited;
        df_ext_iterator<MachineBasicBlock*> dfi = df_ext_begin(entry, visited);
        df_ext_iterator<MachineBasicBlock*> endi = df_ext_end(entry, visited);
        for (; dfi != endi; ++dfi) {
            this->visit(*dfi);
        }
    }
    return false;
}
